// -*-c++-*-

/******************************************************
 * @COPYRIGHT:Jun Masaki @ iesl,ies,sit 2009 
 ******************************************************/

////////////////////////////////////////////////////////

#ifndef ABSTRACT_PG_METHOD_H
#define ABSTRACT_PG_METHOD_H

#include <string>
#include <vector>
#include <rcsc/player/player_agent.h>

class Learning_Draiver;
class Learning_Item;

/*******************************************************
 *
 *******************************************************/
class Abstract_PGMethod {
  friend class Learning_Driver;

private:

protected:

  bool M_initialized_flag;

  bool M_executable_flag;

  bool M_learning_flag;

  bool M_saving_flag;

  int M_unum;

  std::string M_settingFileName;

  std::string M_learningDataFileName;

  std::vector<Learning_Item> M_settingItems;  

  std::vector<Learning_Item> M_dataItems;

  std::vector<double> M_actionProbabilities;

  std::vector< std::vector<double> > M_formula_eweights;

public:

  //////////////////////////////////////////////////////////

private:

  bool isInitialized(void) const
  {
    return M_initialized_flag;
  }

  bool isExecutable(void) const
  {
    return M_executable_flag;
  }

  bool learns(void) const
  {
    return M_learning_flag;
  }

  bool writesData(void) const
  {
    return M_saving_flag;
  }

  bool setMyUnum( const int unum )
  {
    if(0 <= unum && unum <= 11){
      M_unum = unum;
      return true;
    }
    return false;
  }

  bool setDefaultSettings(void);

  virtual bool processWhenInitialized(void)=0;

  const std::string & getSettingFileName(void) const
  {
    return M_settingFileName;
  }

  Learning_Item * getSettingItem( const int item_num );

  const std::string & getLearningDataFileName(void) const
  {
    return M_learningDataFileName;
  }

  Learning_Item * getLearningDataItem( const int item_num );

  virtual bool processWhenReadData(void)=0;

  virtual bool makeAction( const rcsc::PlayerAgent * agent )=0;

  virtual int getActionSize(void) const =0;

  virtual double getTemperature(void) const =0;

  virtual int getHeuristicsSize(void) const =0;

  virtual double getWeight( const int heuristics )const =0;

  virtual double U( const int heuristics 
		    , const int action 
		    , const rcsc::PlayerAgent * agent )=0;

  bool storeProb( const double probability )
  {
    M_actionProbabilities.push_back(probability);
    return true;
  }

  virtual bool setAction( const int action )=0;

  virtual bool isEpisodeEnd( const rcsc::PlayerAgent * agent )=0;

  virtual bool updateRewards( const rcsc::PlayerAgent * agent )=0;

  virtual bool updateWeights( const rcsc::PlayerAgent * agent )=0;

  virtual bool endTheEpisode( const rcsc::PlayerAgent * agent )=0;

  virtual bool processWhenSavedLearningData(void)=0;

protected:

public:

  virtual bool processWhenExecuted( const rcsc::PlayerAgent * agent )=0;

};


/*******************************************************
 *
 *******************************************************/

class Learning_Item {

private:
  
  std::string M_name;

  int M_param_int;

  double M_param_double;

  bool M_param_bool;

  std::string M_param_string;

  bool M_param_setted_flag;

  bool M_no_bool_param_flag;

  bool M_pass_flag;

  ////////////////////////////////////////////////////////

private:

public:

  Learning_Item( const std::string & name , const bool pass_flag )
    :M_name( name )
  {
    M_param_setted_flag = false; 
    M_pass_flag = pass_flag;
  }

  const std::string & getName(void) const
  {
    return M_name;
  }

  bool passes(void) const
  {
    return M_pass_flag;
  }

  bool hasSettedParam(void) const
  {
    return M_param_setted_flag;
  }

  bool hasSettedNoBool(void) const
  {
    return M_no_bool_param_flag;
  }

  bool setParam( const double param );

  double getParam( const double ) const
  {
    return M_param_double;
  }

  bool setParam( const bool param );

  const bool getParam( const bool ) const
  {
    return M_param_bool;
  }

  bool setParam( const std::string & param );

  const std::string & getParam( const std::string & ) const
  {
    return M_param_string;
  }
};

#endif
