// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"
#include "bhv_fs_opponent_pass_cut_move.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_BasicMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_BasicMove"
                        ,__FILE__, __LINE__ );

    // tackle
    if ( Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) )
    {
        return true;
    }

    const rcsc::WorldModel & wm = agent->world();

    // check ball owner

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    //offensive intercept
    if ( ! wm.existKickableTeammate()
         && wm.ball().pos().x > -35
         && ( self_min <= 4
              || ( self_min < mate_min + 1
                   && self_min < opp_min + 4 )
              )
         )	
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d:offensive intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept2008().execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }
    //defensive intercept
    else if ( ! wm.existKickableTeammate()
         && self_min < mate_min + 1
         && self_min < opp_min + 2 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d:defensive intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept2008(false).execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }

    const double dash_power = getDashPower( agent, M_home_pos );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    const rcsc::PlayerObject * nearest_mate
        = static_cast< rcsc::PlayerObject * >( 0 );
     nearest_mate = wm.teammatesFromBall().front();

//前からブロック
    //Block
    if(doBlockDribble(agent)
       &&wm.existKickableOpponent()
       &&!wm.existKickableTeammate()
       &&wm.self().pos().x < wm.ball().pos().x
       &&wm.ball().distFromSelf()<=10
       &&self_min < mate_min
       &&wm.ball().pos().x > -30 )
     {
      return true;
/*
      rcsc::Vector2D block_point(wm.ball().inertiaPoint( self_min + 2 ));
      if ( wm.ball().pos().y > 0 )
        block_point.y-=2;
      else
        block_point.y+=2;

  std::cout << wm.self().unum() << ":[" << wm.time().cycle()
	    << "] block move from front:(" 
	    << block_point.x << "," << block_point.y << ")"
	    << std::endl;

      if ( ! rcsc::Body_GoToPoint( block_point, dist_thr, dash_power
                                 ).execute( agent ) )
       {
        rcsc::Body_TurnToBall().execute( agent );
       }
       agent->setNeckAction( new rcsc::Neck_TurnToBall() );
       return true;*/
     }

//後ろからブロック
    //後ろからの追っかけ 基本的に後ろからは追いつきにくい
    if(wm.existKickableOpponent()
       &&!wm.existKickableTeammate()
       &&wm.self().pos().x > wm.ball().pos().x - 2
       &&wm.ball().distFromSelf()<=10)
     {
      rcsc::Vector2D block_point(wm.ball().inertiaPoint( self_min + 10 ));
      if ( wm.ball().pos().y > 0 )
        block_point.y-=2;
      else
        block_point.y+=2;

  std::cout << wm.self().unum() << ":[" << wm.time().cycle()
	    << "] block move from back:(" 
	    << block_point.x << "," << block_point.y << ")"
	    << std::endl;

      if ( rcsc::Body_GoToPoint(block_point, dist_thr, rcsc::ServerParam::i().maxDashPower()
                               ).execute( agent ) )
       {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return true;
       }       
     }      

    //PassCut
    if(wm.existKickableOpponent()
       &&wm.self().pos().x < wm.ball().pos().x
       &&wm.self().pos().x > -20
       &&wm.ball().distFromSelf()>5)
     {
      if(Bhv_FSOpponentPassCutMove(M_home_pos).execute(agent))
        return true;
     }

    // go to home position

    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( M_home_pos );

    if ( ! rcsc::Body_GoToPoint( M_home_pos, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Bhv_BasicMove::getDashPower( const rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = rcsc::ServerParam::i().maxDashPower();
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( //wm.defenseLineX() > wm.self().pos().x
	wm.existKickableOpponent()
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
      {
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  __FILE__": correct DF line. keep max power" );
      // keep max power
      dash_power = rcsc::ServerParam::i().maxDashPower();
      }
    else if ( ! wm.existKickableTeammate()
              && opp_min < self_min
              && opp_min < mate_min
              && target_point.x < wm.self().pos().x
              && target_point.x < -25.0
              && wm.ball().inertiaPoint( opp_min ).x < -25.0 )
      {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": defense dash" );
        dash_power = rcsc::ServerParam::i().maxDashPower();
      }
    else if ( s_recover_mode )
      {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

	rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": recovering" );
      }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
      {
        dash_power = rcsc::ServerParam::i().maxDashPower();
	rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": in offside area. dash_power=%.1f",
                            dash_power );
      }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 
	      && wm.countOpponentsIn(rcsc::Circle2D(wm.ball().pos()
						    ,5
					 )
				     ,5,false
		  )<=1)
    {
	dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": exist kickable teammate. dash_power=%.1f",
                            dash_power );
    }
    // normal
    else
    {
	dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": normal mode dash_power=%.1f",
                            dash_power );
    }
    
    return dash_power;
}


bool
Bhv_BasicMove::doBlockDribble( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                      __FILE__": execute(). tackle" );
        return true;
    }

    const int opp_min = wm.interceptTable()->opponentReachCycle();
    const rcsc::PlayerObject * opp = wm.interceptTable()->fastestOpponent();

    if ( ! opp )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                      __FILE__": execute(). no fastest opponent" );
        return false;
    }

    const rcsc::PlayerType & self_type = wm.self().playerType();
    const double control_area = self_type.kickableArea() - 0.3;
    const double max_moment = rcsc::ServerParam::i().maxMoment();// * ( 1.0 - ServerParam::i().playerRand() );

    const double first_my_speed = wm.self().vel().r();
    const rcsc::Vector2D first_ball_pos = wm.ball().inertiaPoint( opp_min );

    rcsc::Vector2D opp_target( -44.0, 0.0 );
    if ( first_ball_pos.x > -25.0
         && first_ball_pos.absY() > 20.0 )
    {
        opp_target.y = ( first_ball_pos.y > 0.0
                         ? 20.0
                         : -20.0 );
    }

    const rcsc::Vector2D first_ball_vel
        = ( opp_target - first_ball_pos ).setLengthVector( opp->playerTypePtr()->realSpeedMax() );


    rcsc::dlog.addText( rcsc::Logger::TEAM,
                  __FILE__": execute(). virtual intercept. opp_min=%d bpos=(%.1f %.1f) bvel=(%.1f %.1f)",
                  opp_min,
                  first_ball_pos.x, first_ball_pos.y,
                  first_ball_vel.x, first_ball_vel.y );

    rcsc::Vector2D target_point = rcsc::Vector2D::INVALIDATED;
    int total_step = -1;

    for ( int cycle = 1; cycle <= 20; ++cycle )
    {
        int n_turn = 0;
        int n_dash = 0;
        rcsc::AngleDeg dash_angle = wm.self().body();

        rcsc::Vector2D ball_pos = inertia_n_step_point( first_ball_pos,
                                                  first_ball_vel,
                                                  cycle,
                                                  rcsc::ServerParam::i().ballDecay() );
        rcsc::Vector2D my_pos = wm.self().inertiaPoint( cycle + opp_min );
        rcsc::Vector2D target_rel = ball_pos - my_pos;
        double target_dist = target_rel.r();
        rcsc::AngleDeg target_angle = target_rel.th();

        double stamina = wm.self().stamina();
        double effort = wm.self().effort();
        double recovery = wm.self().recovery();

        // turn
        double angle_diff = ( target_angle - wm.self().body() ).abs();
        double turn_margin = 180.0;
        if ( control_area < target_dist )
        {
            turn_margin = rcsc::AngleDeg::asin_deg( control_area / target_dist );
        }
        turn_margin = std::max( turn_margin, 15.0 );
        double my_speed = first_my_speed;
        while ( angle_diff > turn_margin )
        {
            angle_diff -= self_type.effectiveTurn( max_moment, my_speed );
            my_speed *= self_type.playerDecay();
            ++n_turn;
        }

        if ( n_turn > cycle + opp_min )
        {
            continue;
        }

        if ( n_turn > 0 )
        {
            angle_diff = std::max( 0.0, angle_diff );
            dash_angle = target_angle;
            if ( ( target_angle - wm.self().body() ).degree() > 0.0 )
            {
                dash_angle += angle_diff;
            }
            else
            {
                dash_angle += angle_diff;
            }

            self_type.predictStaminaAfterWait( rcsc::ServerParam::i(),
                                               n_turn,
                                               &stamina,
                                               &effort,
                                               recovery );
        }

        // dash
        n_dash = self_type.cyclesToReachDistance( target_dist );
        self_type.predictStaminaAfterDashes( rcsc::ServerParam::i(),
                                             rcsc::ServerParam::i().maxPower(),
                                             n_dash,
                                             &stamina,
                                             &effort,
                                             &recovery );
        if ( stamina < rcsc::ServerParam::i().recoverDecThrValue() + 300.0 )
        {
            continue;
        }

        if ( n_turn + n_dash <= cycle + opp_min )
        {
            target_point = ball_pos;
            total_step = n_turn + n_dash;
            break;
        }
    }

    if ( ! target_point.valid() )
    {
        return false;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                  __FILE__": execute(). virtual intercept. pos=(%.1f %.1f) total_step=%d",
                  target_point.x, target_point.y,
                  total_step );

    if ( ! rcsc::Body_GoToPoint( target_point, 0.5,
                           rcsc::ServerParam::i().maxPower()
                           ).execute( agent ) )
    {
        return false;
    }
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addMessage( "BlockDribVirtualIntercept" );
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    return true;
}
