
/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fs_direct_pass.h"

#include <rcsc/player/player_agent.h>

#include <rcsc/common/server_param.h>

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

/*-------------------------------------------------------------------*/
/*!

*/

bool
Bhv_FSDirectPass::execute( rcsc::PlayerAgent * agent )
{
  /////////////////////////////////////////////////////////////////////
  //executeはbody_passを流用
  rcsc::dlog.addText( rcsc::Logger::ACTION,
		      "%s:%d: FSPass. execute()"
		      ,__FILE__, __LINE__ );
  
  if ( ! agent->world().self().isKickable() )
    {
      std::cerr << __FILE__ << ": " << __LINE__
		<< " not ball kickable!"
		<< std::endl;
      rcsc::dlog.addText( rcsc::Logger::ACTION,
			  "%s:%d:  not kickable"
			  ,__FILE__, __LINE__ );
      return false;
    }
  
  rcsc::Vector2D target_point(50.0, 0.0);
  double first_speed = 0.0;
  int receiver_unum = 0;

  if ( ! get_best_pass( agent->world()
			, &target_point, &first_speed, &receiver_unum ) )
    {
      return false;
    }
  
  agent->debugClient().addMessage( "pass" );
  agent->debugClient().setTarget( target_point );
  
  int kick_step = ( agent->world().gameMode().type() != rcsc::GameMode::PlayOn
		    && agent->world().gameMode().type() != rcsc::GameMode::GoalKick_
		    ? 1
		    : 3 );
  if ( ! rcsc::Body_SmartKick( target_point,
			       first_speed,
			       first_speed * 0.96,
			       kick_step ).execute( agent ) )
    {
      if ( agent->world().gameMode().type() != rcsc::GameMode::PlayOn
	   && agent->world().gameMode().type() != rcsc::GameMode::GoalKick_ )
        {
	  first_speed = std::min( agent->world().self().kickRate() * rcsc::ServerParam::i().maxPower(),
				  first_speed );
	  rcsc::Body_KickOneStep( target_point,
				  first_speed
				  ).execute( agent );
	  rcsc::dlog.addText( rcsc::Logger::ACTION,
			      __FILE__": execute() one step kick" );
        }
      else
        {
	  rcsc::dlog.addText( rcsc::Logger::ACTION,
			      __FILE__": execute() failed to pass kick." );
	  return false;
        }
    }
  
  // reset intention
  agent->setIntention( static_cast< rcsc::SoccerIntention * >( 0 ) );

  if ( agent->config().useCommunication()
       && receiver_unum != rcsc::Unum_Unknown )
    {
      rcsc::dlog.addText( rcsc::Logger::ACTION,
			  __FILE__": execute() set pass communication." );
      rcsc::Vector2D target_buf = target_point - agent->world().self().pos();
      target_buf.setLength( 1.0 );
      
      agent->addSayMessage( new rcsc::PassMessage( receiver_unum,
						   target_point + target_buf,
						   agent->effector().queuedNextBallPos(),
						   agent->effector().queuedNextBallVel() ) );
    }
  
  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
  /*
  std::cout << agent->world().self().unum()
	    << ":[" << agent->world().time().cycle()
	    << "] FSDirectPass:" << receiver_unum << "@(" 
	    << target_point.x << ":" << target_point.y 
	    << ")" << std::endl;
  */
  return true;
}

///////////////////////////////////////////////////////////////////////
// 
bool
Bhv_FSDirectPass::get_best_pass( const rcsc::WorldModel & world
				 , rcsc::Vector2D * target_point 
				 , double * first_speed 
				 , int * receiver_unum )
{ 
  static rcsc::GameTime S_last_calc_time( 0, 0 );
  static bool S_last_calc_valid = false;
  static rcsc::Vector2D S_last_calc_target;
  static double S_last_calc_speed = 0.0;
  static int S_last_calc_receiver_unum = rcsc::Unum_Unknown;
  
  //1サイクルに何回も計算しません
  if ( S_last_calc_time == world.time() )
    {
      if ( S_last_calc_valid )
        {
	  if ( target_point )
            {
	      *target_point = S_last_calc_target;
            }
	  if ( first_speed )
            {
	      *first_speed = S_last_calc_speed;
            }
	  if ( receiver_unum )
            {
	      *receiver_unum = S_last_calc_receiver_unum;
            }
	  return true;
        }
      return false;
    }  


  S_last_calc_time = world.time();
  S_last_calc_valid = false;

  ////////////////////////////////////////////////////////////////////////
  //パスコースを探索
  bool exist_pass_course = true;
  double first_player_posX = -52.5;
  rcsc::PlayerPtrCont::const_iterator 
    player_it = world.teammatesFromSelf().begin();
  const rcsc::PlayerPtrCont::const_iterator
    player_end = world.teammatesFromSelf().end();
  for(;player_it!=player_end;++player_it){
    //////////////////////////////////////////////////////////////////////
    //rejecters
    if ( (*player_it)->goalie() && (*player_it)->pos().x < -22.0 )
      {
	// goalie is rejected.
	continue;
      }
    if ( (*player_it)->posCount() > 3 )
      {
	// low confidence players are rejected.
	continue;
      }
    if ( (*player_it)->pos().x > world.offsideLineX() + 1.0 )
      {
	// offside players are rejected.
	continue;
      }
    if ( (*player_it)->pos().x < world.ball().pos().x - 25.0 )
      {
	// too back
	continue;
      }

    //パス先までの距離と角度
    const double target_dist 
      = world.self().pos().dist( (*player_it)->pos() );
    const double target_angle 
      = ( (*player_it)->pos() - world.self().pos() ).dir().degree();
    
    //初速度計算
    double end_speed  = 1.5;
    double first_speed = 100.0;

    if((*player_it)->pos().x < world.self().pos().x - 5)
      first_speed
	= rcsc::calc_first_term_geom_series_last
	( 0.1,
	  target_dist,
	  rcsc::ServerParam::i().ballDecay() );
    else if((*player_it)->pos().x < world.self().pos().x + 2)
      first_speed
	= rcsc::calc_first_term_geom_series_last
	( 0.7,
	  target_dist,
	  rcsc::ServerParam::i().ballDecay() );
    else
      do{
	first_speed
	  = rcsc::calc_first_term_geom_series_last
	  ( end_speed,
	    target_dist,
	    rcsc::ServerParam::i().ballDecay() );
	if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
	  {
	    break;
	  }
	end_speed -= 0.1;
      }while ( end_speed > 0.8 );

    first_speed = std::min(first_speed,rcsc::ServerParam::i().ballSpeedMax());

    //到達時間を計算
    double arrival_cycle_len 
      = rcsc::calc_length_geom_series( first_speed,
				       target_dist,
				       rcsc::ServerParam::i().ballDecay() );
    if(arrival_cycle_len < 0)
      continue;
    /////////////////////////////////////////////////////////////////////////
    //パスコースが空いてるか探索
    exist_pass_course = true;
    const rcsc::PlayerPtrCont::const_iterator
      opp_end = world.opponentsFromSelf().end();
    for(rcsc::PlayerPtrCont::const_iterator 
	  opp_it = world.opponentsFromSelf().begin();
	opp_it != opp_end;++opp_it)
      {      
	//oppがcycle時にballに到達できるか
	for(int cycle=0;cycle < arrival_cycle_len;cycle++){
	  rcsc::Vector2D ball_through_pos
	    = world.self().pos() 
	    + rcsc::Vector2D::polar2vector(rcsc::calc_sum_geom_series(first_speed
								      ,rcsc::ServerParam::i().ballDecay()
								      ,cycle)
					   ,target_angle);
	  double opp_dist_toThroughPos
	    = ball_through_pos.dist( (*opp_it)->pos() );
	  double opp_reachable_dist 
	    = rcsc::ServerParam::i().defaultPlayerSpeedMax() * cycle;
	  if((*opp_it)->playerTypePtr() != NULL)
	    opp_reachable_dist 
	      = (*opp_it)->playerTypePtr()->playerSpeedMax() * cycle;
	  
	  //oppに到達されてしまうので、このパスコースは空いてない
	  if(opp_dist_toThroughPos < opp_reachable_dist){
	    exist_pass_course = false;
	    break;
	  }
	}
	//パスコースが空いてないので、oppのループを抜ける
	if(!exist_pass_course)
	  break;
	
	//パス先でもoppに到達されるか
	double opp_dist_toTarget
	  = (*player_it)->pos().dist( (*opp_it)->pos() );
	double opp_reachable_dist_ball_arrived
	  =  rcsc::ServerParam::i().defaultPlayerSpeedMax() * arrival_cycle_len;
	if((*opp_it)->playerTypePtr() != NULL)
	  opp_reachable_dist_ball_arrived
	    = (*opp_it)->playerTypePtr()->playerSpeedMax() * arrival_cycle_len;
	if(opp_dist_toTarget < opp_reachable_dist_ball_arrived){
	  exist_pass_course = false;
	  break;
	}
      }
    //パスコースが空いてれば、このパス先を選択。前のプレイヤーほど優先
    if(exist_pass_course && first_player_posX < (*player_it)->pos().x){
      S_last_calc_target = (*player_it)->pos();
      S_last_calc_speed =  first_speed;
      S_last_calc_receiver_unum = (*player_it)->unum();
      S_last_calc_valid = true;
      first_player_posX = (*player_it)->pos().x;
      rcsc::dlog.addText( rcsc::Logger::ACTION,
			  "%s:%d: get_best_pass() target=(%.1f %.1f)"
			  " speed=%.3f  receiver=%d"
			  ,__FILE__, __LINE__,
			  S_last_calc_target.x, S_last_calc_target.y,
			  S_last_calc_speed,
			  S_last_calc_receiver_unum );
      
    }
  }
  if ( S_last_calc_valid )
    {
      if ( target_point )
        {
	  *target_point = S_last_calc_target;
        }
      if ( first_speed )
        {
	  *first_speed = S_last_calc_speed;
        }
      if ( receiver_unum )
        {
	  *receiver_unum = S_last_calc_receiver_unum;
        }
      
      rcsc::dlog.addText( rcsc::Logger::ACTION,
		    "%s:%d: best pass (%.2f, %.2f). speed=%.2f. receiver=%d"
			  ,__FILE__, __LINE__,
			  S_last_calc_target.x, S_last_calc_target.y,
			  S_last_calc_speed, S_last_calc_receiver_unum );
    }
  return S_last_calc_valid;
}
