
/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fs_goto_receivable_area.h"

#include <rcsc/common/server_param.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_turn_to_point.h>

#include <rcsc/player/player_agent.h>
#include "bhv_basic_move.h"

/*-------------------------------------------------------------------*/
/*!

*/

bool
Bhv_FSGotoReceivableArea::execute( rcsc::PlayerAgent * agent )
{
  if(!agent->world().existKickableTeammate()
     || agent->world().ball().pos().dist(agent->world().self().pos()) < 5)
    return false;
  
  const rcsc::Vector2D ball_pos = agent->world().ball().pos();
  
  int startAngle = -100;
  int angleOffset = 50;
  
  int fewest_area_num_first=-1;
  int min_opps_count=5;
  
  //first search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,18.0
			     ,i*angleOffset + startAngle
			     ,(i+1)*angleOffset + startAngle);
    int opps_count 
      = agent->world().countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_first=i;
    } 
  }
  
  if(fewest_area_num_first==-1)
    return false;
  
  int fewest_area_num_second=-1;
  min_opps_count=3;
  //second search
  for(int i=0;i<5;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,15.0
			     ,fewest_area_num_first*angleOffset 
			     + i*angleOffset/5 + startAngle
			     ,fewest_area_num_first*angleOffset 
			     + (i+1)*angleOffset/5 + startAngle);
    int opps_count 
      = agent->world().countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_second=i;
    } 
  }
  
  if(fewest_area_num_second==-1)
    return false;
  
  int fewest_area_num_third=-1;
  min_opps_count=1;
  //third search
  for(int i=0;i<4;i++){
    rcsc::Sector2D checkArea(ball_pos
			     ,0.0,12
			     ,fewest_area_num_first*angleOffset 
			     + fewest_area_num_second*angleOffset/5
			     + i*angleOffset/5/4 + startAngle
			     ,fewest_area_num_first*angleOffset 
			     + fewest_area_num_second*angleOffset/5
			     + (i+1)*angleOffset/5/4 + startAngle);
    int opps_count 
      = agent->world().countOpponentsIn(checkArea,7,false);
    if(opps_count<min_opps_count){
      min_opps_count=opps_count;
      fewest_area_num_third=i;
    } 
  }

  if(fewest_area_num_third==-1){
    rcsc::Vector2D target_pos
      = rcsc::Vector2D().setPolar(10.0
				  ,fewest_area_num_first*angleOffset
				  + fewest_area_num_second*angleOffset/5
				  + angleOffset/5/2 + startAngle);
    target_pos += ball_pos;
    if(target_pos.dist(agent->world().self().pos())>20
       || target_pos.x > agent->world().offsideLineX() + 1
       || target_pos.dist(M_home_pos) > 20)
      return false;
    
    if(!rcsc::Body_GoToPoint(target_pos
			     ,0.5,Bhv_BasicMove::getDashPower(agent,target_pos)
			     ,5,false,false,12).execute(agent)){
      rcsc::Body_TurnToPoint(target_pos).execute( agent );
    }
    /*
    std::cout << agent->world().self().unum() 
	      << ":[" << agent->world().time().cycle()
	      << "] go to recievable area(2nd):(" 
	      << target_pos.x << "," << target_pos.y << ")"
	      << ": ball pos("
	      << ball_pos.x << "," << ball_pos.y << ")"
	      << std::endl;
    */
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    return true;
  }

  rcsc::Vector2D target_pos
    = rcsc::Vector2D().setPolar(10.0
				,fewest_area_num_first*angleOffset
				+ fewest_area_num_second*angleOffset/5
				+ fewest_area_num_third*angleOffset/5/4 
				+ angleOffset/5/4/2 + startAngle);
  target_pos += ball_pos;
  if(target_pos.dist(agent->world().self().pos()) > 20
     || target_pos.x > agent->world().offsideLineX() + 1
     || target_pos.dist(M_home_pos) > 20)
    return false;
			    
  if(!rcsc::Body_GoToPoint(target_pos
			   ,0.5,Bhv_BasicMove::getDashPower(agent,target_pos)
			   ,5,false,false,12).execute(agent)){
    rcsc::Body_TurnToPoint(target_pos).execute( agent );
  }
  /*
  std::cout << agent->world().self().unum() 
	    << ":[" << agent->world().time().cycle()
	    << "] go to recievable area(3rd):(" 
	    << target_pos.x << "," << target_pos.y << ")"
	    << " :ball pos:("
	    << ball_pos.x << "," << ball_pos.y << ")"
	    << std::endl;
  */
  agent->setNeckAction(new rcsc::Neck_TurnToBall());
  return true;
}
