
/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fs_kickin_mark_opponent.h"

#include <rcsc/player/player_agent.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include <rcsc/common/server_param.h>

#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_turn_to_ball.h>

#include "bhv_basic_move.h"
#include <rcsc/common/logger.h>
#include <rcsc/action/body_intercept.h>

/*-------------------------------------------------------------------*/
/*!

*/

bool
Bhv_FSKickinMarkOpponent::execute( rcsc::PlayerAgent * agent ){
  const rcsc::WorldModel & wm = agent->world();

  ////////////////////////////////////////////////////////////////////////
  //ホームポジションから離れすぎてるときは、reject
  if(wm.self().pos().dist(M_home_pos) > 25)
    return false;
  
  if(wm.ball().pos().x > rcsc::ServerParam::i().pitchHalfLength() - rcsc::ServerParam::i().penaltyAreaLength()
     && wm.ball().pos().absY() < rcsc::ServerParam::i().penaltyAreaHalfWidth())
    return goalKickMark(agent);
  

  ////////////////////////////////////////////////////////////////////////
  //マーク対象を探す
  rcsc::Vector2D mark_point = wm.ball().pos();
  bool mark_flag = false;
  const rcsc::PlayerPtrCont::const_iterator
    opps_end = wm.opponentsFromSelf().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  opps_it = wm.opponentsFromSelf().begin();
	opps_it != opps_end; ++opps_it ){

    //パスされそうな敵以外はとばす、遠い敵もとばす
    //敵とボールとの距離 > 30、敵と自分との距離 > 20、ボールから敵までパスコースが空いてない
    //敵とボールとの距離 < 10は絶対領域みたいで入れない
    //home_pos > 20も飛ばす
    if((*opps_it)->goalie()
       || (*opps_it)->pos().dist(wm.ball().pos()) < 10
       || (*opps_it)->pos().dist(wm.ball().pos()) > 30
       || (*opps_it)->pos().dist(wm.self().pos()) > 20
       || wm.existTeammateIn(rcsc::Sector2D(wm.ball().pos()
					 ,0.0
					 ,wm.ball().pos().dist((*opps_it)->pos())
					 ,((*opps_it)->pos() - wm.ball().pos()).dir().degree() - 5
					 ,((*opps_it)->pos() - wm.ball().pos()).dir().degree() + 5
					 )
			  ,5,false)
       || (*opps_it)->pos().dist(M_home_pos) > 25
       )
      continue;

    //味方がすでに、その敵をマークしてないとき、その敵をマークする
    const rcsc::PlayerObject * nearest_mate 
      = wm.getTeammateNearestTo((*opps_it)->pos(),5,NULL);
    if(nearest_mate != NULL
       && (nearest_mate->pos().dist((*opps_it)->pos()) * 0.6 > 
	   wm.self().pos().dist((*opps_it)->pos())
	   || nearest_mate->unum() > wm.self().unum()
	   && nearest_mate->pos().dist((*opps_it)->pos()) * 2.0 > 
	   wm.self().pos().dist((*opps_it)->pos())
	   )
       ){
      mark_flag = true;
      mark_point.x = (9 * (*opps_it)->pos().x + wm.ball().pos().x) / 10;
      mark_point.y = (9 * (*opps_it)->pos().y + wm.ball().pos().y) / 10;
      break;
    }

  }
  
  if(!mark_flag)
    return false;
  
  double dash_power = Bhv_BasicMove::getDashPower(agent,mark_point);
  int cycle=100;
  bool back_mode=false;
  bool save_recovery=true;
  double dir_thr=12;
  
  if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			 ,5,true)
     <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,false)){
    dash_power=std::min(dash_power*1.05,rcsc::ServerParam::i().maxPower());
    save_recovery=false;
  }
  
  //ボールがマーク相手にきそうな時は、ボールに向かっていく
  for(int i=0;i<7;i++){
    if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5){
      save_recovery=false;
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: intercept"
			  ,__FILE__, __LINE__ );
      rcsc::Body_Intercept2008(save_recovery).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      return true;
    }
  }
  
  if(! rcsc::Body_GoToPoint( mark_point,
			     0.25,dash_power
			     ,cycle,back_mode
			     ,save_recovery,dir_thr
			     ).execute( agent )
     ){
    if(wm.self().posPrev().dist(mark_point) > 1)
      rcsc::Body_TurnToPoint(mark_point).execute(agent);
    else
      rcsc::Body_TurnToBall().execute(agent);
  }
  if ( wm.ball().distFromSelf() < 18.0 ){
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
  }else{
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
  }
  return true;

}

bool
Bhv_FSKickinMarkOpponent::goalKickMark( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

 ////////////////////////////////////////////////////////////////////////
  //マーク対象を探す
  rcsc::Vector2D mark_point = wm.ball().pos();
  bool mark_flag = false;
  const rcsc::PlayerPtrCont::const_iterator
    opps_end = wm.opponentsFromSelf().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  opps_it = wm.opponentsFromSelf().begin();
	opps_it != opps_end; ++opps_it ){

    //パスされそうな敵以外はとばす、遠い敵もとばす
    //敵とボールとの距離 > 30、敵と自分との距離 > 20、ボールから敵までパスコースが空いてない
    //敵とボールとの距離 < 10は絶対領域みたいで入れない
    //home_pos > 20も飛ばす
    if((*opps_it)->goalie()
       || (*opps_it)->pos().dist(wm.ball().pos()) > 30
       || wm.existTeammateIn(rcsc::Sector2D(wm.ball().pos()
					 ,0.0
					 ,wm.ball().pos().dist((*opps_it)->pos())
					 ,((*opps_it)->pos() - wm.ball().pos()).dir().degree() - 15
					 ,((*opps_it)->pos() - wm.ball().pos()).dir().degree() + 15
					 )
			  ,5,false)
       )
      continue;
    
    for(rcsc::PlayerPtrCont::const_iterator
	  opps_it_markee = wm.opponentsFromSelf().begin();
	opps_it_markee != opps_end; ++opps_it_markee){
      if((*opps_it_markee)->goalie()
	 || (*opps_it_markee)->pos().dist(wm.ball().pos()) < 10
	 || (*opps_it_markee)->pos().dist(wm.self().pos()) > 20
	 || (*opps_it_markee)->pos().dist((*opps_it)->pos()) > 25
	 || wm.existTeammateIn(rcsc::Sector2D((*opps_it)->pos()
					      ,0.0
					      ,(*opps_it)->pos().dist((*opps_it_markee)->pos())
					      ,((*opps_it_markee)->pos() - (*opps_it)->pos()).dir().degree() - 5
					      ,((*opps_it_markee)->pos() - (*opps_it)->pos()).dir().degree() + 5
					      )
			       ,5,false)
	 || (*opps_it_markee)->pos().dist(M_home_pos) > 25
	 )
	continue;

      //味方がすでに、その敵をマークしてないとき、その敵をマークする
      const rcsc::PlayerObject * nearest_mate 
	= wm.getTeammateNearestTo((*opps_it_markee)->pos(),5,NULL);
      if(nearest_mate != NULL
	 && (nearest_mate->pos().dist((*opps_it_markee)->pos()) * 0.6 > 
	     wm.self().pos().dist((*opps_it_markee)->pos())
	     || nearest_mate->unum() > wm.self().unum()
	     && nearest_mate->pos().dist((*opps_it_markee)->pos()) * 2.0 > 
	     wm.self().pos().dist((*opps_it_markee)->pos())
	     )
	 ){
	mark_flag = true;
	mark_point.x = (9 * (*opps_it_markee)->pos().x + (*opps_it)->pos().x) / 10;
	mark_point.y = (9 * (*opps_it_markee)->pos().y + (*opps_it)->pos().y) / 10;
	break;
      } 
    }
  
  if(!mark_flag)
    break;
  }
  
  if(!mark_flag)
    return false;
  
  double dash_power = Bhv_BasicMove::getDashPower(agent,mark_point);
  int cycle=100;
  bool back_mode=false;
  bool save_recovery=true;
  double dir_thr=12;
  
  if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			 ,5,true)
     <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,false)){
    dash_power=std::min(dash_power*1.05,rcsc::ServerParam::i().maxPower());
    save_recovery=false;
  }
  
  //ボールがマーク相手にきそうな時は、ボールに向かっていく
  for(int i=0;i<7;i++){
    if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5){
      save_recovery=false;
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: intercept"
			  ,__FILE__, __LINE__ );
      rcsc::Body_Intercept2008(save_recovery).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      return true;
    }
  }
  
  if(! rcsc::Body_GoToPoint( mark_point,
			     0.25,dash_power
			     ,cycle,back_mode
			     ,save_recovery,dir_thr
			     ).execute( agent )
     ){
    if(wm.self().posPrev().dist(mark_point) > 1)
      rcsc::Body_TurnToPoint(mark_point).execute(agent);
    else
      rcsc::Body_TurnToBall().execute(agent);
  }
  if ( wm.ball().distFromSelf() < 18.0 ){
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
  }else{
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
  }
  return true;
}
