
/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fs_kickin_move.h"

#include <rcsc/player/player_agent.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include <rcsc/common/server_param.h>

#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_turn_to_ball.h>

#include "bhv_basic_move.h"

#include "bhv_fs_mark_opponent_forward.h"

/*-------------------------------------------------------------------*/
/*!

*/

bool
Bhv_FSKickinMove::execute( rcsc::PlayerAgent * agent ){
  const rcsc::WorldModel & wm = agent->world();

  if(wm.self().unum() == 6 || wm.self().unum() == 7)
    if(Bhv_FSMarkOpponentForward(M_home_pos).execute(agent))
      return true;

  //ボールから遠いときとホームポジションでパスが受けられそうなときは、reject
  if(wm.ball().pos().dist(wm.self().pos()) > 20
     || wm.ball().pos().dist(M_home_pos) > 30
     || !wm.existOpponentIn(rcsc::Sector2D(wm.ball().pos()
					   ,0.0
					   ,wm.ball().pos().dist(M_home_pos)
					   ,(M_home_pos - wm.ball().pos()).dir() - 15
					   ,(M_home_pos - wm.ball().pos()).dir() + 15)
			    ,5,false
			    )
     || !wm.existTeammateIn(rcsc::Circle2D(wm.ball().pos(),3.5)
			    ,5,false)
     || wm.self().pos().absY() > 33
     || wm.self().pos().absX() > 50)
    return false;
  
  //ボールとの距離と角度
  double current_dist 
    = wm.ball().pos().dist(wm.self().pos());
  const double current_angle 
    = ( wm.self().pos() - wm.ball().pos()).dir().degree();

  //パスを受けられそうなら、ボールのほうを向く
  if(!wm.existOpponentIn(rcsc::Sector2D(wm.ball().pos()
					,0.0
					,current_dist
					,current_angle - 15
					,current_angle + 15)
			 ,5,false
			 )
     && !wm.existTeammateIn(rcsc::Circle2D(wm.ball().pos(),2.5)
			    ,5,false)
     && wm.self().pos().dist(wm.ball().pos()) < 12){
    rcsc::Body_TurnToBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    return true;
  }

  ////////////////////////////////////////////////////////////////////////
  //
  bool move_flag = false;
  rcsc::Vector2D target_pos = M_home_pos;
  while(current_dist > 7){
    int sign = -1;
    current_dist -= 1.4;
    for(int angle = 1;angle < 60;angle++){
      sign *= -1;
      if(wm.existOpponentIn(rcsc::Sector2D(wm.ball().pos()
					   ,0.0
					   ,current_dist
					   ,current_angle - 15 + sign * angle
					   ,current_angle + 15 + sign * angle)
			    ,5,false
			    )
	 )
	continue;

      target_pos = 
	wm.ball().pos()+rcsc::Vector2D::polar2vector(current_dist
						     ,current_angle+sign*angle);
      if(target_pos.absY() < 32
	 && target_pos.absX() < 50
	 && target_pos.dist(M_home_pos) < 30
	 && target_pos.dist(wm.self().pos()) < 20
	 && !wm.existTeammateIn(rcsc::Circle2D(target_pos,2.5)
				 ,5,false)
	 && !wm.existOpponentIn(rcsc::Circle2D(target_pos,5)
				,5,false)
	 ){
	move_flag = true;
	break;
      }
    }

    if(move_flag)
      break;
  }

  
  if(!move_flag)
    return false;
  
  double dash_power = Bhv_BasicMove::getDashPower(agent,target_pos);
  const int cycle = 100;
  const bool back_mode = false;
  bool save_recovery = true;
  const double dir_thr = 12;
  
  //ボールがマーク相手にきそうな時は、ボールに向かっていく
  for(int i=0;i<7;i++){
    if(wm.ball().inertiaPoint(i).dist(target_pos) < 1.5){
      
      save_recovery=false;
      dash_power = rcsc::ServerParam::i().maxPower();
      if(! rcsc::Body_GoToPoint( wm.ball().pos(),
				 0.25,dash_power
				 ,cycle,back_mode
				 ,save_recovery,dir_thr
				 ).execute( agent )
	 ){
	rcsc::Body_TurnToPoint(wm.ball().pos()).execute(agent);
      }
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      return true;
    }
  }
  
  if(! rcsc::Body_GoToPoint( target_pos,
			     0.25,dash_power
			     ,cycle,back_mode
			     ,save_recovery,dir_thr
			     ).execute( agent )
     ){
    if(wm.self().posPrev().dist(target_pos) > 1)
      rcsc::Body_TurnToPoint(target_pos).execute(agent);
    else
      rcsc::Body_TurnToBall().execute(agent);
  }
  if ( wm.ball().distFromSelf() < 18.0 ){
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
  }else{
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
  }
  return true;
}
