// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fs_opponent_pass_cut_move.h"

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_hold_ball.h>

#include <rcsc/player/interception.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/soccer_math.h>
#include <rcsc/math_util.h>

#include "bhv_basic_tackle.h"
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/player/intercept_table.h>

#include "bhv_basic_move.h"

//#define DEBUG


std::vector< Bhv_FSOpponentPassCutMove::PassRoute > Bhv_FSOpponentPassCutMove::S_cached_pass_route;

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_FSOpponentPassCutMove::execute( rcsc::PlayerAgent * agent )
{
  if(!agent->world().existKickableOpponent())
    return false;

  rcsc::dlog.addText( rcsc::Logger::ACTION,
		      "%s:%d: Bhv_Opponent_PassCut. execute()"
		      ,__FILE__, __LINE__ );
  
  // tackle
  if ( Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) )
    {
      return true;
    }
  
  const rcsc::WorldModel & wm = agent->world();
  
  // check ball owner
  
  int self_min = wm.interceptTable()->selfReachCycle();
  int mate_min = wm.interceptTable()->teammateReachCycle();
  int opp_min = wm.interceptTable()->opponentReachCycle();
  
  if ( ! wm.existKickableTeammate()
       && ( self_min < 7
	    || ( self_min < mate_min + 1
		 && self_min < opp_min + 3 )
	    )
       )	
    {
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: intercept"
			  ,__FILE__, __LINE__ );
      rcsc::Body_Intercept2008().execute( agent );
      if ( wm.ball().distFromSelf()
	   < rcsc::ServerParam::i().visibleDistance() )
        {
	  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        }
      else
        {
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
      return true;
    }
  
  //go to mark point

  std::vector<rcsc::Vector2D> target_points;
  std::vector<double> first_speeds;
  std::vector<int> receiver_unums;
  
  if ( ! get_opponent_passes( wm
			      , target_points
			      , first_speeds
			      , receiver_unums ) )
    {
      return false;
    }

  bool valid = false;
  rcsc::Vector2D mark_point;

  //std::cout << "target_points size:" << target_points.size() << std::endl;
  std::vector<rcsc::Vector2D>::iterator pos_it;
  std::vector<rcsc::Vector2D>::const_iterator pos_end = target_points.end();
  //std::vector<double>::iterator speed_it;
  //std::vector<double>::const_iterator speed_end = first_speeds.end();
  for(pos_it = target_points.begin()/*, speed_it = first_speeds.begin()*/;
      pos_it != pos_end /*&& speed_it != speed_end*/; ++pos_it/*, ++speed_it*/){
    if(pos_it->x > -30.0){
      mark_point.x = pos_it->x * 2.0/3.0 + wm.ball().pos().x * 1.0/3.0;
      mark_point.y = pos_it->y * 2.0/3.0 + wm.ball().pos().y * 1.0/3.0;
    }else{
      mark_point = *pos_it;
    }
    /*
    std::cout << wm.self().unum() << ":[" << wm.time().cycle()
	      << "] target_points:(" 
	      << mark_point.x << "," << mark_point.y << ")"
	      << std::endl;
    */
    /*
    double arrival_cycle =
      rcsc::calc_length_geom_series(*speed_it
				    ,mark_point.dist(wm.self().pos())
				    ,rcsc::ServerParam::i().ballDecay());
    */
    if(wm.self().pos().dist(mark_point) < 12
       && !wm.existTeammateIn(rcsc::Circle2D(mark_point,2.0)
			      ,5,false)
       && wm.self().pos().dist(M_home_pos) < 25){
      valid = true;
      break;
    }
  }
  if(!valid)
    return false;

  //ボールがマーク相手にきそうな時は、ボールに向かっていく
  for(int i=0;i<7;i++){
    if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5){
      bool save_recovery=false;
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: intercept"
			  ,__FILE__, __LINE__ );
      rcsc::Body_Intercept2008(save_recovery).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      return true;
    }
  }

  const double dash_power = Bhv_BasicMove::getDashPower( agent, mark_point );

  double dist_thr = wm.ball().distFromSelf() * 0.1;
  if ( dist_thr < 1.0 ) dist_thr = 1.0;
  
  agent->debugClient().addMessage( " PassCutMove%.0f", dash_power );
  agent->debugClient().setTarget( mark_point );
  
  if ( ! rcsc::Body_GoToPoint( mark_point, dist_thr, dash_power
			       ).execute( agent ) )
    {
      rcsc::Body_TurnToBall().execute( agent );
    }
  /*
  std::cout << wm.self().unum() << ":[" << wm.time().cycle()
	    << "] pass cut move:(" 
	    << mark_point.x << "," << mark_point.y << ")"
	    << ": ball pos:("
	    << wm.ball().pos().x << "," << wm.ball().pos().y << ")"
	    << std::endl;
  */
  if ( wm.existKickableOpponent()
       && wm.ball().distFromSelf() < 18.0 )
    {
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
  else
    {
      agent->setNeckAction( new rcsc::Neck_ScanField() );
    }

  return true;
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
bool
Bhv_FSOpponentPassCutMove::get_opponent_passes( const rcsc::WorldModel & world,
					   std::vector<rcsc::Vector2D> & target_points,
					   std::vector<double> & first_speeds,
					   std::vector<int> & receiver_unums )
{
  static rcsc::GameTime S_last_calc_time( 0, 0 );
  static bool S_last_calc_valid = false;
  static std::vector<rcsc::Vector2D> S_last_calc_targets;
  static std::vector<double> S_last_calc_speeds;
  static std::vector<int> S_last_calc_receiver_unums;
  

  if ( S_last_calc_time == world.time() )
    {
      if ( S_last_calc_valid )
        {
	  target_points.swap(S_last_calc_targets);
	  first_speeds.swap(S_last_calc_speeds);
	  receiver_unums.swap(S_last_calc_receiver_unums);
	  return true;
        }
      return false;
    }
  
  S_last_calc_time = world.time();
  S_last_calc_valid = false;
  
  const rcsc::PlayerObject * kicker;
  if(world.existKickableOpponent()){
     kicker = world.getOpponentNearestToBall(5,true);
     if(kicker == NULL)
       return false;
  }else{
    return false;
  }

  // create route
  create_routes( world , kicker );
  S_last_calc_targets.clear();
  S_last_calc_speeds.clear();
  S_last_calc_receiver_unums.clear();

  if ( ! S_cached_pass_route.empty() )
    {
      //std::cout << "size before:" << S_cached_pass_route.size()
      //<< std::endl;
      /*
      //comb sort
      unsigned int gap = (int)S_cached_pass_route.size();
      bool swaps = false;
      if(gap < 2)
	swaps = true;

      while(!swaps || gap > 1) {
	gap /= 1.3;
	if(gap < 1)
	  break;
	if( gap==9 || gap==10 )
	  gap = 11;

	for (unsigned int i = 0;i + gap < S_cached_pass_route.size(); i++){
	  if(S_cached_pass_route.at(i + gap).score_ 
	     > S_cached_pass_route.at(i).score_) {
	    PassRoute temp = S_cached_pass_route.at(i + gap);
	    S_cached_pass_route.at(i + gap) = S_cached_pass_route.at(i);
	    S_cached_pass_route.at(i) = temp;
	    swaps = true;
	  }
	}
      }
      */
      
      std::sort(S_cached_pass_route.begin(),S_cached_pass_route.end()
		,PassRouteScoreCompGreater());
      //std::cout << "size after:" << S_cached_pass_route.size()
      //<< std::endl;

      std::vector< PassRoute >::const_iterator 
	it = S_cached_pass_route.begin();
      std::vector< PassRoute >::const_iterator 
	end = S_cached_pass_route.end();
      for(;it!=end;++it){
	//std::cout << "score:" << it->score_ << std::endl;
	S_last_calc_targets.push_back(it->receive_point_);
	S_last_calc_speeds.push_back(it->first_speed_);
	S_last_calc_receiver_unums.push_back(it->receiver_->unum());
      }

      S_last_calc_valid = true;
    }


    if ( S_last_calc_valid )
    {
      target_points.swap(S_last_calc_targets);
      first_speeds.swap(S_last_calc_speeds);
      receiver_unums.swap(S_last_calc_receiver_unums);
    }

    return S_last_calc_valid;
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
Bhv_FSOpponentPassCutMove::create_routes( const rcsc::WorldModel & world,
				     const rcsc::PlayerObject * kicker)
{
  // reset old info
  S_cached_pass_route.clear();
  
  //基準座標を修正
  // loop candidate teammates
  const rcsc::PlayerPtrCont::const_iterator
    opp_end = world.opponentsFromBall().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  it = world.opponentsFromBall().begin();
	it != opp_end;
	++it )
    {
      if ( (*it)->goalie() && (*it)->pos().x > 22.0 )
        {
	  // goalie is rejected.
	  continue;
        }
      if ( (*it)->pos().x < world.defenseLineX() - 1.0 )
        {
	  // offside players are rejected.
	  continue;
        }
      if ( (*it)->pos().x > world.ball().pos().x + 25.0 )
        {
	  // too back
	  continue;
        }
      
      // create & verify each route
      create_direct_pass( world, kicker, *it );
      create_lead_pass( world, kicker, *it );
      if ( kicker->pos().x < world.defenseLineX() + 20.0 )
        {
	  create_through_pass( world, kicker, *it );
        }
    }

    ////////////////////////////////////////////////////////////////
    // evaluation
  evaluate_routes( world, kicker );
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
Bhv_FSOpponentPassCutMove::create_direct_pass( const rcsc::WorldModel & world,
					  const rcsc::PlayerObject * kicker,
					  const rcsc::PlayerObject * receiver )
{
  static const double MAX_DIRECT_PASS_DIST
    = 0.8 * rcsc::inertia_final_distance( rcsc::ServerParam::i().ballSpeedMax(),
				    rcsc::ServerParam::i().ballDecay() );
#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "Create_direct_pass() to %d(%.1f %.1f)",
		      receiver->unum(),
		      receiver->pos().x, receiver->pos().y );
#endif

  // out of pitch?
  if ( receiver->pos().absX() > rcsc::ServerParam::i().pitchHalfLength() - 3.0
       || receiver->pos().absY() > rcsc::ServerParam::i().pitchHalfWidth() - 3.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ out of pitch" );
#endif
      return;
    }

  /////////////////////////////////////////////////////////////////
  // too far
  if ( receiver->pos().dist(kicker->pos()) > MAX_DIRECT_PASS_DIST )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ over max distance %.2f > %.2f",
			  receiver->pos().dist(kicker->pos()),
			  MAX_DIRECT_PASS_DIST );
#endif
      return;
    }
  // too close
  if ( receiver->pos().dist(kicker->pos()) < 6.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ too close. dist = %.2f  canceled",
			  receiver->pos().dist(kicker->pos()) );
#endif
      return;
    }
  /////////////////
  if ( receiver->pos().x > kicker->pos().x - 5.0 )
    {
      if ( ( receiver->pos() - kicker->pos() ).dir().abs() < 40.0
	   || ( receiver->pos().x < world.offsideLineX() - 10.0 
		&& receiver->pos().x < 0.0 )
	   )
        {
	  // "DIRECT: not defender." <<;
        }
      else
	{
	  // "DIRECT: back.;
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "__ looks back pass. DF Line=%.1f. canceled",
			      world.offsideLineX() );
#endif
	  return;
        }
    }
  
  // not safety area
  if ( receiver->pos().x > 20.0 )
    {
      if ( receiver->pos().x < kicker->pos().x - 13.0 )
        {
	  // safety clear??
        }
      else if ( receiver->pos().x < kicker->pos().x - 5.0
		&& receiver->pos().absY() > 20.0
		&& std::fabs(receiver->pos().y - kicker->pos().y) < 20.0
	)
        {
	  // safety area
        }
      else
        {
	  // dangerous
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "__ receiver is in dangerous area. canceled" );
#endif
	  return;
        }
    }

  /////////////////////////////////////////////////////////////////
  
  // set base target
  rcsc::Vector2D base_player_pos = receiver->pos();
  if ( receiver->velCount() < 3 )
    {
      rcsc::Vector2D fvel = receiver->vel();
      fvel /= rcsc::ServerParam::i().defaultPlayerDecay();
      fvel *= std::min( receiver->velCount() + 1, 2 );
      base_player_pos += fvel;
    }

  const rcsc::Vector2D receiver_rel = base_player_pos - world.ball().pos();
  const double receiver_dist = receiver_rel.r();
  const rcsc::AngleDeg receiver_angle = receiver_rel.th();

#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "__ receiver. predict pos(%.2f %.2f) rel(%.2f %.2f)"
		      "  dist=%.2f  angle=%.1f",
		      base_player_pos.x, base_player_pos.y,
		      receiver_rel.x, receiver_rel.y,
		      receiver_dist,
		      receiver_angle.degree() );
  
#endif
  
  double end_speed = 1.5;
  double first_speed = 100.0;
  do
    {
      first_speed
	= rcsc::calc_first_term_geom_series_last
	( end_speed,
	  receiver_dist,
	  rcsc::ServerParam::i().ballDecay() );
      if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
        {
	  break;
        }
      end_speed -= 0.1;
    }
  while ( end_speed > 0.8 );


  if ( first_speed > rcsc::ServerParam::i().ballSpeedMax() )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ ball first speed= %.3f.  too high. canceled",
			  first_speed );
#endif
      return;
    }


    // add strictly direct pass
  if ( verify_direct_pass( world,
			   kicker,
			   receiver,
			   base_player_pos,
			   receiver_dist,
			   receiver_angle,
			   first_speed ) )
    {
      S_cached_pass_route
	.push_back( PassRoute( DIRECT,
			       kicker,
			       receiver,
			       base_player_pos,
			       first_speed,
			       can_kick_by_one_step( world,
						     kicker,
						     first_speed,
						     receiver_angle )
			       )
		    );
    }
  
  // add kickable edge points
  double kickable_angle_buf = 360.0 * ( rcsc::ServerParam::i().defaultKickableArea()
					/ (2.0 * M_PI * receiver_dist) );
  first_speed *= rcsc::ServerParam::i().ballDecay();

  // right side
  rcsc::Vector2D target_new = world.ball().pos();
  rcsc::AngleDeg angle_new = receiver_angle;
  angle_new += kickable_angle_buf;
  target_new += rcsc::Vector2D::polar2vector(receiver_dist, angle_new);

  if ( verify_direct_pass( world,
			   kicker,
			   receiver,
			   target_new,
			   receiver_dist,
			   angle_new,
			   first_speed ) )
    {
      S_cached_pass_route
	.push_back( PassRoute( DIRECT,
			       kicker,
			       receiver,
			       target_new,
			       first_speed,
			       can_kick_by_one_step( world,
						     kicker,
						     first_speed,
						     angle_new )
			       )
		    );
      
    }
  // left side
  target_new = world.ball().pos();
  angle_new = receiver_angle;
  angle_new -= kickable_angle_buf;
  target_new += rcsc::Vector2D::polar2vector( receiver_dist, angle_new );
  
  if ( verify_direct_pass( world,
			   kicker,
			   receiver,
			   target_new,
			   receiver_dist,
			   angle_new,
			   first_speed ) )
    {
        S_cached_pass_route
            .push_back( PassRoute( DIRECT,
				   kicker,
                                   receiver,
                                   target_new,
                                   first_speed,
                                   can_kick_by_one_step( world,
							 kicker,
                                                         first_speed,
                                                         angle_new )
                                   )
                        );
	rcsc::dlog.addText( rcsc::Logger::PASS,
			    "Pass Success direct unum=%d pos=(%.1f %.1f). first_speed= %.1f",
			    receiver->unum(),
			    target_new.x, target_new.y,
			    first_speed );
    }
  else
    {
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "Pass Failed direct unum=%d pos=(%.1f %.1f). first_speed= %.1f",
			  receiver->unum(),
			  target_new.x, target_new.y,
			  first_speed );
    }
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
Bhv_FSOpponentPassCutMove::create_lead_pass( const rcsc::WorldModel & world,
					const rcsc::PlayerObject * kicker,
					const rcsc::PlayerObject * receiver )
{
  static const double MAX_LEAD_PASS_DIST
    = 0.7 * rcsc::inertia_final_distance( rcsc::ServerParam::i().ballSpeedMax(),
					  rcsc::ServerParam::i().ballDecay() );
  static const
    rcsc::Rect2D shrinked_pitch( rcsc::Vector2D( 3.0,
						 -rcsc::ServerParam::i().pitchHalfWidth() + 3.0 ),
				 rcsc::Size2D( rcsc::ServerParam::i().pitchLength() - 6.0,
					       rcsc::ServerParam::i().pitchWidth() - 6.0 ) );
  
  static const double receiver_dash_speed = 0.8;
  
#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "Create_lead_pass() to %d(%.1f %.1f)",
		      receiver->unum(),
		      receiver->pos().x, receiver->pos().y );
#endif

  /////////////////////////////////////////////////////////////////
  // too far
  if ( receiver->pos().dist(kicker->pos()) > MAX_LEAD_PASS_DIST )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ over max distance %.2f > %.2f",
			  receiver->pos().dist(kicker->pos()), MAX_LEAD_PASS_DIST );
#endif
      return;
    }
  // too close
  if ( receiver->pos().dist(kicker->pos()) < 2.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ too close %.2f",
			  receiver->pos().dist(kicker->pos()) );
#endif
      return;
    }
  if ( receiver->pos().x > kicker->pos().x + 15.0
       && receiver->pos().x > -15.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is back cancel" );
#endif
      return;
    }
  
  if( receiver->pos().x > 10.0
      && std::fabs( receiver->pos().y - kicker->pos().y ) > 20.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is in our field. or Y diff is big" );
#endif
      return;
    }
  
  const rcsc::Vector2D receiver_rel = receiver->pos() - world.ball().pos();
  const double receiver_dist = receiver_rel.r();
  const rcsc::AngleDeg receiver_angle = receiver_rel.th();
  
  const double circum = 2.0 * receiver_dist * M_PI;
  const double angle_step_abs = std::max( 5.0, 360.0 * ( 2.0 / circum ));
  /// angle loop
  double total_add_angle_abs;
  int count;
  for ( total_add_angle_abs = angle_step_abs, count = 0;
	total_add_angle_abs < 31.0 && count < 5;
	total_add_angle_abs += angle_step_abs, ++count )
    {
      /////////////////////////////////////////////////////////////
      // estimate required first ball speed
      const double dash_step
	= ( (angle_step_abs / 360.0) * circum ) / receiver_dash_speed + 2.0;
      
      double end_speed = 1.2 - 0.11 * count; // Magic Number
      double first_speed = 100.0;
      double ball_steps_to_target = 100.0;
      do
        {
	  first_speed
	    = rcsc::calc_first_term_geom_series_last
	    ( end_speed,
	      receiver_dist,
	      rcsc::ServerParam::i().ballDecay() );
	  if ( first_speed < rcsc::ServerParam::i().ballSpeedMax() )
            {
	      break;
            }
	  ball_steps_to_target
	    = rcsc::calc_length_geom_series( first_speed,
					     receiver_dist,
					     rcsc::ServerParam::i().ballDecay() );
	  if ( dash_step < ball_steps_to_target )
            {
	      break;
            }
	  end_speed -= 0.1;
        }
      while ( end_speed > 0.5 );
      
      
      if ( first_speed > rcsc::ServerParam::i().ballSpeedMax() )
        {
	  continue;
        }
      
      /////////////////////////////////////////////////////////////
      // angle plus minus loop
      for ( int i = 0; i < 2; i++ )
        {
	  rcsc::AngleDeg target_angle = receiver_angle;
	  if ( i == 0 )
            {
	      target_angle -= total_add_angle_abs;
            }
	  else
            {
	      target_angle += total_add_angle_abs;
            }
	  
	  // check dir confidence
	  int max_count = 100, ave_count = 100;
	  world.dirRangeCount( target_angle, 20.0,
			       &max_count, NULL, &ave_count );
	  if ( max_count > 9 || ave_count > 3 )
            {
	      continue;
            }
	  
	  const rcsc::Vector2D target_point
	    = world.ball().pos()
	    + rcsc::Vector2D::polar2vector(receiver_dist, target_angle);

	  /////////////////////////////////////////////////////////////////
	  // ignore back pass
	  if ( target_point.x > 0.0
	       && target_point.x >  kicker->pos().x )
            {
	      continue;
            }

	  if ( target_point.x > 0.0
	       && target_point.x > receiver->pos().x + 3.0 )
            {
	      continue;
            }
	  if ( target_point.x > receiver->pos().x + 6.0 )
            {
	      continue;
            }
	  // out of pitch
	  if ( ! shrinked_pitch.contains( target_point ) )
            {
	      continue;
            }
	  // not safety area
	  if ( target_point.x < 10.0 )
            {
	      if ( target_point.x > world.offsideLineX() - 10.0 )
                {
		  continue;
                }
	      else if ( target_point.x < kicker->pos().x - 20.0
			&& std::fabs(target_point.y - kicker->pos().y) < 20.0 )
                {
		  // safety clear ??
                }
	      else if ( target_point.x < kicker->pos().x - 5.0 // forward than me
			&& std::fabs( target_point.y - kicker->pos().y ) < 20.0
			) // out side of me
                {
		  // safety area
                }
	      else if ( target_point.x < world.offsideLineX() - 20.0 )
                {
		  // safety area
                }
	      else
                {
		  // dangerous
		  continue;
                }
            }
	  /////////////////////////////////////////////////////////////////
	  
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "__ lead pass to (%.1f %.1f). first_speed= %.3f. angle = %.1f",
			      target_point.x, target_point.y,
			      first_speed, target_angle.degree() );
#endif
	  // add lead pass route
	  // this methid is same as through pass verification method.
	  if ( verify_through_pass( world,
				    kicker,
				    receiver,
				    receiver->pos(),
				    target_point,
				    receiver_dist,
				    target_angle,
				    first_speed,
				    ball_steps_to_target ) )
            {
	      S_cached_pass_route
		.push_back( PassRoute( LEAD,
				       kicker,
				       receiver,
				       target_point,
				       first_speed,
				       can_kick_by_one_step( world,
							     kicker,
							     first_speed,
							     target_angle ) ) );
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "Pass Success lead unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f",
				  receiver->unum(),
				  target_point.x, target_point.y,
				  target_angle.degree(),
				  first_speed );
            }
	  else
            {
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "Pass Failed lead unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f",
				  receiver->unum(),
				  target_point.x, target_point.y,
				  target_angle.degree(),
				  first_speed );
	      
            }
        }
    }
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
Bhv_FSOpponentPassCutMove::create_through_pass(const rcsc::WorldModel & world,
					  const rcsc::PlayerObject * kicker,
					  const rcsc::PlayerObject * receiver)
{
  static const double MAX_THROUGH_PASS_DIST
    = 0.9 * rcsc::inertia_final_distance( rcsc::ServerParam::i().ballSpeedMax(),
					  rcsc::ServerParam::i().ballDecay() );
  
  ////////////////////////////////////////////////////////////////
  static const
    rcsc::Rect2D shrinked_pitch( rcsc::Vector2D( 3.0,
						 -rcsc::ServerParam::i().pitchHalfWidth() + 3.0 ),
				 rcsc::Size2D( rcsc::ServerParam::i().pitchLength() - 6.0,
					       rcsc::ServerParam::i().pitchWidth() - 6.0 ) );
  
  static const double receiver_dash_speed = 1.0;
  
  static const double S_min_dash = 5.0;
  static const double S_max_dash = 25.0;
  static const double S_dash_range = S_max_dash - S_min_dash;
  static const double S_dash_inc = 4.0;
  static const int S_dash_loop
    = static_cast< int >( std::ceil( S_dash_range / S_dash_inc ) ) + 1;
  
  static const rcsc::AngleDeg S_min_angle = -20.0;
  static const rcsc::AngleDeg S_max_angle = 20.0;
  static const double S_angle_range = ( S_min_angle - S_max_angle ).abs();
  static const double S_angle_inc = 10.0;
  static const int S_angle_loop
    = static_cast< int >( std::ceil( S_angle_range / S_angle_inc ) ) + 1;
  
#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "Create_through_pass() to %d(%.1f %.1f)",
		      receiver->unum(),
		      receiver->pos().x, receiver->pos().y );
#endif
  
  /////////////////////////////////////////////////////////////////
  // check receiver position
  if ( receiver->pos().x < world.defenseLineX() + 0.5 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is offside" );
#endif
      return;
    }
  if ( receiver->pos().x > kicker->pos().x + 10.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is back" );
#endif
      return;
    }
  if ( std::fabs( receiver->pos().y - kicker->pos().y ) > 35.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver Y diff is big" );
#endif
      return;
    }
  if ( world.offsideLineX() > 0.0
       && receiver->pos().x > world.offsideLineX() + 15.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is near to offsied line" );
#endif
      return;
    }
  if ( world.defenseLineX() > 30.0
       && receiver->pos().x > world.defenseLineX() + 15.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver is far from defense line" );
#endif
      return;
    }
  if ( (receiver->pos() - kicker->pos()).dir().abs() > 135.0 )
    {
#ifdef DEBUG
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "__ receiver angle is too back" );
#endif
      return;
    }
  
  // angle loop
  rcsc::AngleDeg dash_angle = S_min_angle;
  for ( int i = 0; i < S_angle_loop; ++i, dash_angle += S_angle_inc )
    {
      const rcsc::Vector2D base_dash = rcsc::Vector2D::polar2vector( 1.0, dash_angle );
      
      // dash dist loop
      double dash_dist = S_min_dash;
      for ( int j = 0; j < S_dash_loop; ++j, dash_dist += S_dash_inc )
        {
	  rcsc::Vector2D target_point = base_dash;
	  target_point *= dash_dist;
	  target_point += receiver->pos();
	  
	  if ( ! shrinked_pitch.contains( target_point ) )
            {
	      // out of pitch
	      continue;
            }
	  
	  if ( target_point.x > kicker->pos().x - 3.0 )
            {
	      continue;
            }
	  
	  const rcsc::Vector2D target_rel = target_point - world.ball().pos();
	  const double target_dist = target_rel.r();
	  const rcsc::AngleDeg target_angle = target_rel.th();
	  
	  // check dir confidence
	  int max_count = 100, ave_count = 100;
	  world.dirRangeCount( target_angle, 20.0,
			       &max_count, NULL, &ave_count );
	  if ( max_count > 9 || ave_count > 3 )
            {
	      continue;
            }
	  
	  if ( target_dist > MAX_THROUGH_PASS_DIST ) // dist range over
            {
	      continue;
            }
	  
	  if ( target_dist < dash_dist ) // I am closer than receiver
            {
	      continue;
            }
	  
	  const double dash_step = dash_dist / receiver_dash_speed;// + 5.0;//+2.0
	  
	  double end_speed = 0.81;//0.65
	  double first_speed = 100.0;
	  double ball_steps_to_target = 100.0;
	  do
            {
	      first_speed
		= rcsc::calc_first_term_geom_series_last
		( end_speed,
		  target_dist,
		  rcsc::ServerParam::i().ballDecay() );
	      if ( first_speed > rcsc::ServerParam::i().ballSpeedMax() )
                {
		  end_speed -= 0.1;
		  continue;
                }
	      
	      ball_steps_to_target
		= rcsc::calc_length_geom_series( first_speed,
						 target_dist,
						 rcsc::ServerParam::i().ballDecay() );
	      if ( dash_step < ball_steps_to_target )
                {
		  break;
                }
	      
	      end_speed -= 0.1;
            }
	  while ( end_speed > 0.5 );
	  
	  if ( first_speed > rcsc::ServerParam::i().ballSpeedMax()
	       || dash_step > ball_steps_to_target )
            {
	      continue;
            }
	  
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "__ throug pass to (%.1f %.1f). first_speed= %.3f",
			      target_point.x, target_point.y,
			      first_speed );
#endif
	  if ( verify_through_pass( world,
				    kicker,
				    receiver,
				    receiver->pos(),
				    target_point, target_dist, target_angle,
				    first_speed,
				    ball_steps_to_target ) )
            {
	      S_cached_pass_route
		.push_back( PassRoute( THROUGH,
				       kicker,
				       receiver,
				       target_point,
				       first_speed,
				       can_kick_by_one_step( world,
							     kicker,
							     first_speed,
							     target_angle )
				       )
			    );
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "Pass Success through pass unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f dash_step=%.1f ball_step=%.1f",
				  receiver->unum(),
				  target_point.x, target_point.y,
				  target_angle.degree(),
				  first_speed,
				  dash_step,
				  ball_steps_to_target );
            }
	  else
            {
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "Pass Failed through unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f dash_step-%.1f ball_step=%.1f",
				  receiver->unum(),
				  target_point.x, target_point.y,
				  target_angle.degree(),
				  first_speed,
				  dash_step,
				  ball_steps_to_target );
            }
        }
    }
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
bool
Bhv_FSOpponentPassCutMove::verify_direct_pass( const rcsc::WorldModel & world,
			       const rcsc::PlayerObject * kicker,
                               const rcsc::PlayerObject * /*receiver*/,
                               const rcsc::Vector2D & target_point,
                               const double & target_dist,
                               const rcsc::AngleDeg & target_angle,
                               const double & first_speed )
{
  static const double player_dash_speed = 1.0;
  
  const rcsc::Vector2D first_vel = rcsc::Vector2D::polar2vector( first_speed, target_angle );
  const rcsc::AngleDeg minus_target_angle = -target_angle;
  const double next_speed = first_speed * rcsc::ServerParam::i().ballDecay();

#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "____ verify direct pass to(%.1f %.1f). first_speed=%.3f. angle=%.1f",
		      target_point.x, target_point.y,
		      first_speed, target_angle.degree() );
#endif

  const rcsc::PlayerPtrCont::const_iterator
    mate_end = world.teammatesFromBall().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  it = world.teammatesFromBall().begin();
	it != mate_end;
	++it )
    {
      if ( (*it)->posCount() > 10 ) continue;
      if ( (*it)->isGhost() && (*it)->posCount() >= 4 ) continue;
      
      const double virtual_dash
	= player_dash_speed * 0.8 * std::min( 5, (*it)->posCount() );

      if ( ( ((*it)->pos()-kicker->pos()).dir() - target_angle ).abs() > 100.0 )
        {
	  continue;
        }
      
      if ( (*it)->pos().dist2( target_point ) < 3.0 * 3.0 )
        {
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "______ opp%d(%.1f %.1f) is already on target point(%.1f %.1f).",
			      (*it)->unum(),
			      (*it)->pos().x, (*it)->pos().y,
			      target_point.x, target_point.y );
#endif
	  return false;
        }

      rcsc::Vector2D ball_to_opp = (*it)->pos();
      ball_to_opp -= world.ball().pos();
      ball_to_opp -= first_vel;
      ball_to_opp.rotate( minus_target_angle );

      if ( 0.0 > ball_to_opp.x && ball_to_opp.x > -target_dist )
        {
	  double opp2line_dist = ball_to_opp.absY();
	  opp2line_dist -= virtual_dash;
	  opp2line_dist -= rcsc::ServerParam::i().defaultKickableArea();
	  opp2line_dist -= 0.1;
	  
	  if ( opp2line_dist < 0.0 )
            {
#ifdef DEBUG
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "______ opp%d(%.1f %.1f) can reach pass line. rejected. vdash=%.1f",
				  (*it)->unum(),
				  (*it)->pos().x, (*it)->pos().y,
				  virtual_dash );
#endif
	      return false;
            }
	  
	  const double ball_steps_to_project
	    = rcsc::calc_length_geom_series( next_speed,
					     ball_to_opp.x,
					     rcsc::ServerParam::i().ballDecay() );

	  if ( ball_steps_to_project < 0.0
	       || opp2line_dist / player_dash_speed < ball_steps_to_project )
            {
#ifdef DEBUG
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "______ opp%d(%.1f %.1f) can reach pass line."
				  " ball reach step to project= %.1f",
				  (*it)->unum(),
				  (*it)->pos().x, (*it)->pos().y,
				  ball_steps_to_project );
#endif
	      return false;
            }
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "______ opp%d(%.1f %.1f) cannot intercept.",
			      (*it)->unum(),
			      (*it)->pos().x, (*it)->pos().y );
#endif
        }
    }
  
#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "__ Success!" );
#endif
  return true;
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
bool
Bhv_FSOpponentPassCutMove::verify_through_pass( const rcsc::WorldModel & world,
					   const rcsc::PlayerObject * kicker,
					   const rcsc::PlayerObject * /*receiver*/,
					   const rcsc::Vector2D & receiver_pos,
					   const rcsc::Vector2D & target_point,
					   const double & target_dist,
					   const rcsc::AngleDeg & target_angle,
					   const double & first_speed,
					   const double & /*reach_step*/ )
{
  static const double player_dash_speed = 1.0;
  
  const rcsc::Vector2D first_vel = rcsc::Vector2D::polar2vector( first_speed, target_angle );
  const rcsc::AngleDeg minus_target_angle = -target_angle;
  const double next_speed = first_speed * rcsc::ServerParam::i().ballDecay();
  
  const double receiver_to_target = receiver_pos.dist( target_point );
  
  bool very_aggressive = false;
  if( target_point.x < -28.0
      && target_point.x < world.ball().pos().x - 20.0 )
    {
      very_aggressive = true;
    }
  else  if ( target_point.x < world.defenseLineX() - 15.0
	     && target_point.x < world.ball().pos().x - 15.0 )
    {
      very_aggressive = true;
    }
  else if ( target_point.x < -38.0
	    && target_point.x < world.defenseLineX()
	    && target_point.absY() < 14.0 )
    {
      very_aggressive = true;
    }
  
  const rcsc::PlayerPtrCont::const_iterator
    mate_end = world.teammatesFromBall().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  it = world.teammatesFromBall().begin();
	it != mate_end;
	++it )
    {
        if ( (*it)->posCount() > 10 ) continue;

        if ( (*it)->goalie() )
        {
	  if ( target_point.absY() > rcsc::ServerParam::i().penaltyAreaWidth() - 3.0
	       || target_point.x > -rcsc::ServerParam::i().theirPenaltyAreaLineX() - 2.0 )
            {
                continue;
            }
        }

        const double virtual_dash
            = player_dash_speed * std::min( 2, (*it)->posCount() );
        const double opp_to_target = (*it)->pos().dist( target_point );
        double dist_rate = ( very_aggressive ? 0.8 : 1.0 );
        double dist_buf = ( very_aggressive ? 0.5 : 2.0 );

        if ( opp_to_target - virtual_dash < receiver_to_target * dist_rate + dist_buf )
        {
#ifdef DEBUG
	  rcsc::dlog.addText( rcsc::Logger::PASS,
			      "______ opp%d(%.1f %.1f) is closer than receiver.",
			      (*it)->unum(),
			      (*it)->pos().x, (*it)->pos().y );
#endif
	  return false;
        }

	rcsc::Vector2D ball_to_opp = (*it)->pos();
        ball_to_opp -= world.ball().pos();
        ball_to_opp -= first_vel;
        ball_to_opp.rotate(minus_target_angle);

        if ( 0.0 > ball_to_opp.x && ball_to_opp.x > -target_dist )
        {
            double opp2line_dist = ball_to_opp.absY();
            opp2line_dist -= virtual_dash;
            opp2line_dist -= rcsc::ServerParam::i().defaultKickableArea();
            opp2line_dist -= 0.1;
            if ( opp2line_dist < 0.0 )
            {
#ifdef DEBUG
	      rcsc::dlog.addText( rcsc::Logger::PASS,
				  "______ opp%d(%.1f %.1f) is already on pass line.",
				  (*it)->unum(),
				  (*it)->pos().x, (*it)->pos().y );
#endif
	      return false;
            }
	    
            const double ball_steps_to_project
	      = rcsc::calc_length_geom_series( next_speed,
                                                 ball_to_opp.x,
					       rcsc::ServerParam::i().ballDecay() );
            if ( ball_steps_to_project < 0.0
                 || opp2line_dist / player_dash_speed < ball_steps_to_project )
            {
#ifdef DEBUG
	      rcsc::dlog.addText( rcsc::Logger::PASS,
                              "______ opp%d(%.1f %.1f) can reach pass line."
                              " ball reach step to project= %.1f",
                              (*it)->unum(),
                              (*it)->pos().x, (*it)->pos().y,
                              ball_steps_to_project );
#endif
                return false;
            }
	    
        }
    }

#ifdef DEBUG
  rcsc::dlog.addText( rcsc::Logger::PASS,
		      "__ Success!" );
#endif
  return true;
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
Bhv_FSOpponentPassCutMove::evaluate_routes( const rcsc::WorldModel & world, 
				       const rcsc::PlayerObject * kicker)
{
  const rcsc::AngleDeg min_angle = -45.0;
  const rcsc::AngleDeg max_angle = 45.0;

  const std::vector< PassRoute >::iterator it_end = S_cached_pass_route.end();
  for ( std::vector< PassRoute >::iterator it = S_cached_pass_route.begin();
	it != it_end;
	++it )
    {
      //-----------------------------------------------------------
      double opp_dist_rate = 1.0;
      {
	double opp_dist = 100.0;
	world.getTeammateNearestTo( it->receive_point_, 20, &opp_dist );
	opp_dist_rate = std::pow( 0.99, std::max( 0.0, 30.0 - opp_dist ) );
      }
      //-----------------------------------------------------------
      double x_diff_rate = 1.0;
      {
	double x_diff = it->receive_point_.x - kicker->pos().x;
	x_diff_rate = std::pow( 0.98,
				std::max( 0.0, 30.0 - x_diff ) );
      }
      //-----------------------------------------------------------
      double receiver_move_rate = 1.0;
      //-----------------------------------------------------------
      double pos_conf_rate = std::pow( 0.98, it->receiver_->posCount() );
      //-----------------------------------------------------------
      double dir_conf_rate = 1.0;
      {
	rcsc::AngleDeg pass_angle = ( it->receive_point_ - kicker->pos() ).th();
	int max_count = 0;
	world.dirRangeCount( pass_angle, 20.0, &max_count, NULL, NULL );
	
	dir_conf_rate = std::pow( 0.95, max_count );
      }
      //-----------------------------------------------------------
      double offense_rate
	= std::pow( 0.98,
		    std::max( 5.0, std::fabs( it->receive_point_.y
					      - world.ball().pos().y ) ) );
      //-----------------------------------------------------------
      const rcsc::Sector2D sector( it->receive_point_,
				   0.0, 10.0,
				   min_angle, max_angle );
      
      // opponent check with goalie
      double front_space_rate = 1.0;
      if ( world.existTeammateIn( sector, 10, true ) )
        {
	  front_space_rate = 0.95;
        }
	
      //-----------------------------------------------------------
      it->score_ = 1000.0;
      it->score_ *= opp_dist_rate;
      it->score_ *= x_diff_rate;
      it->score_ *= receiver_move_rate;
      it->score_ *= pos_conf_rate;
      it->score_ *= dir_conf_rate;
      it->score_ *= offense_rate;
      it->score_ *= front_space_rate;
      
      if ( it->one_step_kick_ )
	{
	  it->score_ *= 1.2;
	}
      
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "PASS Score %6.2f -- to%d(%.1f %.1f) recv_pos(%.1f %.1f) type %d "
			    " speed=%.2f",
			  it->score_,
			  it->receiver_->unum(),
			  it->receiver_->pos().x, it->receiver_->pos().y,
			  it->receive_point_.x, it->receive_point_.y,
			  it->type_,
			  it->first_speed_ );
      rcsc::dlog.addText( rcsc::Logger::PASS,
			  "____ opp_dist=%.2f x_diff=%.2f pos_conf=%.2f"
			    " dir_conf=%.2f space=%.1f %s",
			  opp_dist_rate, x_diff_rate, pos_conf_rate,
			  dir_conf_rate, front_space_rate,
			  ( it->one_step_kick_ ? "one_step" : "" ) );
    }
  
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
bool
Bhv_FSOpponentPassCutMove::can_kick_by_one_step( const rcsc::WorldModel & world,
				 const rcsc::PlayerObject * kicker,
                                 const double & first_speed,
                                 const rcsc::AngleDeg & target_angle )
{
  rcsc::Vector2D required_accel = rcsc::Vector2D::polar2vector( first_speed, target_angle );
  required_accel -= world.ball().vel();
  return ( kicker->playerTypePtr()->kickRate(rcsc::ServerParam::i()
					   ,world.ball().pos().dist(kicker->pos())
					     ,((world.ball().pos() - kicker->pos()).dir() - kicker->body()).degree()) * rcsc::ServerParam::i().maxPower()
	   > required_accel.r() );
}
