// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_pre_process.h"

#include "strategy.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_before_kick_off.h>
#include <rcsc/action/bhv_emergency.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/bhv_shoot.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/view_synch.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/audio_memory.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/geom/sector_2d.h>
#include <rcsc/action/body_smart_kick.h>

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_PreProcess::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_PreProcess"
                        ,__FILE__, __LINE__ );

    //////////////////////////////////////////////////////////////
    // freezed by tackle effect
    if ( agent->world().self().isFreezed() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: tackle wait. expires= %d"
                            ,__FILE__, __LINE__,
                            agent->world().self().tackleExpires() );
        // face neck to ball
        agent->setViewAction( new rcsc::View_Synch() );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }

    //////////////////////////////////////////////////////////////
    // BeforeKickOff or AfterGoal. should jump to the initial position
    if ( agent->world().gameMode().type() == rcsc::GameMode::BeforeKickOff
         || agent->world().gameMode().type() == rcsc::GameMode::AfterGoal_ )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: before_kick_off"
                            ,__FILE__, __LINE__ );
        rcsc::Vector2D move_point = M_strategy.getBeforeKickOffPos( agent->config().playerNumber() );
        //agent->setViewAction( new rcsc::View_Synch() );
        agent->setViewAction( new rcsc::View_Wide() );

 	//---おまけモーション---
	static int last_our_points = 0;
	static int last_opp_points = 0;
	int current_our_points = 0;
	int current_opp_points = 0;
	if(agent->world().isOurLeft()){
	  current_our_points = agent->world().gameMode().scoreLeft();
	  current_opp_points = agent->world().gameMode().scoreRight();
	}else{
	      current_our_points = agent->world().gameMode().scoreRight();
	      current_opp_points = agent->world().gameMode().scoreLeft();
	}
	if(agent->world().setplayCount()<40)
	  {
	    if(current_our_points > last_our_points)
	      {
		//---勝利モーション---
		int r = agent->world().setplayCount()%360;
		int Unum = agent->world().self().unum();
		move_point.x = 10*cos((Unum+r)*3.14/11)-20;
		move_point.y =  5*sin((Unum+r)*3.14/11);	      }
	    else if(current_opp_points > last_opp_points)
	      {
		//敗北モーション
		int setted_move_points_x[]={
		  -20,-19,-18,-17,-16,-15,-14,-13,-12,-11,
		  -11,-12,-13,-14,-15,-16,-17,-18,-19,-20
		};
		int setted_move_points_y[]={
		  -5,-4,-3,-2,-1,0,1,2,3,4,
		  -5,-4,-3,-2,-1,0,1,2,3,4
		};
		int count = agent->world().setplayCount();
		int unum = agent->world().self().unum()*2;
		int index = (count + unum)%20;
		
		move_point.x=setted_move_points_x[index]*2;
		move_point.y=setted_move_points_y[index]*2;
	      }
	  }else{
	  last_our_points = current_our_points;
	  last_opp_points = current_opp_points;
	}
 	//-----------

        rcsc::Bhv_BeforeKickOff( move_point ).execute( agent );
        return true;
    }

    //////////////////////////////////////////////////////////////
    // my pos is unknown
    if ( ! agent->world().self().posValid() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: invalid my pos"
                            ,__FILE__, __LINE__ );
        // included change view
        rcsc::Bhv_Emergency().execute( agent );
        return true;
    }
    //////////////////////////////////////////////////////////////
    // ball search
    // included change view
    int count_thr = 5;
    if ( agent->world().self().goalie() ) count_thr = 10;
    if ( agent->world().ball().posCount() > count_thr
         || agent->world().ball().rposCount() > count_thr + 3 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: search ball"
                            ,__FILE__, __LINE__ );
        //agent->setViewAction( new rcsc::View_Synch() );
        agent->setViewAction( new rcsc::View_Wide() );
        rcsc::Bhv_NeckBodyToBall().execute( agent );
        return true;
    }

    //////////////////////////////////////////////////////////////
    // default change view

    agent->setViewAction( new rcsc::View_Synch() );


    //////////////////////////////////////////////////////////////
    // check queued action
    if ( agent->doIntention() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: do queued intention"
                            ,__FILE__, __LINE__ );
        return true;
    }

    //////////////////////////////////////////////////////////////
    // check shoot chance
    if ( agent->world().gameMode().type() != rcsc::GameMode::IndFreeKick_
         && agent->world().self().isKickable()
         && rcsc::Bhv_Shoot2008().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": shooted" );

        // reset intention
        agent->setIntention( static_cast< rcsc::SoccerIntention * >( 0 ) );
        return true;
    }

    //////////////////////////////////////////////////////////////
    // check simultaneous kick
    if ( agent->world().gameMode().type() == rcsc::GameMode::PlayOn
         && ! agent->world().self().goalie()
         && agent->world().self().isKickable()
         && agent->world().existKickableOpponent() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: simultaneous kick"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "SimultaneousKick" );
        


        if ( agent->world().self().pos().x > 45.0
             && agent->world().self().pos().absY() < 10.0 )
        {
          rcsc::Vector2D shoot_pos( rcsc::ServerParam::i().pitchHalfLength(), 0.0 );
          const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
          if(goalie->pos().absY()>=0) shoot_pos.y=-rcsc::ServerParam::i().goalHalfWidth()+0.3;
          else shoot_pos.y=rcsc::ServerParam::i().goalHalfWidth()-0.3;
	  rcsc::Body_KickOneStep( shoot_pos,
                                rcsc::ServerParam::i().ballSpeedMax()
                                ).execute( agent );
          agent->setNeckAction( new rcsc::Neck_ScanField() );
          return true;
        }
        else if( agent->world().self().pos().absY() < 10.0 )
        {
	  rcsc::Body_KickOneStep(rcsc::Vector2D(wm.self().pos().x+10,wm.self().pos().y),
                                rcsc::ServerParam::i().ballSpeedMax()
                                ).execute( agent );
          agent->setNeckAction( new rcsc::Neck_ScanField() );
          return true;
        }
        else
        {
              rcsc::Vector2D target_point(47.0,32.0);
	      if(wm.self().pos().y<0)
		target_point *= -1;
	      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
	      double thr = 0.96;
	      int step = 2;
	      double end_speed = 0.35;
	      double target_dist = target_point.dist(wm.ball().pos());

	      if(!wm.existOpponentIn(rcsc::Circle2D(target_point,7)
				     ,7,true)
		 && !wm.existOpponentIn(rcsc::Sector2D(wm.self().pos()
						       ,target_point.dist(wm.self().pos())*0.1
						       ,target_point.dist(wm.self().pos())
						       ,(target_point - wm.self().pos()).dir() - 7
						       ,(target_point - wm.self().pos()).dir() + 7)
					,7,false)
		 )
		end_speed = 0.1;
	      first_speed
		= rcsc::calc_first_term_geom_series_last
		( end_speed,
		  target_dist,
		  rcsc::ServerParam::i().ballDecay() );
	      		
	      first_speed = std::min(first_speed
				     ,rcsc::ServerParam::i().ballSpeedMax());
	      rcsc::Body_SmartKick(target_point,
				   first_speed,
				   first_speed * thr,
				   step).execute(agent);
              agent->setNeckAction( new rcsc::Neck_ScanField() );
              return true;
        }
    }

    //////////////////////////////////////////////////////////////
    // check communication intention
    if ( agent->world().audioMemory().passTime() == agent->world().time()
         && ! agent->world().audioMemory().pass().empty()
         && ( agent->world().audioMemory().pass().front().receiver_
              == agent->world().self().unum() )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: communication receive"
                            ,__FILE__, __LINE__ );
        doReceiveMove( agent );

        return true;
    }

    return false;
}


/*-------------------------------------------------------------------*/
/*!

*/
void
Bhv_PreProcess::doReceiveMove( rcsc::PlayerAgent * agent )
{
    agent->debugClient().addMessage( "IntentionRecv" );

    const rcsc::WorldModel & wm = agent->world();
    int self_min = wm.interceptTable()->selfReachCycle();

    if ( ( ! wm.existKickableTeammate()
           && self_min < 6 )
         || wm.audioMemory().pass().empty() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: PreProcess. Receiver. intercept cycle=%d"
                            ,__FILE__, __LINE__,
                            self_min );
        agent->debugClient().addMessage( "Intercept_1" );
        rcsc::Body_Intercept2008().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        return;
    }

    rcsc::Vector2D receive_pos = wm.audioMemory().pass().front().receive_pos_;

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: PreProcess. Receiver. intercept cycle=%d. go to receive point"
                        ,__FILE__, __LINE__,
                        self_min );
    agent->debugClient().setTarget( receive_pos );
    agent->debugClient().addMessage( "GoTo" );
    rcsc::Body_GoToPoint( receive_pos,
                          0.5,
                          rcsc::ServerParam::i().maxDashPower()
                          ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
}
