// -*-c++-*-

/*********************************************************************
 * @COPYRIGHT:Jun Masaki @ iesl,ies,sit 2009 
 *********************************************************************/

////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_selection_pass.h"

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "learning_driver.h"
#include "pg_method_selection_pass.h"
#include <rcsc/math_util.h>
#include <rcsc/player/say_message_builder.h>

/*-------------------------------------------------------------------*/

bool
Bhv_SelectionPass::execute( rcsc::PlayerAgent * agent )
{
  rcsc::dlog.addText( rcsc::Logger::TEAM,
		      "%s:%d: Bhv_SelectionPass"
		      ,__FILE__, __LINE__ );
  
  ////////////////////////////////////////////////////////////////////
  //Body_Pass().execute()を流用
  if ( ! agent->world().self().isKickable() )
    {
      std::cerr << __FILE__ << ": " << __LINE__
		<< " not ball kickable!"
		<< std::endl;
      rcsc::dlog.addText( rcsc::Logger::ACTION,
		    "%s:%d:  not kickable"
		    ,__FILE__, __LINE__ );
      return false;
    }
  
  rcsc::Vector2D target_point(50.0, 0.0);
  double first_speed = 0.0;
  int receiver_unum = 0;
  bool continue_action = false;
  
  //selection_passに選ばれたreceiver_numとtarget_pointを返す
  if ( ! get_selection_pass( agent, 
			     &target_point, &first_speed, &receiver_unum, 
			     &continue_action) )
    {
      return false;
    }

  agent->debugClient().addMessage( "selection-pass" );
  agent->debugClient().setTarget( target_point );

  int kick_step = ( agent->world().gameMode().type() != rcsc::GameMode::PlayOn
		    && agent->world().gameMode().type() != rcsc::GameMode::GoalKick_
		    ? 1 : 3 );
  double thr = 0.96;
  if ( ! rcsc::Body_SmartKick( target_point,
			       first_speed,
			       first_speed * thr,
			       kick_step ).execute( agent ) )
    {
      if ( agent->world().gameMode().type() != rcsc::GameMode::PlayOn
	   && agent->world().gameMode().type() != rcsc::GameMode::GoalKick_ )
	{
	  first_speed = std::min( agent->world().self().kickRate() 
				  * rcsc::ServerParam::i().maxPower(),
				  first_speed );
	  rcsc::Body_KickOneStep( target_point,
				  first_speed
				  ).execute( agent );
	  rcsc::dlog.addText( rcsc::Logger::ACTION,
			      __FILE__": execute() one step kick" );
	}
      else
	{
	  rcsc::dlog.addText( rcsc::Logger::ACTION,
			      __FILE__": execute() failed to pass kick." );
	  return false;
	}
    }
  
  // reset intention
  agent->setIntention( static_cast< rcsc::SoccerIntention * >( 0 ) );

  if ( agent->config().useCommunication()
       && receiver_unum != rcsc::Unum_Unknown )
    {
      rcsc::dlog.addText( rcsc::Logger::ACTION,
		    __FILE__": execute() set pass communication." );
      rcsc::Vector2D target_buf = target_point - agent->world().self().pos();
      target_buf.setLength( 1.0 );
      
      agent->addSayMessage( new rcsc::PassMessage( receiver_unum,
						   target_point + target_buf,
						   agent->effector().queuedNextBallPos(),
						   agent->effector().queuedNextBallVel() ) );
    }
  //////////////////////////////////////////////////////////////////////
  
  agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
  //行動が実行されたときの処理
  if(!continue_action)
    PGMethod_SelectionPass::getInstance().processWhenExecuted(agent);
  std::cout << agent->world().self().unum() 
	    << ":[" << agent->world().time().cycle()
	    << "] executed action:target:"<< receiver_unum 
	    <<"@(" << target_point.x << ":" << target_point.y 
	    << ") first_speed:"
	    << first_speed << std::endl;
  return true;
}

bool
Bhv_SelectionPass::get_selection_pass( rcsc::PlayerAgent * agent
				       ,rcsc::Vector2D * target_point
				       ,double * first_speed
				       ,int * receiver_unum
				       ,bool * continue_action
				       )
{
  ////////////////////////////////////////////////////////////////////
  //Body_Pass::get_best_passを流用
  static rcsc::GameTime S_last_calc_time( 0, 0 );
  static bool S_last_calc_valid = false;
  static rcsc::Vector2D S_last_calc_target_point;
  static double S_last_calc_speed = 0.0;
  static int S_last_calc_receiver_unum = rcsc::Unum_Unknown;
  static int S_last_calc_cycle = agent->world().time().cycle();
  
  if(continue_action)
    *continue_action = false;
  if ( S_last_calc_time == agent->world().time() 
       )
    {
      if ( S_last_calc_valid )
        {
	  if ( target_point )
            {
	      *target_point = S_last_calc_target_point;
            }
	  if ( first_speed )
            {
	      *first_speed = S_last_calc_speed;
            }
	  if ( receiver_unum )
            {
	      *receiver_unum = S_last_calc_receiver_unum;
            }
	  return true;
        }
      return false;
    }
  
  if ( agent->world().time().cycle() - S_last_calc_cycle < 3
       && agent->world().time().cycle() > 3
       )
    {
      if ( S_last_calc_valid )
        {
	  if ( target_point )
            {
	      *target_point = S_last_calc_target_point;
            }
	  if ( first_speed )
            {
	      *first_speed = S_last_calc_speed;
            }
	  if ( receiver_unum )
            {
	      *receiver_unum = S_last_calc_receiver_unum;
            }
	  std::cout << agent->world().self().unum() 
		    << ":[" << agent->world().time().cycle() 
		    << "] Maybe, the last action have not been executed yet, because last action cycle (" 
		    << S_last_calc_cycle
		    << ") is near current cycle."
		    << std::endl;
	   if(continue_action)
	     *continue_action = true;
	  return true;
        }
      return false;
    }

  S_last_calc_time = agent->world().time();
  S_last_calc_valid = false;
  S_last_calc_cycle = agent->world().time().cycle();

  ///////////////////////////////////////////////////////////////////
  //
  if(Learning_Driver(PGMethod_SelectionPass::getInstance()).execute( agent )){
    S_last_calc_target_point
      = PGMethod_SelectionPass::getInstance().getTargetPoint();
    S_last_calc_receiver_unum 
      = PGMethod_SelectionPass::getInstance().getTargetUnum();

    if(S_last_calc_target_point == rcsc::Vector2D::INVALIDATED
       || S_last_calc_receiver_unum == rcsc::Unum_Unknown)
      return false;

    double receiver_dist = agent->world().self().pos().dist(S_last_calc_target_point);
    double end_speed = 1.5;
    double first_speed_ = 100.0;
    do{
      first_speed_
	= rcsc::calc_first_term_geom_series_last
	( end_speed,
	  receiver_dist,
	  rcsc::ServerParam::i().ballDecay() );
      if ( first_speed_ < rcsc::ServerParam::i().ballSpeedMax() )
	{
	  break;
	}
      end_speed -= 0.1;
    }while ( end_speed > 0.8 );

    S_last_calc_speed 
      = std::min(first_speed_,rcsc::ServerParam::i().ballSpeedMax());
    S_last_calc_valid = true;
    rcsc::dlog.addText( rcsc::Logger::ACTION,
			"%s:%d: get_selection_pass() target=(%.1f %.1f)"
			" speed=%.3f  receiver=%d"
			,__FILE__, __LINE__,
			S_last_calc_target_point.x, S_last_calc_target_point.y,
			S_last_calc_speed,
			S_last_calc_receiver_unum );
    std::cout << agent->world().self().unum() 
	      << ":[" << agent->world().time().cycle()
	      << "] selected action:target:"<< S_last_calc_receiver_unum 
	      <<"@(" << S_last_calc_target_point.x << ":" << S_last_calc_target_point.y 
	      << ") first_speed:"
	      << S_last_calc_speed << std::endl;
  }

  if ( S_last_calc_valid )
    {
      if ( target_point )
        {
	  *target_point = S_last_calc_target_point;
        }
      if ( first_speed )
        {
	  *first_speed = S_last_calc_speed;
        }
      if ( receiver_unum )
        {
	  *receiver_unum = S_last_calc_receiver_unum;
        }
      
      rcsc::dlog.addText( rcsc::Logger::ACTION,
			  "%s:%d: best pass (%.2f, %.2f). speed=%.2f. receiver=%d"
			  ,__FILE__, __LINE__,
			  S_last_calc_target_point.x, S_last_calc_target_point.y,
			  S_last_calc_speed, S_last_calc_receiver_unum );
    }
  
  return S_last_calc_valid;
}

bool
Bhv_SelectionPass::prepare( rcsc::PlayerAgent * agent )
{
  return Learning_Driver(PGMethod_SelectionPass::getInstance()).prepare(agent);
}
