// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_set_play_goal_kick.h"

#include "bhv_set_play.h"
#include "bhv_prepare_set_play_kick.h"
#include "bhv_go_to_static_ball.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include <rcsc/player/player_agent.h>

#include <rcsc/common/logger.h>

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/common/server_param.h>
#include "bhv_fs_clear_ball.h"

#include "bhv_fs_kickin_move.h"

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_SetPlayGoalKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_SetPlayGoalKick"
                        ,__FILE__, __LINE__ );

    if ( isKicker( agent ) )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_SetPlayGoalKick::isKicker( const rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    if(wm.self().goalie()){
      return true;
    }
    if ( wm.setplayCount() < 30 )
      {
        return false;
      }

    if ( wm.self().distFromBall() > 20 )
      {
    	return false;
      }

    if ( wm.teammatesFromBall().empty() )
      {
        return true;
      }

    const rcsc::PlayerPtrCont::const_iterator 
      end = wm.teammatesFromBall().end();
    for ( rcsc::PlayerPtrCont::const_iterator 
	    it = wm.teammatesFromBall().begin();
          it != end; ++it )
      {
        if ( ! (*it)->goalie()
	       && (*it)->distFromBall() < wm.ball().distFromSelf()
	     )
	  {
	    // exist other kicker
	    return false;
	  }
      }
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Bhv_SetPlayGoalKick::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel &wm=agent->world();

  static int S_scan_count = -5;
  rcsc::Vector2D face_target( 0.0, 0.0 );
  
  // go to point to kick the ball
  if ( Bhv_GoToStaticBall( 0.0 ).execute( agent ) )
    {
      return;
    }

  // already kick point
  if ( ( wm.ball().angleFromSelf()
	 - wm.self().body() ).abs() > 3.0 )
    {
      rcsc::Body_TurnToBall().execute( agent );
      agent->setNeckAction( new rcsc::Neck_ScanField() );
      return;
    }
  
  // wait1
  if ( S_scan_count < 0 )
    {
      S_scan_count++;
      rcsc::Body_TurnToPoint( face_target ).execute( agent );
      agent->setNeckAction( new rcsc::Neck_ScanField() );
      return;
    }
  
    // wait2
  if ( S_scan_count < 30
       && wm.teammatesFromSelf().empty() )
    {
      S_scan_count++;
      rcsc::Body_TurnToPoint( face_target ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        return;
    }
  
  S_scan_count = -5;

  //キック先の選定
  rcsc::Vector2D farPos=wm.self().pos();
  const rcsc::PlayerPtrCont::const_iterator team_end = wm.teammatesFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator team_it = wm.teammatesFromSelf().begin();
	team_it != team_end;
	++team_it)
  {
	rcsc::Segment2D passCource(wm.ball().pos(),(*team_it)->pos());

	const rcsc::PlayerPtrCont::const_iterator opp_end = wm.opponentsFromSelf().end();
	for(rcsc::PlayerPtrCont::const_iterator
	   	 opp_it = wm.opponentsFromSelf().begin();
		 opp_it != opp_end;
		 ++opp_it)
	{
		if((*team_it)->distFromSelf()<30 
		   &&(*team_it)->distFromSelf()>10
		   &&passCource.dist((*opp_it)->pos()) + wm.self().pos().dist((*opp_it)->pos())*0.1 > 5
                   &&wm.countOpponentsIn(rcsc::Circle2D((*team_it)->pos(),10),5,false)==0
                   &&!(*team_it)->goalie()
  		   &&wm.self().pos().dist((*team_it)->pos()) > wm.self().pos().dist(farPos))
		{
			farPos=(*team_it)->pos();
	        }
	}
  }

  if(wm.self().pos().dist(farPos)>5
     &&rcsc::Body_KickOneStep(farPos
		,rcsc::ServerParam::i().ballSpeedMax()).execute(agent))
  {
	agent->setNeckAction( new rcsc::Neck_ScanField() );
	return;
  }

  //2回目
  for(rcsc::PlayerPtrCont::const_iterator team_it = wm.teammatesFromSelf().begin();
	team_it != team_end;
	++team_it)
  {
	rcsc::Segment2D passCource(wm.ball().pos(),(*team_it)->pos());

	const rcsc::PlayerPtrCont::const_iterator opp_end = wm.opponentsFromSelf().end();
	for(rcsc::PlayerPtrCont::const_iterator
	   	 opp_it = wm.opponentsFromSelf().begin();
		 opp_it != opp_end;
		 ++opp_it)
	{
		if((*team_it)->distFromSelf()<30 
		   &&(*team_it)->distFromSelf()>5
		   &&passCource.dist((*opp_it)->pos()) + wm.self().pos().dist((*opp_it)->pos())*0.1 > 5
                   &&wm.countOpponentsIn(rcsc::Circle2D((*team_it)->pos(),3),5,false)==0
                   &&!(*team_it)->goalie()
  		   &&wm.self().pos().dist((*team_it)->pos()) > wm.self().pos().dist(farPos))
		{
			farPos=(*team_it)->pos();
	        }
	}
  }

  if(wm.self().pos().dist(farPos)>5
     &&rcsc::Body_KickOneStep(farPos
		,rcsc::ServerParam::i().ballSpeedMax()).execute(agent))
  {
	agent->setNeckAction( new rcsc::Neck_ScanField() );
	return;
  }

    // clear
  Bhv_FSClearBall().execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Bhv_SetPlayGoalKick::doMove( rcsc::PlayerAgent * agent )
{
  if(Bhv_FSKickinMove(M_home_pos).execute(agent))
    return;
    double dash_power
        = Bhv_SetPlay::get_set_play_dash_power( agent );
    double dist_thr = agent->world().ball().distFromSelf() * 0.07;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    rcsc::Vector2D move_point = M_home_pos;
    move_point.y += agent->world().ball().pos().y * 0.5;

    int i=0;
    while(agent->world().existOpponentIn(rcsc::Circle2D(move_point,10)
					 ,5,false)
	  && move_point.dist(M_home_pos)<10){
      if(i%2==0)
	move_point.x-=0.5;
      else
	move_point.y*=0.8;

      i++;
    }
    if ( ! rcsc::Body_GoToPoint( move_point,
                                 dist_thr,
                                 dash_power
                                 ).execute( agent ) )
    {
        // already there
        rcsc::Body_TurnToBall().execute( agent );
    }
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
}
