// -*-c++-*-

/******************************************************
 * @COPYRIGHT:Jun Masaki @ iesl,ies,sit 2009 
 ******************************************************/

////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "pg_method_selection_pass.h"

#include <boost/lexical_cast.hpp>
#include <cmath>
#include <stdexcept>
#include <rcsc/player/intercept_table.h>
#include "learning_driver.h"
#include <iostream>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>

//0:No debug log, 1:few, 2:middle, 3:max
#define DEBUG_LEARN_SP 0

/*-------------------------------------------------------------------*/

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::processWhenInitialized(void)
{
  M_initialized_flag = true;
  std::vector<Learning_Item>::const_iterator 
    it = M_settingItems.begin();
  std::vector<Learning_Item>::const_iterator 
    end = M_settingItems.end();
  int checker = 0;
  for(;it != end;++it){
    if(it->hasSettedParam()){
      if(it->hasSettedNoBool()){
	if(it->getName() == "epsilon"
	   && (checker & 8) != 8){
	  M_epsilon = it->getParam(0.0);
	  checker+=8;
	}else if(it->getName() == "temperature"
		 && (checker & 16) != 16){
	  M_temperature = it->getParam(0.0);
	  checker+=16;
	}
      }else{
	if(it->getName() == "Using" 
	   && (checker & 1) != 1){
	  M_executable_flag = it->getParam(true);
	  checker+=1;
	}else if(it->getName() == "Learning"
		 && (checker & 2) != 2){
	  M_learning_flag = it->getParam(true);
	  checker+=2;
	}else if(it->getName() == "Saving"
		 && (checker & 4) != 4){
	  M_saving_flag = it->getParam(true);
	  checker+=4;
	}
      } 
    }
  }

  std::cout.setf(std::ios::boolalpha);
  std::cout << M_unum << ":PGMethod_SelectionPass initialized." << std::endl
	    << "     -Useing:" << M_executable_flag << std::endl
	    << "   -Learning:" << M_learning_flag << std::endl
	    << "     -Saving:" << M_saving_flag << std::endl
	    << "    -epsilon:" << M_epsilon << std::endl
	    << "-temperature:" << M_temperature << std::endl;

  if(checker != 31)
    return false;

  if(M_executable_flag){
    try{
      M_learningDataFileName = "csv/learning_sp_" 
	+ boost::lexical_cast<std::string>(M_unum) + ".csv";
    }catch(boost::bad_lexical_cast & ex){
      std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
      return false;
    }
    M_dataItems.push_back(Learning_Item("date",true)); 
    M_dataItems.push_back(Learning_Item("end_time",true)); 
    M_dataItems.push_back(Learning_Item("num_all",false)); 
    M_dataItems.push_back(Learning_Item("num_one",true));
    M_dataItems.push_back(Learning_Item("start_cycle",true));
    M_dataItems.push_back(Learning_Item("end_cycle",true));
    M_dataItems.push_back(Learning_Item("actions",true));
    Learning_Item temperature_data("temerature",true);
    temperature_data.setParam(M_temperature);
    M_dataItems.push_back(temperature_data);
    Learning_Item epsilon_data("epsilon",true);
    epsilon_data.setParam(M_epsilon);
    M_dataItems.push_back(epsilon_data);
    M_dataItems.push_back(Learning_Item("reward",true));
    
    heuristics_cont::const_iterator heuristics_it = M_heuristics_ptrs.begin();
    heuristics_cont::const_iterator heuristics_end = M_heuristics_ptrs.end();
    for(;heuristics_it != heuristics_end;++heuristics_it){
      M_dataItems.push_back(Learning_Item(heuristics_it->getName(),false));
    }
  }
  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::processWhenReadData(void)
{
  int checker=0;

  std::vector<Learning_Item>::iterator item_it = M_dataItems.begin();
  std::vector<Learning_Item>::const_iterator item_end = M_dataItems.end();
  heuristics_cont::const_iterator heuristics_it = M_heuristics_ptrs.begin();
  heuristics_cont::const_iterator heuristics_end = M_heuristics_ptrs.end();
  for(;item_it!=item_end;++item_it){
    if(item_it->hasSettedParam()){
      if(item_it->hasSettedNoBool()){
#if DEBUG_LEARN_SP >= 2
	std::cout << M_unum 
		  << ":" << item_it->getName()
		  << " was setted as " << item_it->getParam(0.0)
		  << std::endl;

#endif
	if(item_it->getName() == "num_all"
	   && (checker & 1) != 1){
	  M_all_learning_count = (int)item_it->getParam(0.0);
	  checker+=1;
	}
	int offset;
	for(heuristics_it=M_heuristics_ptrs.begin() , offset=2;
	    heuristics_it!=heuristics_end;
	    ++heuristics_it , offset*=2){
	  if(item_it->getName() == heuristics_it->getName() 
	     && (checker & offset) != offset){
	    M_weights.push_back(item_it->getParam(0.0));
	    checker+=offset;
	  }
	}
      }
    }else{
#if DEBUG_LEARN_SP >= 1
      std::cout << M_unum 
		<< ":" << item_it->getName()
		<< ":This param was not setted now or not required."
		<< " As required, it'll be given default value."
		<< std::endl;

#endif
      if(item_it->getName() == "num_all"
	 && (checker & 1) != 1){
	M_all_learning_count = 0;
	checker+=1;
      }
      int offset;
      for(heuristics_it=M_heuristics_ptrs.begin() , offset=2;
	  heuristics_it!=heuristics_end;
	  ++heuristics_it , offset*=2){
	if(item_it->getName() == heuristics_it->getName() 
	   && (checker & offset) != offset){
	  M_weights.push_back(0.0);
	  checker+=offset;
	}
      }
    }
  }
  if(checker!=std::pow(2.0,(int)M_heuristics_ptrs.size() + 1) - 1)
    return false;

  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::makeAction( const rcsc::PlayerAgent * agent )
{
  M_settedAction_num = -1;
  if(!M_actions.empty())
    M_actions.clear();

    rcsc::PlayerPtrCont::const_iterator 
      player_it = agent->world().teammatesFromSelf().begin();
    const rcsc::PlayerPtrCont::const_iterator
      player_end = agent->world().teammatesFromSelf().end();
    for(;player_it!=player_end;++player_it){
      if ( (*player_it)->goalie() && (*player_it)->pos().x < -22.0 )
        {
	  // goalie is rejected.
	  continue;
        }
      
      if ( (*player_it)->posCount() > 12 ) // 3->12
        {
	  // low confidence players are rejected.
	  continue;
        }

      M_actions.push_back(action_cont::value_type(Action_inner((*player_it)->unum()
							       ,(*player_it)->pos())));
#if DEBUG_LEARN_SP >=2
      std::cout << M_unum
		<< ":[" << agent->world().time().cycle()
		<< "] make an action:" << (*player_it)->unum() << "@(" 
		<< (*player_it)->pos().x << ":" << (*player_it)->pos().y 
		<< ")" << std::endl;
#endif
    }
  
  if(M_actions.empty())
    return false;
  
  M_Us.clear();
  M_Us.resize(M_actions.size()
	      ,std::vector<double>(M_heuristics_ptrs.size()));

  return true;
}

///////////////////////////////////////////////////////////////////////
//
double 
PGMethod_SelectionPass::getWeight( const int heuristics ) const
{
  if(heuristics < 0 || (int)M_weights.size() <= heuristics)
    return 0.0;
  try{
    return M_weights.at(heuristics);
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
  return 0.0;
}

///////////////////////////////////////////////////////////////////////
//
double 
PGMethod_SelectionPass::U( const int heuristics 
			   , const int action 
			   , const rcsc::PlayerAgent * agent )
{
  if(heuristics < 0 || (int)M_heuristics_ptrs.size() <= heuristics)
    return 0.0;
  if(action < 0 || (int)M_actions.size() <= action)
    return 0.0;

  double value;
  try{ 
    value = (this->*M_heuristics_ptrs.at(heuristics).getPointer())(agent
								   ,action);
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }

  value = std::max(0.0,std::min(value,10.0));
  try{
    if(M_Us.size()!=M_actions.size() 
       || M_Us.at(action).size()!=M_heuristics_ptrs.size()){
      M_Us.resize(M_actions.size()
		  ,std::vector<double>(M_heuristics_ptrs.size()));
    }
    M_Us.at(action).at(heuristics)=value;
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
  return value;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::setAction( const int action )
{
  if(action<0 || (int)M_actions.size() <= action){
    M_settedAction_num = -1;
    return false;
  }
#if DEBUG_LEARN_SP >=2
  std::cout << M_unum
	    << ":set an action:" << action
	    << std::endl;
#endif
  M_settedAction_num = action;
  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::isEpisodeEnd( const rcsc::PlayerAgent * agent )
{
  static int last_ourBall_cycle;

  if(M_beginningAction_cycle == -1)
    return false;
  int self_min = agent->world().interceptTable()->selfReachCycle();
  int mate_min = agent->world().interceptTable()->teammateReachCycle();
  int opp_min = agent->world().interceptTable()->opponentReachCycle();
  if(agent->world().existKickableTeammate()
     || mate_min < opp_min+4
     || self_min < opp_min+4){
    last_ourBall_cycle = agent->world().time().cycle();
  }

  if(agent->world().time().cycle() - last_ourBall_cycle > 10){
    if(M_beginningAction_cycle > last_ourBall_cycle){
#if DEBUG_LEARN_SP >= 2
      std::cout << M_unum << ":[" << agent->world().time().cycle() 
		<< "] Maybe, the last action was not executed, because beginning episode cycle (" 
		<< M_beginningAction_cycle 
		<< ") is later than last our ball cycle ("
		<< last_ourBall_cycle 
		<< ")." 
		<< std::endl;
#endif
      M_beginningAction_cycle = -1;
      return false;
    }

    std::cout << M_unum << ":[" << agent->world().time().cycle() 
	      << "] the episode is end." 
	      << " last our ball cycle:" << last_ourBall_cycle
	      << std::endl;
    return true;
  }
    
#if DEBUG_LEARN_SP >= 3
      std::cout << M_unum << ":[" << agent->world().time().cycle() 
		<< "] continue the episode: biginning episode cycle:" 
		<< M_beginningAction_cycle 
		<< " last our ball cycle:"
		<< last_ourBall_cycle  
		<< std::endl;
#endif

  return false;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::updateRewards( const rcsc::PlayerAgent * agent )
{
  if(M_beginningAction_cycle == -1)
    return false;

  int episode_len = agent->world().time().cycle() - M_beginningAction_cycle;
  
#if DEBUG_LEARN_SP >= 2
  std::cout << M_unum << ":[" << agent->world().time().cycle() 
	    << "] episode length:" 
	    << episode_len 
	    << std::endl;
#endif
  M_reward = episode_len;


  /*
  if(episode_len <= 20)
    M_reward = 0.0;
  else
    M_reward = 1.0 - 20.0/(double)episode_len;
  */
  /*
  if(episode_len <= 0)
    M_reward = 1.0;
  else
    M_reward = 1.0/(double)episode_len;
  */
  std::vector<Learning_Item>::iterator item_it = M_dataItems.begin();
  std::vector<Learning_Item>::const_iterator item_end = M_dataItems.end();
  for(;item_it!=item_end;++item_it){
    if(item_it->getName() == "reward"){
      item_it->setParam(M_reward);
      break;
    }
  }

  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::updateWeights( const rcsc::PlayerAgent * agent )
{ 
  if(M_beginningAction_cycle == -1)
    return false;
  double eweight_sum = 0.0;

  std::vector<Learning_Item>::iterator item_it;
  std::vector<Learning_Item>::const_iterator item_end = M_dataItems.end();

  std::vector< std::vector<double> >::const_iterator e_it;
  std::vector< std::vector<double> >::const_iterator
    e_end = M_formula_eweights.end();
  for(unsigned int i=0; i < M_heuristics_ptrs.size(); i++){
    eweight_sum=0.0;
    for(e_it = M_formula_eweights.begin();
	e_it!=e_end;++e_it){

      try{
	eweight_sum += e_it->at(i);
#if DEBUG_LEARN_SP >= 2
	std::cout << M_unum << ":[" << agent->world().time().cycle() 
		  << "] heuristics:" << i
		  << " :e of each action:"
		  << e_it->at(i)
		  << std::endl;
#endif
      }catch(std::out_of_range & ex){
	std::cerr << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	return false;
      }

    }

    try{
      double weight = M_weights.at(i) + M_epsilon * M_reward * eweight_sum;
#if DEBUG_LEARN_SP >=2
      std::cout << M_unum << ":[" << agent->world().time().cycle() << "]"
		<< " update weight:"
		<< " heuristics:" << i
		<< " weight after" << weight
		<< " weight before:" << M_weights.at(i)
		<< " epsilon:" << M_epsilon
		<< " reward:" << M_reward
		<< " e:" << eweight_sum
		<< std::endl;
#endif
      if(!isinf(weight) && !isnan(weight)){
	M_weights.at(i) = weight;
      }
#if DEBUG_LEARN_SP >=2
      else{
	std::cout << M_unum << ":[" << agent->world().time().cycle() << "]"
		  << " update weight:"
		  << " nan or inf was detected:"
		  << " heuristics:" << i
		  << " weight before:" << M_weights.at(i)
		  << " epsilon:" << M_epsilon
		  << " reward:" << M_reward
		  << " e:" << eweight_sum
		  << std::endl;
      }
#endif
      
      if(M_weights.at(i) < 0.0) 
	M_weights.at(i) = 0.0;

      for(item_it=M_dataItems.begin();
	  item_it!=item_end;++item_it){
	
	if(item_it->getName() == M_heuristics_ptrs.at(i).getName()){  
	  item_it->setParam(M_weights.at(i));	  
	  break;
	}
      }

    }catch(std::out_of_range & ex){
      std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
      return false;
    }
  }
  
  M_formula_eweights.clear();
  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::processWhenSavedLearningData(void)
{
  return true;
}

///////////////////////////////////////////////////////////////////////
//
bool
PGMethod_SelectionPass::endTheEpisode( const rcsc::PlayerAgent * agent )
{
  if(M_beginningAction_cycle == -1)
    return false;

  struct tm *date;
  time_t now;
  int year, month, day;
  int hour, minute, second;
  
  /* 現在の日時を取得 */
  time(&now);
  date = localtime(&now);
  
  year = date->tm_year + 1900;
  month = date->tm_mon + 1;
  day = date->tm_mday;
  hour = date->tm_hour;
  minute = date->tm_min;
  second = date->tm_sec;
  
    M_all_learning_count++;
    M_game_learning_count++;
    std::vector<Learning_Item>::iterator item_it = M_dataItems.begin();
    std::vector<Learning_Item>::const_iterator item_end = M_dataItems.end();
    int checker = 0;
    for(;item_it!=item_end;++item_it){
      try{
	if(item_it->getName() == "num_all" && (checker&1)!=1){
	  item_it->setParam(boost::lexical_cast<std::string>(M_all_learning_count));
	  checker+=1;
	}else if(item_it->getName() == "num_one" && (checker&2)!=2){
	  item_it->setParam(boost::lexical_cast<std::string>(M_game_learning_count));
	  checker+=2;
	}else if(item_it->getName() == "end_cycle" && (checker&4)!=4){
	  item_it->setParam(boost::lexical_cast<std::string>(agent->world().time().cycle()));
	  checker+=4;
	}else if(item_it->getName() == "date" && (checker&8)!=8){
	  std::string date 
	    = boost::lexical_cast<std::string>(year)
	    + "/"
	    + boost::lexical_cast<std::string>(month)
	    + "/"
	    + boost::lexical_cast<std::string>(day);
	  item_it->setParam(date);
	  checker+=8;
	}else if(item_it->getName() == "end_time" && (checker&16)!=16){
	  std::string time_now 
	    = boost::lexical_cast<std::string>(hour)
	    + ":"
	    + boost::lexical_cast<std::string>(minute)
	    + ":"
	    + boost::lexical_cast<std::string>(second);
	  item_it->setParam(time_now);
	  checker+=16;
	}
      }catch(boost::bad_lexical_cast & ex){
	std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	return false;
      }
    }

    if(checker == 31){
      std::cout << M_unum << ":[" << agent->world().time().cycle() 
		<< "] end the episode." 
		<< std::endl;
      M_beginningAction_cycle = -1;
      return true;
    }

    return false;
}

///////////////////////////////////////////////////////////////////////
//
bool 
PGMethod_SelectionPass::processWhenExecuted( const rcsc::PlayerAgent * agent )
{
  bool first_action_valid = false;
  std::vector<Learning_Item>::iterator item_it;
  std::vector<Learning_Item>::const_iterator item_end = M_dataItems.end();
  if(M_beginningAction_cycle == -1){
    M_beginningAction_cycle = agent->world().time().cycle();
    int checker = 0;
    for(item_it=M_dataItems.begin();
	item_it!=item_end;++item_it){
      try{
	if(item_it->getName()=="actions"){
	  std::string first_action 
	    = boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).unum())
	    + "@("
	    + boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).pos().x)
	    + ":"
	    + boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).pos().y)
	    + ") ";
	  item_it->setParam(first_action);
	  checker+=2;
	}else if(item_it->getName()=="start_cycle"){
	  item_it->setParam(boost::lexical_cast<std::string>(agent->world().time().cycle()));
	checker+=3;
	}
      }catch(std::out_of_range & ex){
	std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	return false;
      }catch(boost::bad_lexical_cast & ex){
	std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	return false;
      }
    }

    if(checker==5){
      std::cout << M_unum 
		<< ":[" << M_beginningAction_cycle 
		<< "] start episode." 
		<< std::endl;
      first_action_valid = true;
      M_formula_eweights.clear();
    }else{
      M_beginningAction_cycle = -1;
      return false;
    }
  }
  if(!first_action_valid){
    for(item_it = M_dataItems.begin(); item_it != item_end; ++item_it){
      if(item_it->getName()=="actions"){
	try{
	  std::string action_series = item_it->getParam(std::string("string"));
	  action_series 
	    += boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).unum())
	    + "@("
	    + boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).pos().x)
	    + ":"
	    + boost::lexical_cast<std::string>(M_actions.at(M_settedAction_num).pos().y)
	    + ") ";
	  item_it->setParam(action_series);
	}catch(std::out_of_range & ex){
	  std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	  return false;
	}catch(boost::bad_lexical_cast & ex){
	  std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	  return false;
	}
#if DEBUG_LEARN_SP >= 1
	std::cout << M_unum << ":[" << agent->world().time().cycle() 
		  << "] new action was executed." 
		  << " beginning episode cycle:" << M_beginningAction_cycle
		  << std::endl;
#endif
	break;
      }
    }
  }
  
  
  std::vector<double> formula_eweight;
  double expectation;
  double e_heuristics;
  for(unsigned int i=0;i < M_heuristics_ptrs.size();i++){
    expectation = 0.0;
    try{
      for(unsigned int a=0;a < M_actions.size();a++){
	expectation += M_Us.at(a).at(i) * M_actionProbabilities.at(a);
      }
      e_heuristics 
	= (M_Us.at(M_settedAction_num).at(i) - expectation)
	/ M_temperature;
      if(isinf(e_heuristics) && isnan(e_heuristics)){
	e_heuristics = 0.0;
#if DEBUG_LEARN_SP >= 1
      std::cout << M_unum << ":[" << agent->world().time().cycle() 
		<< "] calc e:"
		<< " heuristics:" << i
		<< " nan or inf was detected."
		<< std::endl;
#endif
      }
      formula_eweight.push_back(e_heuristics);
#if DEBUG_LEARN_SP >= 1
      std::cout << M_unum << ":[" << agent->world().time().cycle() 
		<< "] calc e:"
		<< " heuristics:" << i
		<< " e:" << e_heuristics
		<< " selected action's U:" << M_Us.at(M_settedAction_num).at(i)
		<< " expectation:" << expectation
		<< " temperature:" << M_temperature
		<< std::endl;
#endif
    }catch(std::out_of_range & ex){
      std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
      return false;
    }
  }
  M_formula_eweights.push_back(formula_eweight);
  M_Us.clear();
  M_actionProbabilities.clear();
  return true;
}

///////////////////////////////////////////////////////////////////////
//
const rcsc::Vector2D &
PGMethod_SelectionPass::getTargetPoint(void) const
{
  if(M_settedAction_num < 0 || (int)M_actions.size() <= M_settedAction_num)
    return rcsc::Vector2D::INVALIDATED;

  try{
    return M_actions.at(M_settedAction_num).pos();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return rcsc::Vector2D::INVALIDATED;
  }
  return rcsc::Vector2D::INVALIDATED;
}

///////////////////////////////////////////////////////////////////////
//
int 
PGMethod_SelectionPass::getTargetUnum(void) const
{
  if(M_settedAction_num < 0 || (int)M_actions.size() <= M_settedAction_num)
    return rcsc::Unum_Unknown;

  try{
    return M_actions.at(M_settedAction_num).unum();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return rcsc::Unum_Unknown;
  }
  return rcsc::Unum_Unknown;
}

///////////////////////////////////////////////////////////////////////
// パスコースに敵はいるか
double 
PGMethod_SelectionPass::existEnemiesOnPassCourse( const rcsc::PlayerAgent * agent 
						  , const int action )
{ 
  rcsc::Vector2D target_pos;
  try{
    if(!(0 <= action && action < (int)M_actions.size())
       && M_actions.at(action).unum() == agent->world().self().unum())
      return 0.0;

    target_pos = M_actions.at(action).pos();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
  const double target_dist 
    = agent->world().self().pos().dist( target_pos );
  const double target_angle 
    = ( target_pos - agent->world().self().pos() ).dir().degree();
  
  const int count_30Degree =
    agent->world().countOpponentsIn(rcsc::Sector2D(agent->world().self().pos()
						  ,0.0,target_dist
						  ,target_angle - 30
						  ,target_angle + 30)
				   ,5,true);
  if(count_30Degree == 0)
    return 10.0;

  const int count_20Degree =
    agent->world().countOpponentsIn(rcsc::Sector2D(agent->world().self().pos()
						  ,0.0,target_dist
						  ,target_angle - 20
						  ,target_angle + 20)
				   ,5,true);
  if(count_20Degree == 0)
    return 10.0 - count_30Degree * 1;

  const int count_10Degree =
    agent->world().countOpponentsIn(rcsc::Sector2D(agent->world().self().pos()
						   ,0.0,target_dist
						   ,target_angle - 10
						   ,target_angle + 10)
				    ,5,true);
  
  return 10.0 
    - (count_30Degree - count_20Degree) * 1
    - (count_20Degree - count_10Degree) * 3
    - count_10Degree * 5;
}

///////////////////////////////////////////////////////////////////////
// レシーバと最近接の敵との距離から
double 
PGMethod_SelectionPass::distFromNearestEnemyToPassPoint( const rcsc::PlayerAgent * agent 
							 , const int action )
{
  rcsc::Vector2D target_pos;
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  try{
    target_pos = M_actions.at(action).pos();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }

  double min_dist = 1000.0;

  const rcsc::PlayerPtrCont::const_iterator 
    opp_end = agent->world().opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator 
	opp_it = agent->world().opponentsFromSelf().begin();
      opp_it != opp_end;
      ++opp_it )
    {
      if(*opp_it == NULL) 
	continue;		//いなければ飛ばす
      
      const double dist = target_pos.dist( (*opp_it)->pos() );
      if(min_dist > dist)
	{
	  min_dist = dist;
	}
    }
  if(min_dist > 200.0) 
    return 0.0;

  if(min_dist < 25)
    return min_dist/2.5;
  else
    return 10.0;
  
  return 0.0;
}

///////////////////////////////////////////////////////////////////////
//パス先に敵が何人いるか
double
PGMethod_SelectionPass::countEnemiesNearPassPoint( const rcsc::PlayerAgent * agent 
						  , const int action )
{
  int theRestEnemies = 10;
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  
  rcsc::Vector2D target_pos;
  try{
    target_pos = M_actions.at(action).pos();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }  

  const rcsc::PlayerPtrCont::const_iterator
    opp_end = agent->world().opponentsFromSelf().end();
  for(rcsc::PlayerPtrCont::const_iterator 
	opp_it = agent->world().opponentsFromSelf().begin();
      opp_it != opp_end;
      ++opp_it)
    {
      if(*opp_it == NULL) 
	continue;		//いなければ飛ばす
 
      if(target_pos.dist((*opp_it)->pos()) < 10.0)
	{
	  theRestEnemies--;
	}
    }
  
  if(theRestEnemies<0)
    return 0.0;

  return (double)theRestEnemies;
}

///////////////////////////////////////////////////////////////////////
//
double 
PGMethod_SelectionPass::distBetweenGoalAndPassPoint( const rcsc::PlayerAgent * agent 
						     , const int action )
{
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  
  const rcsc::Vector2D goal_pos(52.5, 0.0);
  rcsc::Vector2D target_pos;
  try{
    target_pos = M_actions.at(action).pos();
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }  

  double goal_dist = (goal_pos - target_pos).r();

  if(goal_dist > 50.0)
    return 0.0;
  return 10 - goal_dist/5.0;
}

///////////////////////////////////////////////////////////////////////
//パス相手の位置の信頼性によって判断
double 
PGMethod_SelectionPass::recieverConf( const rcsc::PlayerAgent * agent 
				      , const int action )
{
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  try{
    if(M_actions.at(action).unum() == agent->world().self().unum())
      return 9.0;
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }

  double player_conf = 0.0;
  try{
    if(agent->world().teammate(M_actions.at(action).unum()) != NULL)
      player_conf = 
	1 - agent->world().teammate(M_actions.at(action).unum())->posCount()/10;
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
#if DEBUG_LEARN_SP >= 3 
  std::cout << M_unum << ":["
	    << agent->world().time().cycle() 
	    << "] recieverConf: action:" << action
	    << " player posCount:" << (1 - player_conf) * 10
	    <<std::endl;    
#endif  
  if(player_conf < 0.0)
    player_conf = 0.0;

  return 10.0 * player_conf;
}

///////////////////////////////////////////////////////////////////////
//パス先のポジションによって判断
double 
PGMethod_SelectionPass::recieverRole( const rcsc::PlayerAgent * agent 
				      , const int action )
{
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  try{
    //自分自身へのパス(ドリブル)の時は10を返す
    if(M_actions.at(action).unum() == agent->world().self().unum())
      return 10.0;
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
  try{
    switch( M_actions.at(action).unum() ){
    case 1:
      return 0.0;
    case 2:
    case 3:
      return 1.0;
    case 4:
    case 5:
      //DFにパス
      return 3.0;
    case 6:
      return 5.0;
    case 7:
    case 8:
      //MFにパス
      return 7.0;
    case 9:
    case 10:
      return 9.0;
    case 11:
      //FWにパス
      return 10.0;
    default:
      return 0.0;
    }
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }

  return 0.0;
}

///////////////////////////////////////////////////////////////////////
//
double 
PGMethod_SelectionPass::distFromSelfToPassPoint( const rcsc::PlayerAgent * agent 
						 , const int action )
{
  if(!(0 <= action && action < (int)M_actions.size()))
    return 0.0;
  try{
    if(M_actions.at(action).unum() == agent->world().self().unum())
      return 10.0;
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }

  double target_dist; 
  try{
    target_dist 
      = agent->world().self().pos().dist( M_actions.at(action).pos() );
  }catch(std::out_of_range & ex){
    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
    return 0.0;
  }
  if(target_dist > 50.0)
    return 0.0;
  return (50 - target_dist)/5.0;
}
