// -*-c++-*-

/******************************************************
 * @COPYRIGHT:Jun Masaki @ iesl,ies,sit 2009 
 ******************************************************/

////////////////////////////////////////////////////////

#ifndef PG_METHOD_SELECTION_PASS_H
#define PG_METHOD_SELECTION_PASS_H

#include "abstract_pg_method.h"
#include "singleton_interface.h"

#include <rcsc/player/player_agent.h>

class Learning_Driver;

//This class is defined as singleton class
class PGMethod_SelectionPass 
  : public Abstract_PGMethod , protected Singleton_Interface {
  friend class Learning_Driver;

private:

  double M_epsilon;

  double M_temperature;

  double M_reward;

  int M_beginningAction_cycle;

  int M_all_learning_count;
  
  int M_game_learning_count;
  

  class Action_inner { 
  private:
    int M_unum;
    rcsc::Vector2D M_pos;
  public:
    Action_inner( const int unum , const rcsc::Vector2D & pos ):M_pos(pos){
      M_unum = unum;
    }
    const int unum() const{
      return M_unum;
    }
    const rcsc::Vector2D & pos() const{
      return M_pos;
    }
  };

  typedef std::vector<Action_inner> action_cont;

  action_cont M_actions;

  int M_settedAction_num;

  std::vector<double> M_weights;

  typedef double (PGMethod_SelectionPass::*heuristics_funcPtr)( const rcsc::PlayerAgent * 
								, const int );

  class Heuristics_inner {
  private:
    std::string M_name;
    heuristics_funcPtr M_funcPtr;
  public:
    Heuristics_inner( const std::string & name 
		      ,heuristics_funcPtr funcPtr):M_name(name){
      M_funcPtr=funcPtr;
    }
    const std::string & getName(void) const{
      return M_name;
    }
    heuristics_funcPtr getPointer(void) const{
      return M_funcPtr;
    }
  };
  
  typedef std::vector<Heuristics_inner> heuristics_cont;
  
  heuristics_cont M_heuristics_ptrs;

  std::vector< std::vector<double> > M_Us;

  int M_last_action_cycle;

  //////////////////////////////////////////////////////////

private:

  PGMethod_SelectionPass(void)
  {
    M_settingFileName = "learning.conf";
    M_settingItems.push_back(Learning_Item("Using",true));
    M_settingItems.push_back(Learning_Item("Learning",true));
    M_settingItems.push_back(Learning_Item("Saving",true));
    M_settingItems.push_back(Learning_Item("epsilon",true));
    M_settingItems.push_back(Learning_Item("temperature",true));

    M_heuristics_ptrs.push_back(Heuristics_inner("existEnemiesOnPassCourse"
						 ,&PGMethod_SelectionPass::existEnemiesOnPassCourse));
    M_heuristics_ptrs.push_back(Heuristics_inner("distFromNearestEnemyToPassPoint"
						 ,&PGMethod_SelectionPass::distFromNearestEnemyToPassPoint));
    M_heuristics_ptrs.push_back(Heuristics_inner("countEnemiesNearPassPoint"
						 ,&PGMethod_SelectionPass::countEnemiesNearPassPoint));
    M_heuristics_ptrs.push_back(Heuristics_inner("distBetweenGoalAndPassPoint"
						 ,&PGMethod_SelectionPass::distBetweenGoalAndPassPoint));
    M_heuristics_ptrs.push_back(Heuristics_inner("recieverConf"
						 ,&PGMethod_SelectionPass::recieverConf));
    M_heuristics_ptrs.push_back(Heuristics_inner("recieverRole"
						 ,&PGMethod_SelectionPass::recieverRole));
    M_heuristics_ptrs.push_back(Heuristics_inner("distFromSelfToPassPoint"
						 ,&PGMethod_SelectionPass::distFromSelfToPassPoint));
    
    M_initialized_flag = false;
    M_executable_flag = false;
    M_learning_flag = false;
    M_saving_flag = false;
    M_unum = 0;
    M_settedAction_num = 0;
    M_all_learning_count = 0;
    M_game_learning_count = 0;
    M_beginningAction_cycle = -1;
  }

  PGMethod_SelectionPass (const PGMethod_SelectionPass & copyConstructer ){}

  PGMethod_SelectionPass & operator=( const PGMethod_SelectionPass & copyOperator ){}

  bool processWhenInitialized(void);

  bool processWhenReadData(void);

  bool makeAction( const rcsc::PlayerAgent * agent );

  int getActionSize(void) const
  {
    return M_actions.size();
  }

  double getTemperature(void) const
  {
    return M_temperature;
  }

  int getHeuristicsSize(void) const
  {
    return M_heuristics_ptrs.size();
  }

  double getWeight( const int heuristics ) const;

  double U( const int heuristics 
	    , const int action 
	    , const rcsc::PlayerAgent * agent );

  bool setAction( const int action );

  bool isEpisodeEnd( const rcsc::PlayerAgent * agent );

  bool updateRewards( const rcsc::PlayerAgent * agent );

  bool updateWeights( const rcsc::PlayerAgent * agent );

  bool endTheEpisode( const rcsc::PlayerAgent * agent );

  bool processWhenSavedLearningData(void);

  double existEnemiesOnPassCourse( const rcsc::PlayerAgent * agent 
				   , const int action );

  double distFromNearestEnemyToPassPoint( const rcsc::PlayerAgent * agent 
					  , const int action );

  double countEnemiesNearPassPoint( const rcsc::PlayerAgent * agent 
				    , const int action );

  double distBetweenGoalAndPassPoint( const rcsc::PlayerAgent * agent 
				      , const int action );

  double recieverConf( const rcsc::PlayerAgent * agent 
		       , const  int action );

  double recieverRole( const rcsc::PlayerAgent * agent 
		       , const int action );

  double distFromSelfToPassPoint( const rcsc::PlayerAgent * agent 
				  , const int action );

public:

  static  PGMethod_SelectionPass & getInstance(void){
    static PGMethod_SelectionPass singleton;
    return singleton;
  }

  bool processWhenExecuted( const rcsc::PlayerAgent * agent );

  const rcsc::Vector2D & getTargetPoint(void) const;

  int getTargetUnum(void) const;

};

#endif
