// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_center_back.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include "bhv_basic_tackle.h"
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include "bhv_fs_clear_ball.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::execute( rcsc::PlayerAgent * agent )
{
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }

  if ( kickable )
    {
      doKick( agent );
    }
  else
    {
      doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  case Strategy::BA_CrossBlock:
  case Strategy::BA_Stopper:
  case Strategy::BA_Danger:
    if(wm.countOpponentsIn(rcsc::Circle2D(wm.self().pos(),10),7,false)>=2){
      Bhv_FSClearBall().execute(agent);
      break;
    }

  case Strategy::BA_DribbleBlock:
  case Strategy::BA_DefMidField:
  case Strategy::BA_DribbleAttack:
  case Strategy::BA_OffMidField:
  case Strategy::BA_Cross:
  case Strategy::BA_ShootChance:
  default:
    Bhv_BasicOffensiveKick().execute( agent );
    break;
  }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doMove( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();

  int ball_reach_step = 0;
  if ( ! wm.existKickableTeammate()
       && ! wm.existKickableOpponent() )
    {
      ball_reach_step
	= std::min( wm.interceptTable()->teammateReachCycle(),
		    wm.interceptTable()->opponentReachCycle() );
    }
  const rcsc::Vector2D base_pos
    = wm.ball().inertiaPoint( ball_reach_step );

  rcsc::Vector2D home_pos
    = formation().getPosition( agent->config().playerNumber(),
			       base_pos );
  if ( rcsc::ServerParam::i().useOffside() )
    {
      home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

  rcsc::dlog.addText( rcsc::Logger::ROLE,
		      "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
		      ,__FILE__,
		      home_pos.x, home_pos.y,
		      base_pos.x, base_pos.y );

  switch ( Strategy::get_ball_area( base_pos ) ) {
    //////////////////////////////////////////////
  case Strategy::BA_Danger:
  case Strategy::BA_CrossBlock:
    if(wm.existKickableOpponent()){
      if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	break;

      bool save_recovery=true; 
      if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			     ,5,true)
	 <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			       ,5,false)){
	save_recovery=false;
      }
      
      int self_min = wm.interceptTable()->selfReachCycle();
      int mate_min = wm.interceptTable()->teammateReachCycle();
      int opp_min = wm.interceptTable()->opponentReachCycle();
      
      if ( ! wm.existKickableTeammate()
	   && ( self_min <= 7
		|| ( self_min <= mate_min +1
		     && self_min < opp_min+7 )
		)
	   )
	{
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: intercept"
			      ,__FILE__, __LINE__ );
	  rcsc::Body_Intercept2008(save_recovery).execute( agent );
	  if ( wm.ball().distFromSelf()
	       < rcsc::ServerParam::i().visibleDistance() ){
	    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	  }else{
	    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	  }
	  break;
	}
    }
    
    if(markFreeSpaceAtOurGoal(agent))
      break;
    
    Bhv_BasicMove( home_pos ).execute( agent );
    break;
    ////////////////////////////////////////////////
  case Strategy::BA_DribbleBlock:
  case Strategy::BA_DefMidField:
    if(wm.existKickableOpponent()){
      if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	break;

      int cycle=100;
      bool save_recovery=true;
      double dir_thr=12;
      
      if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			     ,5,true)
	 <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			       ,5,false)){
	save_recovery=false;
      }
      
      int self_min = wm.interceptTable()->selfReachCycle();
      int mate_min = wm.interceptTable()->teammateReachCycle();
      int opp_min = wm.interceptTable()->opponentReachCycle();

      if ( ! wm.existKickableTeammate()
	   && ( self_min <= 12
		|| ( self_min < mate_min + 1
		     && self_min < opp_min + 5 )
		)
	   )
	{
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: intercept"
			      ,__FILE__, __LINE__ );
	  rcsc::Body_Intercept2008(save_recovery).execute( agent );
	  if ( wm.ball().distFromSelf()
	       < rcsc::ServerParam::i().visibleDistance() )
	    {
	      agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	    }
	  else
	    {
	      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	    }
	  break;
	}

      double distToPoint;
      const rcsc::PlayerObject *opp=wm.getOpponentNearestTo(wm.ball().pos(),
							  5,&distToPoint);
      rcsc::Vector2D point;
      if(opp){
	point=opp->pos();
	point.x--;
	point.y*=0.95;
      }else{
	point=home_pos;
      }
      const double dash_power = Bhv_BasicMove::getDashPower(agent,point);
      double dist_thr = wm.ball().distFromSelf() * 0.1;
      if ( dist_thr < 1.0 ) dist_thr = 1.0;

      agent->debugClient().addMessage( " defenceMove%.0f", dash_power );
      agent->debugClient().setTarget( point );

      if ( ! rcsc::Body_GoToPoint( point, dist_thr, dash_power
				   ,cycle,false
				   ,save_recovery,dir_thr
				   ).execute( agent ) )
	{
	  if(wm.ball().posCount()<3){
	    rcsc::Body_TurnToBall().execute( agent );
	  }else{
	    point.x-=5;
	    point.y*=0.9;
	    if ( ! rcsc::Body_GoToPoint( point, dist_thr, dash_power
					 ,cycle,false
					 ,save_recovery,dir_thr
					 ).execute( agent ) ){
	      rcsc::Body_TurnToBall().execute( agent );
	    }
	  }
	}

      if ( wm.existKickableOpponent()
	   && wm.ball().distFromSelf() < 18.0 )
	{
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	}
      else
	{
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	}
      break;
    }
    Bhv_BasicMove( home_pos ).execute( agent );
    break;
    ////////////////////////////////////////
  case Strategy::BA_Stopper:
  case Strategy::BA_DribbleAttack:
  case Strategy::BA_OffMidField:
  case Strategy::BA_Cross:
  case Strategy::BA_ShootChance:
  default:
    Bhv_BasicMove( home_pos ).execute( agent );
    break;
  }
}

bool
RoleCenterBack::markFreeSpaceAtOurGoal( rcsc::PlayerAgent * agent ){
  const rcsc::WorldModel & wm = agent->world();

  ///////////////////////////////////////////////////////////////////
  //フリースペースへいくかどうか判定
  if(wm.ball().distFromSelf()>10.0
     && wm.ball().pos().x < -30
     && ( wm.ball().pos().absY()>7 
	  || wm.existTeammateIn(rcsc::Circle2D(wm.ball().pos(),5)
				,5,true)
	  && !wm.existKickableTeammate()
	  )
     ){
    ///////////////////////////////////////////////////////////////////
    //どの場所へいくか判定用。行き先の初期値
    bool right_valid=false;
    bool center_valid=false;
    bool left_valid=false;
    rcsc::Vector2D right_pos(-49.75,-rcsc::ServerParam::i().goalWidth()/2 + 2);
    rcsc::Vector2D center_pos(-49.75,0);
    rcsc::Vector2D left_pos(-49.75,rcsc::ServerParam::i().goalWidth()/2 - 2);

    ///////////////////////////////////////////////////////////////////
    //他のチームメートを見て、行き先を判定
    rcsc::PlayerPtrCont::const_iterator
      team_it = wm.teammatesFromSelf().begin();
    const rcsc::PlayerPtrCont::const_iterator
      team_end = wm.teammatesFromSelf().end();
    for(;team_it!=team_end;++team_it){
      ///////////////////////////////////////////////////////////////////
      //右の領域にチームメートがいるかどうか
      if(rcsc::Rect2D(rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				     ,-(rcsc::ServerParam::i().goalHalfWidth()
					+ 1))
		      ,rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				      + 3.5
				      ,-(rcsc::ServerParam::i().goalHalfWidth() 
					 + 1 - 5))
		      ).contains((*team_it)->pos())){
	right_valid=true;
	right_pos=(*team_it)->pos();
	continue;
      }
      ///////////////////////////////////////////////////////////////////
      //真ん中の領域にチームメートがいるかどうか
      if(rcsc::Rect2D(rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				     ,-(rcsc::ServerParam::i().goalHalfWidth()
					+ 1 -5))
		      ,rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				      + 3.5
				      ,(rcsc::ServerParam::i().goalHalfWidth() 
					 + 1 - 5))
		      ).contains((*team_it)->pos())){
	center_valid=true;
	center_pos=(*team_it)->pos();
	continue;
      }
      ///////////////////////////////////////////////////////////////////
      //左の領域にチームメートがいるかどうか
      if(rcsc::Rect2D(rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				     ,(rcsc::ServerParam::i().goalHalfWidth()
					+ 1 - 5))
		      ,rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				      + 3.5
				      ,(rcsc::ServerParam::i().goalHalfWidth() 
					 + 1))
		      ).contains((*team_it)->pos())){
	left_valid=true;
	left_pos=(*team_it)->pos();
      }
    }
  
    //どの領域にもチームメートがいるなら、別の行動をする
    if(right_valid && center_valid && left_valid)
      return false;
    ////////////////////////////////////////////////////
    //if(!right_valid || !center_valid || !left_valid)

    double dash_power=0;
    int cycle=100;
    bool back_mode=false;
    bool save_recovery=true;
    double dir_thr=12;
    if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,true)
       <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			     ,5,false)){
      dash_power=rcsc::ServerParam::i().maxPower();
      save_recovery=false;
    }

    //真ん中が空いてる
    if(!center_valid){
      	rcsc::Vector2D mark_point(-49.75,0.0);
	//両サイドがいるなら、両サイドの選手の中間
      if(right_valid && left_valid){\
	mark_point =rcsc::Vector2D(-49.75
				  ,(right_pos.y+left_pos.y)/2);
      }

      //ボールがマーク相手にきそうな時は、ボールに向かっていく
      double nearest_opp_dist = 10;
      const rcsc::PlayerObject * nearest_opp 
	= wm.getOpponentNearestTo(mark_point
				  ,5,&nearest_opp_dist);
      for(int i=0;i<7;i++){
	if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5
	   || (nearest_opp != NULL
	       && nearest_opp_dist < 7
	       && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	  save_recovery=false;
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: intercept"
			      ,__FILE__, __LINE__ );
	  rcsc::Body_Intercept2008(save_recovery).execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	  return true;
	}
      }

      dash_power = std::max(dash_power
			    ,Bhv_BasicMove::getDashPower(agent,mark_point));
      if(! rcsc::Body_GoToPoint( mark_point,
				 0.25,dash_power
				 ,cycle,back_mode
				 ,save_recovery,dir_thr
				 ).execute( agent )
	 ){
	  if(wm.self().posPrev().dist(mark_point)>1)
	    rcsc::Body_TurnToPoint(mark_point).execute(agent);
	  else
	    rcsc::Body_TurnToBall().execute(agent);
      }
	if ( wm.ball().distFromSelf() < 18.0 ){
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	}else{
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	}
	return true;
    }
    
    ////////////////////////////////////////////////////
    //if(center_valid)

    //真ん中にしかチームメートがいない(自分側サイドに移動)
    if(!right_valid && !left_valid){
      if(wm.self().pos().y<0){
	rcsc::Vector2D mark_point(-49.75
				 ,-rcsc::ServerParam::i().goalWidth()/2 + 2);

	//ボールがマーク相手にきそうな時は、ボールに向かっていく
	double nearest_opp_dist = 10;
	const rcsc::PlayerObject * nearest_opp 
	  = wm.getOpponentNearestTo(mark_point
				,5,&nearest_opp_dist);
	for(int i=0;i<7;i++){
	  if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5
	     || (nearest_opp != NULL
		 && nearest_opp_dist < 7
		 && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	    save_recovery=false;
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: intercept"
				,__FILE__, __LINE__ );
	    rcsc::Body_Intercept2008(save_recovery).execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	  return true;
	  }
	}

	dash_power = std::max(dash_power
			      ,Bhv_BasicMove::getDashPower(agent,mark_point));
	if(! rcsc::Body_GoToPoint( mark_point,
				   0.25,dash_power
				   ,cycle,back_mode
				   ,save_recovery,dir_thr
				     ).execute( agent )
	   ){
	  if(wm.self().posPrev().dist(mark_point)>1)
	    rcsc::Body_TurnToPoint(mark_point).execute(agent);
	  else
	    rcsc::Body_TurnToBall().execute(agent);
	}
	if ( wm.ball().distFromSelf() < 18.0 ){
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	}else{
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	}
	  return true;
      }else{
	rcsc::Vector2D mark_point(-49.75
				 ,rcsc::ServerParam::i().goalWidth()/2 - 2);

	//ボールがマーク相手にきそうな時は、ボールに向かっていく
	double nearest_opp_dist = 10;
	const rcsc::PlayerObject * nearest_opp 
	  = wm.getOpponentNearestTo(mark_point
				    ,5,&nearest_opp_dist);
	for(int i=0;i<7;i++){
	  if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5
	     || (nearest_opp != NULL
		 && nearest_opp_dist < 7
		 && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	    save_recovery=false;
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: intercept"
				,__FILE__, __LINE__ );
	    rcsc::Body_Intercept2008(save_recovery).execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	    return true;
	  }
	}
	
	dash_power = std::max(dash_power
			      ,Bhv_BasicMove::getDashPower(agent,mark_point));
	if(! rcsc::Body_GoToPoint( mark_point,
				   0.25,dash_power
				     ,cycle,back_mode
				   ,save_recovery,dir_thr
				   ).execute( agent )
	   ){
	  if(wm.self().posPrev().dist(mark_point)>1)
	    rcsc::Body_TurnToPoint(mark_point).execute(agent);
	  else
	    rcsc::Body_TurnToBall().execute(agent);
	  }
	if ( wm.ball().distFromSelf() < 18.0 ){
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	}else{
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	}
	return true;
      }
    }
    
    //右が空いてる
    if(!right_valid){ // && center_valid && left_valid
      rcsc::Vector2D mark_point(-49.75
			       ,-rcsc::ServerParam::i().goalWidth()/2 + 2);

      //ボールがマーク相手にきそうな時は、ボールに向かっていく
      double nearest_opp_dist = 10;
      const rcsc::PlayerObject * nearest_opp 
	= wm.getOpponentNearestTo(mark_point
				  ,5,&nearest_opp_dist);
      for(int i=0;i<7;i++){
	if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5
	   || (nearest_opp != NULL
	       && nearest_opp_dist < 7
	       && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	  save_recovery=false;
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: intercept"
			      ,__FILE__, __LINE__ );
	  rcsc::Body_Intercept2008(save_recovery).execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	  return true;
	}
      }

      dash_power = std::max(dash_power
			    ,Bhv_BasicMove::getDashPower(agent,mark_point));
      if(! rcsc::Body_GoToPoint( mark_point,
				 0.25,dash_power
				 ,cycle,back_mode
				 ,save_recovery,dir_thr
				 ).execute( agent )
	 ){
	if(wm.self().posPrev().dist(mark_point)>1)
	  rcsc::Body_TurnToPoint(mark_point).execute(agent);
	else
	  rcsc::Body_TurnToBall().execute(agent);
      }
      if ( wm.ball().distFromSelf() < 18.0 ){
	agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      }else{
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      }
      return true;
    }
    //左が空いてる
    if(!left_valid){ // && center_valid && right_valid
      rcsc::Vector2D mark_point(-49.75
				,rcsc::ServerParam::i().goalWidth()/2 - 2);
      
      //ボールがマーク相手にきそうな時は、ボールに向かっていく
      double nearest_opp_dist = 10;
      const rcsc::PlayerObject * nearest_opp 
	= wm.getOpponentNearestTo(mark_point
				  ,5,&nearest_opp_dist);
      for(int i=0;i<7;i++){
	if(wm.ball().inertiaPoint(i).dist(mark_point) < 1.5
	   || (nearest_opp != NULL
	       && nearest_opp_dist < 7
	       && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	  save_recovery=false;
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: intercept"
			      ,__FILE__, __LINE__ );
	  rcsc::Body_Intercept2008(save_recovery).execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	  return true;
	}
      }

      dash_power = std::max(dash_power
			    ,Bhv_BasicMove::getDashPower(agent,mark_point));
      if(! rcsc::Body_GoToPoint( mark_point,
				 0.25,dash_power
				 ,cycle,back_mode
				 ,save_recovery,dir_thr
				 ).execute( agent )
	 ){
	if(wm.self().posPrev().dist(mark_point)>1)
	  rcsc::Body_TurnToPoint(mark_point).execute(agent);
	else
	  rcsc::Body_TurnToBall().execute(agent);
      }
      if ( wm.ball().distFromSelf() < 18.0 ){
	agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      }else{
	agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
      }
      return true;
    }
  }
  return false;
}
