// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_defensive_half.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/basic_actions.h>

#include <rcsc/geom/sector_2d.h>

#include "bhv_fs_opponent_pass_cut_move.h"
#include "bhv_fs_clear_ball.h"

#include "bhv_selection_pass.h"

#include "bhv_fs_goto_receivable_area.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::execute( rcsc::PlayerAgent * agent )
{
  //学習の準備部分(毎サイクルやる)
  Bhv_SelectionPass().prepare(agent);
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
  //通常の処理
  switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  case Strategy::BA_CrossBlock:
  case Strategy::BA_Stopper:
  case Strategy::BA_Danger:
    if(agent->world().countOpponentsIn(rcsc::Circle2D(agent->world().self().pos(),7)
			   ,7,false)>=2){
      if(Bhv_FSClearBall().execute(agent))
	break;
    }
  case Strategy::BA_DribbleBlock:
  case Strategy::BA_DefMidField:
  case Strategy::BA_DribbleAttack:
  case Strategy::BA_OffMidField:
  case Strategy::BA_Cross:
  case Strategy::BA_ShootChance:
  default:
    {
      /////////////////////////////////////////////////////////////////////
      //学習の実行部分(学習的には、こんなことしない方がいんだけど、、、)
      rcsc::Vector2D pass_point;
      if(Bhv_SelectionPass().get_selection_pass(agent
						,&pass_point,NULL,NULL,NULL)
	 && (agent->world().self().pos().x - 10) < pass_point.x
	 && agent->world().countOpponentsIn(rcsc::Circle2D(pass_point,2.5)
					    ,7,true) <=1
	 && agent->world().countOpponentsIn(rcsc::Circle2D(agent->world().self().pos()
							   ,2.0)
					    ,7,true) >=1
	 && Bhv_SelectionPass().execute(agent))
	break;
    }
    Bhv_BasicOffensiveKick().execute( agent );
    break;
  }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleDefensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    //const rcsc::PlayerObject *opp 
    //= wm.getOpponentNearestTo(rcsc::Vector2D(-50,0),5,NULL);
    //.getOpponentNearestToSelf(5);

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y);

    //strategyの適用
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_OffMidField:
    case Strategy::BA_DefMidField:
      /*
	if((wm.self().pos().x<0)&&(wm.ball().distFromSelf()<15.0))
	{
	if (!wm.existKickableTeammate())
	{
	home_pos = rcsc::Vector2D(wm.ball().pos().x,
	wm.ball().pos().y);
	}
	}
      */

      if(wm.ball().pos().dist(wm.self().pos())<4)
	{
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
	  break;
	}
      if(wm.existKickableOpponent()
       &&Bhv_FSOpponentPassCutMove(home_pos).execute(agent))
      break;

      if(Bhv_FSGotoReceivableArea(home_pos).execute(agent))
	break;
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_Stopper:
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Danger:
    {
      if(wm.existKickableOpponent())
	{
	  if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	    break;
	  
	  int self_min = wm.interceptTable()->selfReachCycle();
	  int mate_min = wm.interceptTable()->teammateReachCycle();
	  int opp_min = wm.interceptTable()->opponentReachCycle();
	  
	  if ( ! wm.existKickableTeammate()
	       && ( self_min <= 7
		    || ( self_min <= mate_min + 1
			 && self_min < opp_min+7
			 )
		    )
	       )
	    {
	      rcsc::dlog.addText( rcsc::Logger::TEAM,
				  "%s:%d: intercept"
				  ,__FILE__, __LINE__ );
	      rcsc::Body_Intercept2008().execute( agent );
	      if ( wm.ball().distFromSelf() < rcsc::ServerParam::i().visibleDistance() )
		agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	      else
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	      
	      break;
	    }
	}
      if(markOppAtOurGoal(agent))
	break;
      if(Bhv_FSOpponentPassCutMove(home_pos).execute(agent))
	break;
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }

    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DribbleAttack:
      if(Bhv_FSGotoReceivableArea(home_pos).execute(agent))
	break;
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

bool
RoleDefensiveHalf::markOppAtOurGoal( rcsc::PlayerAgent * agent ){
  const rcsc::WorldModel & wm = agent->world();

  /////////////////////////////////////////////////////////////////////////
  //条件 (
  //自分とボールの距離が7mより遠い、
  //ボールがゴールに近い(x<-30)、
  //ボールがコーナー付近(y>7 or y<-7)
  // ) に合わないときは、reject
  if(!(wm.ball().distFromSelf() > 5.0
       && wm.ball().pos().x < -30
       && wm.ball().pos().absY() > rcsc::ServerParam::i().goalHalfWidth()))
    return false;

  ////////////////////////////////////////////////////////////////////////
  //マーク対象を探す
  rcsc::Vector2D markPoint = wm.ball().pos();
  bool markFlag=false;
  const rcsc::PlayerPtrCont::const_iterator
    opps_end = wm.opponentsFromSelf().end();
  for ( rcsc::PlayerPtrCont::const_iterator
	  opps_it = wm.opponentsFromSelf().begin();
	opps_it != opps_end; ++opps_it ){

    //ゴール前にいるパスされそうな敵以外はとばす
    if ( !((*opps_it)->pos().dist(rcsc::Vector2D(-52.5,0))<20 
	  && !(*opps_it)->isKickable()))
      continue;

    //味方がすでに、その敵をマークしてないとき、その敵をマークする
    if( !wm.existTeammateIn(rcsc::Circle2D((*opps_it)->pos()
					   ,std::max((*opps_it)->pos().dist(wm.ball().pos())*0.1
						     ,1.5)
					   )
			    ,5,true)){
      markFlag = true;
      markPoint.x = (9 * (*opps_it)->pos().x + wm.ball().pos().x) / 10;
      markPoint.y = (9 * (*opps_it)->pos().y + wm.ball().pos().y) / 10;
      break;
    }

  }
  
  if(!markFlag)
    return false;
  
  double dash_power = Bhv_BasicMove::getDashPower(agent,markPoint);
  int cycle=100;
  bool back_mode=false;
  bool save_recovery=true;
  double dir_thr=12;
  
  if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			 ,5,true)
     <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,false)){
    dash_power=std::min(dash_power*1.05,rcsc::ServerParam::i().maxPower());
    save_recovery=false;
  }
  
  //ボールがマーク相手にきそうな時は、ボールに向かっていく
  for(int i=0;i<7;i++){
    if(wm.ball().inertiaPoint(i).dist(markPoint) < 1.5){
      
      save_recovery=false;
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: intercept"
			  ,__FILE__, __LINE__ );
      rcsc::Body_Intercept2008(save_recovery).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToBall() );
      return true;
    }
  }
  
  if(! rcsc::Body_GoToPoint( markPoint,
			     0.25,dash_power
			     ,cycle,back_mode
			     ,save_recovery,dir_thr
			     ).execute( agent )
     ){
    if(wm.self().posPrev().dist(markPoint) > 1)
      rcsc::Body_TurnToPoint(markPoint).execute(agent);
    else
      rcsc::Body_TurnToBall().execute(agent);
  }
  if ( wm.ball().distFromSelf() < 18.0 ){
    agent->setNeckAction( new rcsc::Neck_TurnToBall() );
  }else{
    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
  }
  return true;
}
