// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_fs_attacker.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/audio_codec.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/geom/sector_2d.h>

#include <rcsc/math_util.h>

#include <boost/tokenizer.hpp>

#include "bhv_fs_clear_ball.h"
#include "bhv_fs_shoot.h"
#include "bhv_nutmeg_kick.h"
#include "bhv_fs_through_pass_move.h"
#include "bhv_fs_through_pass_kick.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleFSAttacker::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }
	
    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/

void
RoleFSAttacker::doKick( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject *opp = wm.getOpponentNearestToSelf(5);
    //const rcsc::PlayerObject *goalie = wm.getOpponentGoalie();
    rcsc::Vector2D pass_point;
    rcsc::Vector2D dribble_target(50.0, std::min( wm.self().pos().absY(), rcsc::ServerParam::i().goalHalfWidth()*2/3 ));
    if(wm.self().pos().y < 0)
      dribble_target.y *= -1;
    rcsc::AngleDeg dribble_angle = (dribble_target - wm.self().pos()).th();
    rcsc::Sector2D checkSector( wm.self().pos()
		     ,0.1,7.0	//r=0.1~7
		     ,dribble_angle-15,dribble_angle+15);

    //if(Bhv_NutmegKick().execute(agent))
    //	return;
    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_ShootChance:
      //こいつはここだとパスしないんだね。まぁいいか
      if(Bhv_FSShoot().execute(agent))
      break;

      rcsc::Body_Dribble2008(dribble_target,2.0,
	  		       rcsc::ServerParam::i().maxDashPower(),2,false
			       ).execute( agent );
      agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
      break;

    case Strategy::BA_Cross:
      //-----クロス時-----
      if( rcsc::Circle2D(rcsc::Vector2D(50,0),20).contains(wm.self().pos())
	  &&Bhv_FSShoot().execute(agent))
	break;

      if(countKickableCycleAtSamePoint(agent)<20){
	rcsc::Body_Dribble2008(rcsc::Vector2D(52.0,dribble_target.y)
			       ,2.0,rcsc::ServerParam::i().maxDashPower()
			       ,1,true).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
	break;
      }else{
	if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)){
	  rcsc::Body_Pass().execute(agent);
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}
      }
      Bhv_BasicOffensiveKick().execute( agent );
      break;
      //--------------------------------------------------------
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_DefMidField:
    {
      if( rcsc::Circle2D(rcsc::Vector2D(50,0),20).contains(wm.self().pos())
	 &&Bhv_FSShoot().execute(agent))
	  break;

      //セクター内に敵がいるか判定
      if ( wm.countOpponentsIn(checkSector,3,true) == 0 )
       {
        rcsc::Body_Dribble2008(dribble_target,3.0,
			       rcsc::ServerParam::i().maxDashPower(),5,false
			       ).execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
	break;
       }
/*      else if( wm.countOpponentsIn(checkSector,3,true) <=1 )
       {
        rcsc::Body_Dribble2008(dribble_target,5.0,
			       rcsc::ServerParam::i().maxDashPower(),2,true
			       ).execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );
	break;
       }*/
        if(Bhv_FSThroughPassKick().execute(agent))
	  break;

      if(opp!=NULL && (wm.self().pos().dist(opp->pos())<2.5) &&
	 (wm.self().pos().x + 1 < opp->pos().x))
	{
	  //---Pass---
	  if(rcsc::Body_Pass::get_best_pass(wm,&pass_point,NULL,NULL))
	    {
	      if(rcsc::ServerParam::i().useOffside() 
		     && pass_point.x<wm.offsideLineX()
		     || !rcsc::ServerParam::i().useOffside()
		     && wm.countOpponentsIn(rcsc::Circle2D(pass_point,2.5)
					,7,true) <=1)
		{
		  rcsc::Body_Pass().execute(agent);
		  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		  break;
		}
	    }
	  if(wm.self().pos().x<28.0)
	    {//---適当キック（for Corner）---
	      rcsc::Vector2D target_point(47.0,32.0);
	      if(wm.self().pos().y<0)
		target_point *= -1;
	      double first_speed = rcsc::ServerParam::i().ballSpeedMax();
	      double thr = 0.96;
	      int step = 2;
	      double end_speed = 0.35;
	      double target_dist = target_point.dist(wm.ball().pos());

	      if(!wm.existOpponentIn(rcsc::Circle2D(target_point,7)
				     ,7,true)
		 && !wm.existOpponentIn(rcsc::Sector2D(wm.self().pos()
						       ,target_point.dist(wm.self().pos())*0.1
						       ,target_point.dist(wm.self().pos())
						       ,(target_point - wm.self().pos()).dir() - 7
						       ,(target_point - wm.self().pos()).dir() + 7)
					,7,false)
		 )
		end_speed = 0.1;
	      first_speed
		= rcsc::calc_first_term_geom_series_last
		( end_speed,
		  target_dist,
		  rcsc::ServerParam::i().ballDecay() );
	      		
	      first_speed = std::min(first_speed
				     ,rcsc::ServerParam::i().ballSpeedMax());
	      rcsc::Body_SmartKick(target_point,
				   first_speed,
				   first_speed * thr,
				   step).execute(agent);
	    }
	  else
	    {
		Bhv_BasicOffensiveKick().execute( agent );
		break;
	    }
	  agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	  break;
	}
      else
	{
	   if(countKickableCycleAtSamePoint(agent)<20){
	     //printf("%2d:kickable\n",countKickableCycleAtSamePoint(agent));
	     rcsc::Body_Dribble2008(rcsc::Vector2D(47.0, wm.self().pos().y)
				    ,3.0,rcsc::ServerParam::i().maxPower()
				    ,3,true).execute( agent );
             agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	     break;
	   }else{
	     if(rcsc::Body_Pass::get_best_pass(wm,&pass_point,NULL,NULL))
	       {
		 rcsc::Body_Pass().execute(agent);
		 agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		 break;
	       }
	     if(Bhv_FSClearBall().execute(agent))
	       break;
	   }
	}
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    }
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    default:
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleFSAttacker::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate() && ! wm.existKickableOpponent())
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    //各種サイクルの取得
    int self_cycle = wm.interceptTable()->selfReachCycle();
    int mate_cycle = wm.interceptTable()->teammateReachCycle();
    int opp_cycle = wm.interceptTable()->opponentReachCycle();
    int min_cycle = std::min ( self_cycle, std::min( mate_cycle, opp_cycle ) );

      //---------------
     // エリアごとの処理
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_ShootChance:
    case Strategy::BA_Cross:
     if(Bhv_FSThroughPassMove().execute(agent))
	break;
     if(moveCentering(agent))
        break;
      if(!wm.existKickableTeammate()
         &&!wm.existKickableOpponent()
         &&self_cycle < mate_cycle 
         &&wm.ball().inertiaPoint( min_cycle ).x>wm.offsideLineX() )
       {//-----インターセプト-----
        if(Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) ){break;}
        rcsc::Body_Intercept2008(false).execute( agent );
        agent->setNeckAction(new rcsc::Neck_TurnToBallOrScan());
        break;
       }
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
      
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
     if(Bhv_FSThroughPassMove().execute(agent))
	break;
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}

int
RoleFSAttacker::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}

bool
RoleFSAttacker::moveCentering( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();
  static int catched_cycle = -1;
  static rcsc::Vector2D target_point(0.0,0.0);

  if( agent->world().audioMemory().freeMessageTime().cycle()
      > agent->world().time().cycle() - 3
      ){
    std::vector<rcsc::AudioMemory::FreeMessage> messages =
      agent->world().audioMemory().freeMessage();
    
    std::vector<rcsc::AudioMemory::FreeMessage>::const_iterator 
      mes_end = messages.end();
    for(std::vector<rcsc::AudioMemory::FreeMessage>::iterator mes_it 
	  = messages.begin();
	mes_it != mes_end;
	++mes_it){
      const rcsc::PlayerObject * 
	kicker = agent->world().getTeammateNearestToBall(3,false);
      if(kicker == NULL
	 || kicker->unum() != mes_it->sender_
	 || kicker->unum() == rcsc::Unum_Unknown)
	continue;
      
      // 2文字、4文字に分割してみる。
      const int offsets[] = {2,4};
      boost::offset_separator oss( offsets, offsets+2 );
      boost::tokenizer<boost::offset_separator> tokens( mes_it->message_, oss );
      boost::tokenizer<boost::offset_separator>::iterator 
	token_it = tokens.begin();
      
      if(token_it != tokens.end()
	 && *token_it == "KC"){
	++token_it;
	if(token_it == tokens.end())
	  continue;
	target_point = rcsc::AudioCodec::i().decodeStr4ToPos(*token_it);
	
	catched_cycle = agent->world().time().cycle();
	
	std::cout << agent->world().self().unum()
		  << ":[" << agent->world().time().cycle()
		  << "] MoveCentering:" << mes_it->sender_ 
		  << ":" << mes_it->message_ << " -> (" 
		  << target_point.x << "," << target_point.y 
		  << ")" << std::endl;
	break;
      }
    }
  }
  if(catched_cycle > agent->world().time().cycle() - 10
     &&!wm.existKickableTeammate()
     &&!wm.existKickableOpponent()){
    rcsc::Body_Intercept2008(false).execute( agent );
    agent->setNeckAction(new rcsc::Neck_TurnToBall());
    std::cout << agent->world().self().unum()
	      << ":[" << agent->world().time().cycle()
	      << "] MoveCentering(Intercept)" << std::endl;
    return true;
  }
  return false;
}
