// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_goalie.h"
#include "strategy.h"

#include "bhv_basic_move.h"
#include "bhv_goalie_basic_move.h"
#include "bhv_goalie_chase_ball.h"
#include "bhv_goalie_free_kick.h"

#include <rcsc/formation/formation.h>

#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/world_model.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/rect_2d.h>

#include <rcsc/action/body_pass.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_kick_one_step.h>

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleGoalie::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    static const rcsc::Rect2D 
      our_penalty( rcsc::Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
				   -rcsc::ServerParam::i().penaltyAreaHalfWidth() + 1.0 ),
		   rcsc::Size2D( rcsc::ServerParam::i().penaltyAreaLength() - 1.0,
				 rcsc::ServerParam::i().penaltyAreaWidth() - 2.0 ) );

    //////////////////////////////////////////////////////////////

    // play_on play

    // catchable
    if ( wm.time().cycle() > M_last_goalie_kick_time.cycle() + 5
         && wm.ball().distFromSelf() < agent->world().self().catchableArea() - 0.05
         && our_penalty.contains( agent->world().ball().pos() )
	 && ! wm.existKickableTeammate()
	 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            __FILE__": catchable. ball dist=%.1f, my_catchable=%.1f",
                            agent->world().ball().distFromSelf(),
                            agent->world().self().catchableArea() );
        agent->doCatch();
        agent->setNeckAction( new rcsc::Neck_TurnToBall );
    }
    else if ( agent->world().self().isKickable() )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleGoalie::doKick( rcsc::PlayerAgent * agent )
{
	//BhvGoalieFreekickとほぼ同じ
	const rcsc::WorldModel & wm = agent->world();
	rcsc::Vector2D target_point;
	double ball_speed;

	if  ( rcsc::Body_Pass::get_best_pass( agent->world(),
			&target_point,
			&ball_speed,
			NULL )
	&& target_point.dist( wm.self().pos() ) > 10.0
	)
	{
		double opp_dist = 100.0;
		const rcsc::PlayerObject * opp
		= agent->world().getOpponentNearestTo( target_point, 5, &opp_dist );
		agent->debugClient().addMessage( "KickOppDist%.1f", opp ? opp_dist : 1000.0 );
		if ( ! opp || opp_dist > 3.0 )
		{
			// enforce one step kick
		  rcsc::Body_KickOneStep( target_point,
					  ball_speed,
					  true
					  ).execute( agent ); // enforce mode
		  agent->setNeckAction( new rcsc::Neck_ScanField() );
		  return;
		}
	}
	if(rcsc::Body_Pass().execute( agent ))
	{
	  agent->setNeckAction( new rcsc::Neck_ScanField() );
	  return;
	}
	rcsc::Body_ClearBall().execute( agent );
	agent->setNeckAction( new rcsc::Neck_ScanField() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleGoalie::doMove( rcsc::PlayerAgent * agent )
{
/*
	const rcsc::WorldModel & wm = agent->world();
	int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
             && ! wm.existKickableOpponent() )
    {
            ball_reach_step
                = std::min( wm.interceptTable()->teammateReachCycle(),
                            wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
		= wm.ball().inertiaPoint( ball_reach_step );
    rcsc::Vector2D home_pos
		= formation().getPosition( agent->config().playerNumber(),
                                       base_pos );
    static const
      rcsc::Rect2D our_penalty( rcsc::Vector2D( -rcsc::ServerParam::i().pitchHalfLength(),
						-rcsc::ServerParam::i().penaltyAreaHalfWidth() + 1.0 ),
				rcsc::Size2D( rcsc::ServerParam::i().penaltyAreaLength() - 1.0,
					      rcsc::ServerParam::i().penaltyAreaWidth() - 2.0 ) );
*/
    if ( Bhv_GoalieChaseBall::is_ball_chase_situation( agent ))
    {
        Bhv_GoalieChaseBall().execute( agent );
    }
/*	未実装 自分一人で止めなければいけないとき
	else if ( Bhv_GoalieChaseBall::is_emergency_situation( agent ) )
	{
		Bhv_GoalieEmergencyMove().execute( agent );
	}
*/
    else
    {
    	Bhv_GoalieBasicMove().execute( agent );
    }
}
