// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA

 *EndCopyright
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_offensive_half.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/action/body_smart_kick.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_turn_to_ball.h>

#include "bhv_fs_clear_ball.h"
#include "bhv_selection_pass.h"
#include "bhv_fs_through_pass_kick.h"

#include "bhv_fs_mark_opponent_forward.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::execute( rcsc::PlayerAgent * agent )
{
  //学習の準備部分(毎サイクルやる)
  Bhv_SelectionPass().prepare(agent);
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }
  
  if ( kickable )
    {
      doKick( agent );
    }
  else
    {
      doMove( agent );
    }
  
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel &wm=agent->world();
    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_Cross:{
      rcsc::Vector2D pass_point;
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)
	 && Strategy::get_ball_area(pass_point) 
	 == Strategy::BA_ShootChance
	 &&(rcsc::ServerParam::i().useOffside() 
	    && pass_point.x<wm.offsideLineX()
	    || !rcsc::ServerParam::i().useOffside()))
	{//PassPoint is in ShootChanceArea
	  rcsc::Body_Pass().execute(agent);
	}
      else
	{
	  rcsc::Body_Dribble2008(rcsc::Vector2D(46.0,0),5.0,
				 rcsc::ServerParam::i().maxDashPower()*0.8,
				 3,false).execute( agent );
	}
      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
      break;
    }
    case Strategy::BA_DefMidField:
    case Strategy::BA_OffMidField:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DribbleAttack:
      if(Bhv_FSThroughPassKick().execute(agent))
	break;
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_ShootChance:
    default:
      {
	/////////////////////////////////////////////////////////////////////
	//学習の実行部分(学習的には、こんなことしない方がいんだけど、、、)
	rcsc::Vector2D pass_point;
	if(Bhv_SelectionPass().get_selection_pass(agent
						  ,&pass_point,NULL,NULL,NULL)
	   && (wm.self().pos().x - 10)<pass_point.x
	   && wm.countOpponentsIn(rcsc::Circle2D(pass_point,2.5)
				  ,7,true) <=1
	   && wm.countOpponentsIn(rcsc::Circle2D(wm.self().pos(),2.0)
				  ,7,true) >=1
	   && Bhv_SelectionPass().execute(agent))
	  break;
      }
      if(countKickableCycleAtSamePoint(agent)>20)
	if(Bhv_FSClearBall().execute(agent))
	  break;

	Bhv_BasicOffensiveKick().execute(agent);
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerObject 
      *opp_NearestFromSelf = wm.getOpponentNearestToSelf(10);

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    int self_min = wm.interceptTable()->selfReachCycle();

    if(Bhv_FSMarkOpponentForward(home_pos).execute(agent))
      return;
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
      if(Bhv_FSMarkOpponentForward(home_pos).execute(agent))
	break;
      if(!wm.existKickableTeammate()){
	if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	  break;

#if 0 
	//
	if(wm.ball().distFromSelf() > 7.0
	   && wm.ball().pos().x < -30){
	  //マーク対象を探す
	  rcsc::Vector2D markPoint = wm.ball().pos();
	  bool markFlag=false;
	  double dist;
	  
	  const rcsc::PlayerObject *opps_front = wm.getOpponentNearestTo(rcsc::ServerParam::i().ourTeamGoalPos(),5,&dist);
	  if(opps_front!=NULL){
	    
	    //味方がまだ敵をマークしてないとき、その敵をマークする
	    if(opps_front->pos().dist(wm.self().pos()) < 10
	       &&!wm.existTeammateIn(rcsc::Circle2D(opps_front->pos()
						    ,std::max(opps_front->pos().dist(wm.ball().pos())*0.1
							      ,3.0)
						    )
				     ,5,true)){
	      markFlag = true;
	      markPoint.x = (9 * opps_front->pos().x + wm.ball().pos().x) / 10;
	      markPoint.y = (9 * opps_front->pos().y + wm.ball().pos().y) / 10;
	      std::cout << wm.self().unum() << ":[" << wm.time().cycle()
			<< "] MarkFront:[" << opps_front->unum() << "]" 
		//	    << block_point.x << "," << block_point.y << ")"
			<< std::endl;
	    }else{
	      
	      const rcsc::PlayerPtrCont::const_iterator
		opps_end = wm.opponentsFromSelf().end();
	      for ( rcsc::PlayerPtrCont::const_iterator
		      opps_it = wm.opponentsFromSelf().begin();
		      opps_it != opps_end; ++opps_it ){
		
		int pole_y=7;
		if (home_pos.y < 0) pole_y*=-1;
		//ゴール前にいるパスされそうな敵以外はとばす
		if ( !((*opps_it)->pos().dist(rcsc::Vector2D(-52.5,pole_y))<20 
		       && !((*opps_it)->pos().dist(wm.self().pos())<5)
		       && !(home_pos.y * (*opps_it)->pos().y >= 0)
		       && !(*opps_it)->isKickable()))
		  continue;
		
		//味方がすでに、その敵をマークしてないとき、その敵をマークする
		if( !wm.existTeammateIn(rcsc::Circle2D((*opps_it)->pos()
						       ,std::max((*opps_it)->pos().dist(wm.ball().pos())*0.1
								 ,3.0)
						       )
					,5,true)){
		  markFlag = true;
		  markPoint.x = (9 * (*opps_it)->pos().x + wm.ball().pos().x) / 10;
		  markPoint.y = (9 * (*opps_it)->pos().y + wm.ball().pos().y) / 10;
		  /*
		    std::cout << wm.self().unum() << ":[" << wm.time().cycle()
		    << "] Mark:[" << (*opps_it)->unum() << "]" 
		    //	    << block_point.x << "," << block_point.y << ")"
		    << std::endl;
		  */
		  break;
		}
	      }
	    }
	    if(markFlag){
	      
	      double dash_power = Bhv_BasicMove::getDashPower(agent,markPoint);
	      int cycle=100;
	      bool back_mode=false;
	      bool save_recovery=true;
	      double dir_thr=12;
	      
	      if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
				     ,5,true)
		 <=wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
				       ,5,false)){
		dash_power=std::min(dash_power*1.05,rcsc::ServerParam::i().maxPower());
		save_recovery=false;
	      }
	      
	      //ボールがマーク相手にきそうな時は、ボールに向かっていく
	      for(int i=0;i<7;i++){
		if(wm.ball().inertiaPoint(i).dist(markPoint) < 1.5){
		  
		  save_recovery=false;
		  rcsc::dlog.addText( rcsc::Logger::TEAM,
				      "%s:%d: intercept"
				      ,__FILE__, __LINE__ );
		  rcsc::Body_Intercept2008(save_recovery).execute( agent );
		  agent->setNeckAction( new rcsc::Neck_TurnToBall() );
		  break;
		}
	      }
	      
	      if(! rcsc::Body_GoToPoint( markPoint,
					 0.25,dash_power
					 ,cycle,back_mode
					 ,save_recovery,dir_thr
					 ).execute( agent )
		 ){
		if(wm.self().posPrev().dist(markPoint) > 1)
		  rcsc::Body_TurnToPoint(markPoint).execute(agent);
		else
		  rcsc::Body_TurnToBall().execute(agent);
	      }
	      if ( wm.ball().distFromSelf() < 18.0 ){
		agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	      }else{
		agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	      }
	      break;
	    }
	  }
	}
#endif	
      }
      
/*
      if(wm.existKickableOpponent())
       {
	 if ( wm.ball().pos().x < -36.5 //BA_Danger.MaxX
             || ( wm.ball().pos().x < -1 //BA_DefMidField.MaxX
                 && wm.self().pos().x<wm.ball().pos().x
            )   )
          {
            if ( Bhv_BasicTackle(0.5,90.0).execute(agent)){break;}
            if ( wm.ball().distFromSelf() < 8)	//5もあった
	     {
	       rcsc::Body_Intercept2008().execute( agent );
	       agent->setNeckAction( new rcsc::Neck_TurnToBall );
	       break;
	     }
          }
    	 else
	  {
	    if(opp_NearestFromSelf!=NULL)
	     {
               home_pos = opp_NearestFromSelf->pos();//この行を消してマークをさせたい
	     }
	  }
       }
      if( self_min <= ball_reach_step
         && wm.ball().vel().x < 0
         && wm.ball().distFromSelf() < 12 )
	{ //Tonegawa: 自軍向きのこぼれ玉はとりあえず取っておく ボールのベクトルが自軍向き
	  rcsc::Body_Intercept2008(false).execute( agent );
	  agent->setNeckAction( new rcsc::Neck_TurnToBall );
	  break;
	}
*/
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
      
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

int
RoleOffensiveHalf::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}
