// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_back.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include "bhv_basic_tackle.h"
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include "bhv_fs_clear_ball.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::execute( rcsc::PlayerAgent * agent )
{
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }

  if ( kickable )
    {
      doKick( agent );
    }
  else
    {
      doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel &wm=agent->world();

  switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  case Strategy::BA_Danger:
  case Strategy::BA_Stopper:
    if(wm.countOpponentsIn(rcsc::Circle2D(wm.self().pos(),7),7,false)>=2){
      if(Bhv_FSClearBall().execute(agent))
	break;
    }
  case Strategy::BA_CrossBlock:
  case Strategy::BA_DribbleBlock:
  case Strategy::BA_DefMidField:
  case Strategy::BA_DribbleAttack:
  case Strategy::BA_OffMidField:
  case Strategy::BA_Cross:
  case Strategy::BA_ShootChance:
  default:
    Bhv_BasicOffensiveKick().execute( agent );
    break;
  }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideBack::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
      ///////////////////////////////////////////////////////
    case Strategy::BA_CrossBlock:
      
      //GoalKick時のボールを取りにいく
      if(wm.setplayCount()<100)
	{
	  int self_min = wm.interceptTable()->selfReachCycle();
	  int mate_min = wm.interceptTable()->teammateReachCycle();
	  if(wm.ball().vel().x>0 &&
	     wm.ball().distFromSelf() < 6 &&
	     self_min < mate_min + 3)
	    {
	      rcsc::Body_Intercept2008(true).execute(agent);
	      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan());
	      break;
	    }
	}
      
      //今までの処理
      if(!wm.existKickableTeammate()){
	if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	  break;

	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();

	if ( self_min <= 7
	     || ( self_min < mate_min + 1
		  && self_min < opp_min + 1 )
	     )
	  {
	    
	    bool save_recovery=true;
	    if(wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
				   ,5,true)
	       <wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
				    ,5,false)){
	      save_recovery=false;
	    }
	    
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: intercept"
				,__FILE__, __LINE__ );
	    rcsc::Body_Intercept2008(save_recovery).execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	    break;
	  }
	if(markGoal(agent))
	  break;
      }
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
      ////////////////////////////////////////////////////////////
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
      if(!wm.existKickableTeammate()){
	if( Bhv_BasicTackle(0.5,90.0).execute(agent) )
	  break;
	
	int self_min = wm.interceptTable()->selfReachCycle();
	int mate_min = wm.interceptTable()->teammateReachCycle();
	int opp_min = wm.interceptTable()->opponentReachCycle();
	
	//---opp_min補正---
	int Team_Count=0;
	int Opp_Count =0;
	
	//ボールより手前にいる味方をカウント
	rcsc::PlayerPtrCont::const_iterator
	    T_end = wm.teammatesFromSelf().end();

      for ( rcsc::PlayerPtrCont::const_iterator
	    it = wm.teammatesFromSelf().begin();
	    it != T_end;++it )
	{
	  if((*it)->pos().x<wm.ball().pos().x+5)
	    {
	      Team_Count++;
	      if(Team_Count>3)
		{
		  break;
		}
	    }
	}

      rcsc::PlayerPtrCont::const_iterator
	    O_end = wm.opponentsFromSelf().end();

      for ( rcsc::PlayerPtrCont::const_iterator
	    it = wm.opponentsFromSelf().begin();
	    it != O_end;++it )
	{
	  if((*it)->pos().x<wm.ball().pos().x+5)
	    {
	      Opp_Count++;
	      if(Opp_Count>3)
		{
		  break;
		}
	    }
	}
      int Offset = 4;
      if(Opp_Count>2)
	{
	  Offset=1;
	}
      else if(Opp_Count <= 1 && Team_Count >= 2)
	{
	  Offset=7;
	}

	//---補正終了---
	if ( self_min <= 7
	     || ( self_min < mate_min + 1
		  && self_min < opp_min + Offset)
	     )
	  {
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: intercept"
				,__FILE__, __LINE__ );
	    rcsc::Body_Intercept2008().execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	    break;
	  }
	
	if(rcsc::Rect2D(rcsc::Vector2D(-rcsc::ServerParam::i().pitchHalfLength()
				       ,-rcsc::ServerParam::i().goalHalfWidth())
			,rcsc::Vector2D(-40.0
				       ,rcsc::ServerParam::i().goalHalfWidth())
			).contains( wm.ball().pos() ))
	  {
	    if(! rcsc::Body_GoToPoint(wm.ball().pos()
				      ,0.5
				      ,rcsc::ServerParam::i().maxPower()
				      ,100,false
				      ,false,12
				      ).execute( agent )
	       ){
	      rcsc::Body_TurnToBall();
	    }
	    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
	    break;
	  }
	
	if(markGoal(agent))
	  break;
      }
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_ShootChance:
    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

bool
RoleSideBack::markGoal( rcsc::PlayerAgent * agent){
  const rcsc::WorldModel & wm = agent->world();

  int cycle=100;
  bool back_mode=false;
  bool save_recovery=true;
  double dir_thr=12;
  
  if((wm.ball().pos().dist(rcsc::Vector2D(-50,0))<15
      || wm.self().pos().dist(rcsc::Vector2D(-50,0))>15
      && wm.ball().pos().dist(wm.self().pos())>8
      && wm.ball().pos().x<-34)
     &&wm.countTeammatesIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			   ,5,true)
     <wm.countOpponentsIn(rcsc::Circle2D(rcsc::Vector2D(-50,0),15)
			  ,5,false)){
    save_recovery=false;
  }
  
  /*
  if(wm.self().pos().dist(rcsc::Vector2D(-50,0)) < 5
     && wm.self().pos().x > -50)
    back_mode = true;
  */
  
  //////////////////////////////////////////////////////////////////
  //ボールと逆サイドにいて、ボールから7mは離れてるときに行う
  if(wm.self().posPrev().y * wm.ball().pos().y < 0
     && wm.ball().pos().dist(wm.self().posPrev())>3.5){
    //ゴーリーの場所を探す
    const rcsc::PlayerPtrCont::const_iterator
      end = wm.teammatesFromSelf().end();
    double ourGoaliePosY=100;
    double ourGoaliePosX=-52.5;
    for ( rcsc::PlayerPtrCont::const_iterator
	    it = wm.teammatesFromSelf().begin();
	  it != end;++it ){
      if((*it)->goalie()){
	ourGoaliePosY=(*it)->pos().y;
	ourGoaliePosX=(*it)->pos().x;
	break;
      }
    }
    
    //ゴーリーの場所によって、移動する位置を若干変える
    double targetPointOffsetY = -3;
    if(ourGoaliePosX<-47){
      if(wm.ball().pos().x>-47){
	targetPointOffsetY = -3;
      }else{
	targetPointOffsetY = -4;
      }
    }else{
      targetPointOffsetY = -5;
    }
    rcsc::Vector2D target_point(-49.75
				,rcsc::ServerParam::i().goalWidth()/2 
				+ targetPointOffsetY);
    //ゴーリーと逆サイドにいく
    if(ourGoaliePosY>0)
      target_point.y *= -1;

    //ボールがマーク相手にきそうな時は、ボールに向かっていく
    double nearest_opp_dist = 10;
    const rcsc::PlayerObject * nearest_opp 
      = wm.getOpponentNearestTo(target_point
				,5,&nearest_opp_dist);
    for(int i=0;i<7;i++){
      if(wm.ball().inertiaPoint(i).dist(target_point) < 1.5
	 || (nearest_opp != NULL
	     && nearest_opp_dist < 7
	     && wm.ball().inertiaPoint(i).dist(nearest_opp->pos()) < 1.5)){
	
	save_recovery=false;
	rcsc::dlog.addText( rcsc::Logger::TEAM,
			    "%s:%d: intercept"
			    ,__FILE__, __LINE__ );
	rcsc::Body_Intercept2008(save_recovery).execute( agent );	
	agent->setNeckAction( new rcsc::Neck_TurnToBall() );
	return true;
      }
    }

    ////////////////////////////////////////////////////////////
    //ゴーリーが見つかったときのみ行う
    if(ourGoaliePosY<100){
      double dash_power = Bhv_BasicMove::getDashPower(agent,target_point);
      if(! rcsc::Body_GoToPoint(target_point
				,0.5
				,dash_power
				,cycle,back_mode
				,save_recovery,dir_thr
				).execute( agent )
	 ){
	if(wm.self().posPrev().dist(target_point) > 1.5){
	  rcsc::Body_TurnToPoint(target_point).execute(agent);
	}else{
	  rcsc::Body_TurnToBall();
	}
      }
      agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

      return true;
    }
  }
  return false;
}
