// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_forward.h"

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"

#include <rcsc/formation/formation.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/free_message.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/neck_turn_to_goalie_or_scan.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_go_to_point.h>

#include <rcsc/action/body_smart_kick.h>

#include "bhv_fs_clear_ball.h"
#include "bhv_fs_shoot.h"
#include "bhv_fs_through_pass_kick.h"
#include "bhv_fs_through_pass_move.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::execute( rcsc::PlayerAgent * agent )
{
  
  bool kickable = agent->world().self().isKickable();
  if ( agent->world().existKickableTeammate()
       && agent->world().teammatesFromBall().front()->distFromBall()
       < agent->world().ball().distFromSelf() )
    {
      kickable = false;
    }
  
    if ( kickable )
      {
        doKick( agent );
      }
    else
      {
        doMove( agent );
      }
    
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doKick( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::Vector2D pass_point;
    rcsc::Vector2D dribble_target(50.0, std::min( wm.self().pos().absY(), rcsc::ServerParam::i().goalHalfWidth()*2/3 ));
    if(wm.self().pos().y < 0)
      dribble_target.y *= -1;

    //nearest opponent
    //const rcsc::PlayerObject *Opp = wm.getOpponentNearestToSelf(10);
    double opp_dist = wm.getDistOpponentNearestToSelf(10);
    //RaidAttack
    bool force = false;

    double opp_offside_x = 0;
    const rcsc::PlayerPtrCont::const_iterator 
      end = wm.opponentsFromSelf().end();
    for(rcsc::PlayerPtrCont::const_iterator 
	  it = wm.opponentsFromSelf().begin();
	it != end;
	++it)
      {
	if(opp_offside_x < (*it)->pos().x
	   && !(*it)->goalie())
	  {
	    opp_offside_x = (*it)->pos().x;
	  }
     }

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
 //---------------------------ゴール間際--------------------------------
    case Strategy::BA_ShootChance:
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL))
       {
	  if(pass_point.dist(rcsc::Vector2D(52.5,0))<
	     wm.self().pos().dist(rcsc::Vector2D(52.5,0))-5)//パラメータ追加(真崎)無駄パスにならない？
	    {//--ゴールに近い味方がいたらパス--
	      if(rcsc::ServerParam::i().useOffside() 
		 && pass_point.x<wm.offsideLineX()
		 || !rcsc::ServerParam::i().useOffside()){
		rcsc::Body_Pass().execute(agent);
		agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		break;
	      }
	    }
       }
      if(wm.self().pos().dist(rcsc::Vector2D(52.5,0))<16.0)
       {
        if(Bhv_FSShoot().execute(agent))
          break;

	if(opp_dist<=5.0)force=true;
	rcsc::Body_Dribble2008(rcsc::Vector2D(52.0,0),5.0,
			     rcsc::ServerParam::i().maxDashPower(),
			     2,force).execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToGoalieOrScan() );   
	break;
       }
      Bhv_BasicOffensiveKick().execute( agent );
      break;
      //---------------------------------------------------------------
      //------------------------クロスフィールド-----------------------------
    case Strategy::BA_Cross:{
      if(rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL)
	 && Strategy::get_ball_area(pass_point) 
	 == Strategy::BA_ShootChance
	 &&(rcsc::ServerParam::i().useOffside() 
	    && pass_point.x<wm.offsideLineX()
	    || !rcsc::ServerParam::i().useOffside()))
	{//PassPoint is in ShootChanceArea
	  rcsc::Body_Pass().execute(agent);
	  agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	  break;
	}

      if( kickCentering( agent ) )
      {//センタリング
        break;
      }
      
      rcsc::Vector2D dribble_pos(52.0,0);
      if(wm.self().pos().x < -50) dribble_pos.y=wm.self().pos().y;
      
      rcsc::Body_Dribble2008(dribble_pos,1.0,
			     rcsc::ServerParam::i().maxDashPower(),
			     1,true).execute( agent );
      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
      break;
    }
      //-----------------------オフェンスフィールド--------------------------
    case Strategy::BA_OffMidField:
    case Strategy::BA_DribbleAttack:
      if(Bhv_FSThroughPassKick().execute(agent))
	break;
      //オフサイドラインを越えてたら、ゴールまで1直線
      if(wm.self().pos().x>opp_offside_x){
	if(!rcsc::Body_Dribble2008(dribble_target,4.0,
				   rcsc::ServerParam::i().maxDashPower(),5,
				   force).execute( agent )){
	}else{
	  agent->setNeckAction(new rcsc::Neck_TurnToGoalieOrScan());
	  break;
	}
      }
      //---パスが出来るのであればパス---
      if(rcsc::Body_Pass::get_best_pass(wm, &pass_point, NULL, NULL)
	 &&opp_dist<=5.0)
	{
	  if( wm.self().pos().dist(pass_point) > 6.0
	      && (wm.self().pos().x - 4.0) < pass_point.x
	      && (rcsc::ServerParam::i().useOffside() 
		  && pass_point.x<wm.offsideLineX()
		  || !rcsc::ServerParam::i().useOffside())
	      && wm.countOpponentsIn(rcsc::Circle2D(pass_point,3.0)
				     ,7,true)<=1 )
	    {
	      rcsc::Body_Pass().execute(agent);
	      agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
	      break;
	    }
	}
      if(opp_dist<=10.0)force=true;//回避モード //パラメータ変更(真崎)
      if(wm.self().pos().x<25.0)
	{//---SideLine Dribble---
	  if(wm.self().pos().y> 0)dribble_target.y= 31.0;
	  if(wm.self().pos().y< 0)dribble_target.y=-31.0;
	}
      else
	{
	  dribble_target.y = wm.self().pos().y;
	}
      if(countKickableCycleAtSamePoint(agent)<12){
	//printf("%2d:kickable\n",countKickableCycleAtSamePoint(agent));
	rcsc::Body_Dribble2008(dribble_target
			       ,2.0,rcsc::ServerParam::i().maxDashPower()
			       ,3,force).execute( agent );
      }else{
	//ドリブルできないときは、てきとークリア
	if(Bhv_FSClearBall().execute(agent))
	  break;
      }
      
      agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
      break;

    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleBlock:
      if(Bhv_FSThroughPassKick().execute(agent))
	break;
	//---Deffense region ---
	 //---パスが出きるのであればパス---
	 if(rcsc::Body_Pass::get_best_pass(wm, &pass_point, NULL, NULL)
	    &&opp_dist<=8.0)
	    { 
		if(wm.self().pos().dist(pass_point)>7.0 && 
		   (wm.self().pos().x - 5.0)<pass_point.x)
		{
		    rcsc::Body_Pass().execute(agent);
		    agent->setNeckAction(new rcsc::Neck_TurnToLowConfTeammate());
		    break;
		}
	    }
         //---そうでなければ強行突破---
	 if(opp_dist<=18.0)
	   force=true;//回避モード
	 //---SideLine Dribble---
	 dribble_target.x=10.0;
	 if(wm.self().pos().y> 0)dribble_target.y= 22.0;
	 if(wm.self().pos().y< 0)dribble_target.y=-22.0;
	  rcsc::Body_Dribble2008(dribble_target,4.0,
			     rcsc::ServerParam::i().maxDashPower(),2,
			     force).execute( agent );
	 agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	 break;
  //--------------------------その他の領域------------------------------
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    default:
      Bhv_BasicOffensiveKick().execute( agent );
      break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    //各種サイクルの取得
    int self_cycle = wm.interceptTable()->selfReachCycle();
    int mate_cycle = wm.interceptTable()->teammateReachCycle();
    int opp_cycle = wm.interceptTable()->opponentReachCycle();
    int min_cycle = std::min ( self_cycle, std::min( mate_cycle, opp_cycle ) );

    //Strategyの部分
    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_ShootChance:
    case Strategy::BA_Cross:
    case Strategy::BA_DefMidField:
     if(Bhv_FSThroughPassMove().execute(agent))
	break;
      if(!wm.existKickableTeammate()
         &&!wm.existKickableOpponent()
         &&self_cycle < mate_cycle 
         &&wm.ball().inertiaPoint( min_cycle ).x>wm.offsideLineX() )
       {//-----インターセプト-----
        if(Bhv_BasicTackle( 0.5, 90.0 ).execute( agent ) ){break;}
        rcsc::Body_Intercept2008(false).execute( agent );
        agent->setNeckAction(new rcsc::Neck_TurnToBall());
        break;
       }

	//-- カウンターのために設置　自分より前(-α)をドリブルしてる仲間がいるなら良いパスがもらえる位置までダッシュ --
	if( ( wm.existKickableTeammate()
		|| mate_cycle < opp_cycle )
		&& wm.ball().pos().x > wm.self().pos().x - 5
		&& wm.ball().pos().y * wm.self().pos().y > 0 )
	{
		rcsc::Body_GoToPoint(rcsc::Vector2D(wm.offsideLineX()-1.0, home_pos.y)
				  ,1.0,rcsc::ServerParam::i().maxPower()
				  ).execute(agent);
		agent->setNeckAction(new rcsc::Neck_TurnToBall());
		break;
	}
	
	Bhv_BasicMove( home_pos ).execute( agent );
	break;

    case Strategy::BA_DribbleBlock:
      if(Bhv_FSThroughPassMove().execute(agent))
	break;
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    default:
      Bhv_BasicMove( home_pos ).execute( agent );
      break;
    }
}

int
RoleSideForward::countKickableCycleAtSamePoint(rcsc::PlayerAgent * agent){
  static int startCycle = 0;
  static int prevCycle = 0;
  static rcsc::Vector2D point(0,0);

  if(agent->world().self().isKickable()
     && point.dist(agent->world().ball().pos()) < 3.5
     && agent->world().time().cycle() - prevCycle <= 7){
    prevCycle = agent->world().time().cycle();
  }else{
    startCycle = agent->world().time().cycle();
    prevCycle = agent->world().time().cycle();
    point = agent->world().ball().pos();
  }
  //printf("start:%2d,prev:%2d(%.2lf,%.2lf)\n"
  // ,startCycle
  // ,prevCycle,point.x,point.y);
  return agent->world().time().cycle()-startCycle;
}

bool
RoleSideForward::kickCentering( rcsc::PlayerAgent * agent )
{
  //センタリング 090704
  const rcsc::WorldModel & wm = agent->world();
  if ( !agent->config().useCommunication() )
    return false;

  rcsc::Circle2D checkCircle(rcsc::ServerParam::i().theirTeamGoalPos(),15);
  if( wm.countTeammatesIn(checkCircle,3,false)==0
      //|| wm.countOpponentsIn(checkCircle,3,false>3)
    )
    return false;
  
  double dist;
  const rcsc::PlayerObject *mate_front = wm.getTeammateNearestTo(rcsc::ServerParam::i().theirTeamGoalPos(),5,&dist);
  if(mate_front==NULL)
    return false;

  rcsc::Vector2D target_pos = mate_front->pos();
  double min_dist_from_segment=0;
  double best_target_pos_x = 0;
  
  for(int i=mate_front->pos().x + 3;i>mate_front->pos().x;i--)
    {
      double dist_from_segment=99;
      
      rcsc::Segment2D centering_course(wm.ball().pos(),target_pos);
      const rcsc::PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
      for(rcsc::PlayerPtrCont::const_iterator 
	    it = wm.opponentsFromSelf().begin();
	  it != end;
	  ++it)
	{
          double tmp_dist = centering_course.dist((*it)->pos()) + wm.self().pos().dist((*it)->pos())*0.1;
	  if ( tmp_dist < dist_from_segment ){
	    dist_from_segment = tmp_dist;
	  }
	}
      if( dist_from_segment > min_dist_from_segment )
	{
	  min_dist_from_segment = dist_from_segment;
	  best_target_pos_x = i;
	}
    }
  
  //あまりにも無謀
  if(min_dist_from_segment < 0.5) //tune this.超簡易
    return false;

  target_pos.x = best_target_pos_x;

  const double first_speed = rcsc::ServerParam::i().ballSpeedMax();
  if(rcsc::Body_SmartKick(target_pos
			  ,first_speed
			  ,first_speed * 0.96
			  ,2
			  ).execute(agent))
    {
      std::string pos_message =
	rcsc::AudioCodec::i().encodePosToStr4(target_pos); 
      
      agent->addSayMessage( new rcsc::FreeMessage<6>("KC" + pos_message) );
      agent->setNeckAction(new rcsc::Neck_ScanField());
    std::cout << agent->world().self().unum()
	      << ":[" << agent->world().time().cycle()
	      << "] KickCentering(kick):(" 
	      << target_pos.x << "," << target_pos.y 
	      << ")" << std::endl;
      return true;
    }
    return false;
}

